//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//	ChatSelector.cpp : implementation file
//

#include "stdafx.h"
#include "emule.h"
#include "ChatSelector.h"
#include "packets.h"
#include "otherfunctions.h"
#include "Smtp.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif


CChatItem::CChatItem()
{
	m_pClient = NULL;
	m_pLog = NULL;
	m_strMessagePending.Empty();
	m_bNotify = false;
	m_iHistoryIndex = 0;
}

//	CChatSelector

IMPLEMENT_DYNAMIC(CChatSelector, CClosableTabCtrl)
CChatSelector::CChatSelector()
{
	m_pCloseBtn = NULL;
	m_pMessageBox = NULL;
	m_pSendBtn = NULL;
	lastemptyicon = false;
	blinkstate = false;
	m_Timer = 0;
}

CChatSelector::~CChatSelector()
{
}


BEGIN_MESSAGE_MAP(CChatSelector, CClosableTabCtrl)
	ON_WM_TIMER()
	ON_NOTIFY_REFLECT(TCN_SELCHANGE, OnTcnSelchangeChatsel)
	ON_WM_SIZE()
	ON_BN_CLICKED(IDC_CCLOSE, OnBnClickedCclose)
	ON_BN_CLICKED(IDC_CSEND, OnBnClickedCsend)
	ON_NOTIFY(EN_LINK, 124, OnEnLinkMessageBox)
	ON_WM_DESTROY()
END_MESSAGE_MAP()

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CChatSelector message handlers
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::Init()
{
	static const uint16 s_auIconResID[] =
	{
		IDI_CHAT,
		IDI_MESSAGE,
		IDI_MPENDING
	};

	CRect rect, rClose, rSend, rMessage, rChatOut;

	GetClientRect(&rect);
	AdjustRect(FALSE, rect);

	m_pCloseBtn = GetParent()->GetDlgItem(IDC_CCLOSE);
	m_pCloseBtn->SetParent(this);
	m_pCloseBtn->GetWindowRect(&rClose);
	m_pCloseBtn->SetWindowPos( NULL, rect.right - 7 - rClose.Width(), rect.bottom - 7 - rClose.Height(),
	                           rClose.Width(), rClose.Height(), SWP_NOZORDER );
	m_pCloseBtn->EnableWindow(false);

	m_pSendBtn = GetParent()->GetDlgItem(IDC_CSEND);
	m_pSendBtn->SetParent(this);
	m_pSendBtn->GetWindowRect(&rSend);
	m_pSendBtn->SetWindowPos( NULL, rect.right - 7 - rClose.Width() - 7 - rSend.Width(), rect.bottom - 7 - rSend.Height(),
	                          rSend.Width(), rSend.Height(), SWP_NOZORDER );
	m_pSendBtn->EnableWindow(false);

	m_pMessageBox = GetParent()->GetDlgItem(IDC_CMESSAGE);
	m_pMessageBox->SetParent(this);
	m_pMessageBox->GetWindowRect(&rMessage);
	m_pMessageBox->SetWindowPos( NULL, rect.left + 7, rect.bottom - 9 - rMessage.Height(),
	                             rect.right - 7 - rClose.Width() - 7 - rSend.Width() - 21,
	                             rMessage.Height(), SWP_NOZORDER );

	int iTop = rClose.Height() > rSend.Height() ? rClose.Height() : rSend.Height();

	if (iTop < rMessage.Height())
		iTop = rMessage.Height();

	rChatOut = rect;
	rChatOut.top += 7;
	rChatOut.left += 7;
	rChatOut.right -= 18;
	rChatOut.bottom -= 7 - iTop - 7;

	ModifyStyle(0, WS_CLIPCHILDREN);
	chatout.Create(WS_VISIBLE | WS_CHILD | WS_HSCROLL | WS_VSCROLL | ES_MULTILINE | ES_READONLY | ES_NOHIDESEL, rChatOut, this, 124);
	chatout.ModifyStyleEx(0, WS_EX_STATICEDGE, SWP_FRAMECHANGED);
	chatout.AppendHyperLink(_T(""), _T(""), CLIENT_NAME, _T(""));
	chatout.AppendText(CString(" Version " CURRENT_VERSION_LONG " - ") + GetResString(IDS_CHAT_WELCOME));
	chatout.SetEventMask(chatout.GetEventMask() | ENM_LINK);
	chatout.SetFont(&g_eMuleApp.m_pdlgEmule->m_fontDefault);
	chatout.SetTitle(GetResString(IDS_CW_MESSAGES));

	m_imageList.Create(16, 16, ILC_COLOR32 | ILC_MASK, 0, 10);
//	m_imageList.SetBkColor(RGB(255, 0, 255));
	FillImgLstWith16x16Icons(&m_imageList, s_auIconResID, ELEMENT_COUNT(s_auIconResID));

	SetImageList(&m_imageList);
/*
	
	CImageList imageList;
	CBitmap Bitmap, *pOldBitmap;
	CDC *pCtrlDC = GetDC(), TempDC;

	TempDC.CreateCompatibleDC(pCtrlDC);
	Bitmap.CreateCompatibleBitmap(pCtrlDC, 16 * 3, 16);
	pOldBitmap = TempDC.SelectObject(&Bitmap);
	m_imageList.Draw(&TempDC, 0, CPoint(0, 0), ILD_NORMAL);
	m_imageList.Draw(&TempDC, 1, CPoint(16, 0), ILD_NORMAL);
	m_imageList.Draw(&TempDC, 2, CPoint(32, 0), ILD_NORMAL);
	TempDC.SelectObject(pOldBitmap);
	imageList.Create(16, 16, ILC_COLOR32 | ILC_MASK, 0, 10);
	imageList.Add(&Bitmap, RGB(255, 0, 255));
	SetImageList(&imageList);
	imageList.Detach();
	Bitmap.DeleteObject();
	ReleaseDC(pCtrlDC);
*/
	VERIFY((m_Timer = SetTimer(20, 1500, 0)));
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CChatItem* CChatSelector::StartSession(CUpDownClient* pClient)
{
	if (pClient == NULL)
		return NULL;

	EMULE_TRY

	if (GetTabByClient(pClient) != 0xFFFF)
	{
		SetCurSel(GetTabByClient(pClient));
		chatout.SetHyperText(GetItemByClient(pClient)->m_pLog);
		chatout.SetTitle(pClient->GetUserName());
		if ((chatout.m_bAutoScroll))
			chatout.ScrollToLastLine();
		return NULL;
	}

	CChatItem* pChatItem = new CChatItem();
	pChatItem->m_pLog = new CPreparedRTFText();
	pChatItem->m_pClient = pClient;

	COleDateTime timelog(COleDateTime::GetCurrentTime());
	CString sessions = GetResString(IDS_CHAT_START) + CString(pClient->GetUserName()) + _T(" (IP: ") + CString(pClient->GetFullIP()) + _T(") - ") + timelog.Format() + _T("\n");

	pChatItem->m_pLog->AppendText(sessions, RGB(255, 0, 0));
	pClient->SetChatState(MS_CHATTING);

	CString strName;

	if (!pClient->GetUserName().IsEmpty())
	{
		strName = pClient->GetUserName();

		if (strName.GetLength() > 30)
			strName = strName.Left(30) + _T("...");
	}
	else
	{
		strName.Format(_T("(%s)"), GetResString(IDS_UNKNOWN));
	}

	TCITEM newitem;

	newitem.mask = TCIF_PARAM | TCIF_TEXT | TCIF_IMAGE;
	newitem.lParam = (LPARAM)pChatItem;
	newitem.pszText = strName.GetBuffer();
	newitem.cchTextMax = strName.GetLength() + 1;
	newitem.iImage = 0;
	uint16 itemnr = InsertItem(GetItemCount(), &newitem);

	m_pCloseBtn->EnableWindow(true);
	m_pSendBtn->EnableWindow(true);

	if (IsWindowVisible())
	{
		SetCurSel(itemnr);
		chatout.SetHyperText(pChatItem->m_pLog);
		chatout.SetTitle(pClient->GetUserName());
		if ((chatout.m_bAutoScroll))
			chatout.ScrollToLastLine();
	}

	if (g_eMuleApp.m_pdlgEmule->m_pdlgActive == &g_eMuleApp.m_pdlgEmule->m_wndChat)
		m_pMessageBox->SetFocus();

	return pChatItem;

	EMULE_CATCH

	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint16 CChatSelector::GetTabByClient(CUpDownClient* client)
{
	for (int i = 0; i < GetItemCount();i++)
	{
		TCITEM cur_item;

		cur_item.mask = TCIF_PARAM;
		if (GetItem(i, &cur_item) && ((CChatItem*)cur_item.lParam)->m_pClient == client)
			return i;
	}
	return (uint16)-1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CChatItem* CChatSelector::GetItemByClient(CUpDownClient* client)
{
	for (int i = 0; i < GetItemCount();i++)
	{
		TCITEM cur_item;

		cur_item.mask = TCIF_PARAM;
		if (GetItem(i, &cur_item) && ((CChatItem*)cur_item.lParam)->m_pClient == client)
			return(CChatItem*)cur_item.lParam;
	}
	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::ProcessMessage(CUpDownClient* sender, LPCTSTR message)
{
	//filter me?
	CString Cmessage = CString(message).MakeLower();
	CString resToken;
	int curPos = 0;

	if ((Cmessage.Find(_T("emule sans upload @ http://www.chez.com/theworld/")) > -1) || (Cmessage.Find(_T("http://fullspeed.to/mison")) > -1) && (g_eMuleApp.m_pGlobPrefs->IsCounterMeasures()))
	{
		AddDebugLogLine(false, RGB_LOG_DIMMED + _T("Anti-leechermods: Client '%s' (%s) has been banned because of spamming"), sender->GetUserName(), GetClientNameAndVersionString(sender));
		sender->Ban(_T("AUTOMATED MESSAGE: Your client is being banned for spamming and leeching!"), GetResString(IDS_BAN_LEECHERSPAM));
		return;
	}

	resToken = g_eMuleApp.m_pGlobPrefs->GetMessageFilter().Tokenize(_T("|"), curPos);
	while (resToken != "")
	{
		if (Cmessage.Find(resToken.MakeLower()) > -1)
		{
			AddDebugLogLine(false, RGB_LOG_DIMMED + _T("Filtered message %s from client '%s' (%s)"), message, sender->GetUserName(), GetClientNameAndVersionString(sender));
			return;
		}
		resToken = g_eMuleApp.m_pGlobPrefs->GetMessageFilter().Tokenize(_T("|"), curPos);
	};

//	continue
	CChatItem* ci = GetItemByClient(sender);
	bool isNewChatWindow = false;

	if (!ci)
	{
		if (GetItemCount() >= 50)
		{
			AddDebugLogLine(false, RGB_LOG_WARNING_TXT _T("Instant Messaging: Messages limit reached"));		
			return;
		}

		if (g_eMuleApp.m_pGlobPrefs->GetAcceptMessagesFrom() == 4) //no messages
			return;

		if (g_eMuleApp.m_pGlobPrefs->GetAcceptMessagesFrom() == 2) //only friends
			if (!(sender->IsFriend()))
				return;

		if ((g_eMuleApp.m_pGlobPrefs->GetAcceptMessagesFrom() == 3) && (!(sender->IsFriend())))
		{ //only friends
			AddLogLine(false, GetResString(IDS_IM_MSGFROMCHAT) + _T(" %s: %s"), sender->GetUserName(), message);
			sender->SetChatState(MS_NONE);
			return;
		}

		ci = StartSession(sender);
		if (!ci)
			return;
		isNewChatWindow = true;
	}
	COleDateTime	timelog(COleDateTime::GetCurrentTime());
	CString			strTmp;

	strTmp.Format(_T("%s (%s): "), sender->GetUserName(), timelog.Format());
	ci->m_pLog->AppendText(strTmp, RGB(50, 200, 250));
	ci->m_pLog->AppendText(CString(message) + _T('\n'));

	if (g_eMuleApp.m_pGlobPrefs->GetAwayState() == true)
	{
		if ((::GetTickCount() - ci->m_pClient->GetAwayMessageResendCount()) > 3000)
		{ //send again only if 3 secs from last away message
			SendAwayMessage(ci);
			ci->m_pClient->SetAwayMessageResendCount(::GetTickCount());
		}
	}
	if (GetCurSel() != GetTabByClient(sender))
		SetItemState(GetTabByClient(sender), TCIS_HIGHLIGHTED, TCIS_HIGHLIGHTED);
	if (GetCurSel() == GetTabByClient(sender) && GetParent()->IsWindowVisible())
		chatout.SetHyperText(ci->m_pLog);
	else if (g_eMuleApp.m_pGlobPrefs->GetAwayState() == false)
	{
		ci->m_bNotify = true;
		if (isNewChatWindow || g_eMuleApp.m_pGlobPrefs->GetNotifierPopsEveryChatMsg())
		{
			CString MessageText;
			MessageText.Format("%s %s:'%s'\n", GetResString(IDS_TBN_NEWCHATMSG), sender->GetUserName(), message);
			g_eMuleApp.m_pSMTPConnection->SendMail(MessageText, g_eMuleApp.m_pGlobPrefs->GetNotifierPopsEveryChatMsg(), g_eMuleApp.m_pGlobPrefs->IsSMTPInfoEnabled());
			g_eMuleApp.m_pdlgEmule->ShowNotifier(MessageText, TBN_CHAT, false, g_eMuleApp.m_pGlobPrefs->GetNotifierPopsEveryChatMsg());
		}
		isNewChatWindow = false;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CChatSelector::SendMessage(const CString& message)
{
	CChatItem * ci = GetCurrentChatItem();

	if (ci == NULL || ci->m_pClient == NULL)
		return false;

	if (ci->m_strHistoryArray.GetCount() == g_eMuleApp.m_pGlobPrefs->GetMaxChatHistoryLines())
		ci->m_strHistoryArray.RemoveAt(0);
	ci->m_strHistoryArray.Add(CString(message));
	ci->m_iHistoryIndex = ci->m_strHistoryArray.GetCount();

	if (ci->m_pClient->GetChatState() == MS_CONNECTING)
		return false;
#ifdef OLD_SOCKETS_ENABLED
	if (ci->m_pClient->m_pRequestSocket && ci->m_pClient->m_pRequestSocket->IsConnected())
	{
		SendMessagePacket(*ci, message);

		COleDateTime	timelog(COleDateTime::GetCurrentTime());
		CString			strTmp;

		strTmp.Format(_T("%s (%s): "), g_eMuleApp.m_pGlobPrefs->GetUserNick(), timelog.Format());
		ci->m_pLog->AppendText(strTmp, RGB(1, 180, 20));
		ci->m_pLog->AppendText(message + _T('\n'));
	}
	else
	{
		ci->m_pLog->AppendText(_T("*** ") + GetResString(IDS_CONNECTING), RGB(255, 0, 0));
		ci->m_strMessagePending = message;
		ci->m_pClient->SetChatState(MS_CONNECTING);
		ci->m_pClient->TryToConnect();
	}
#endif //OLD_SOCKETS_ENABLED
	if (chatout.GetHyperText() == ci->m_pLog)
		chatout.UpdateSize();
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::ConnectingResult(CUpDownClient* sender, bool success)
{
	CChatItem * ci = GetItemByClient(sender);

	if (ci == NULL || ci->m_pClient == NULL)
		return;
	ci->m_pClient->SetChatState(MS_CHATTING);
	if (!success)
	{
		if (!ci->m_strMessagePending.IsEmpty())
		{
			ci->m_pLog->AppendText(_T(' ') + GetResString(IDS_FAILED) + _T('\n'), RGB(255, 0, 0));
			ci->m_strMessagePending.Empty();	//SyruS i wish compiler would warn on this: delete[] ci->messagepending;
		}
		else
			ci->m_pLog->AppendText(GetResString(IDS_CHATDISCONNECTED) + _T('\n'), RGB(255, 0, 0));
	}
	else
	{
		ci->m_pLog->AppendText(_T(" ok\n"), RGB(255, 0, 0));
		SendMessagePacket(*ci, ci->m_strMessagePending);

		COleDateTime	timelog(COleDateTime::GetCurrentTime());
		CString			strTmp;

		strTmp.Format(_T("%s (%s): "), g_eMuleApp.m_pGlobPrefs->GetUserNick(), timelog.Format());
		ci->m_pLog->AppendText(strTmp, RGB(1, 180, 20));
		ci->m_pLog->AppendText(ci->m_strMessagePending + _T('\n'));
		ci->m_strMessagePending.Empty();
	}
	if (chatout.GetHyperText() == ci->m_pLog)
		chatout.UpdateSize();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::DeleteAllItems()
{
	for (int i = 0; i < GetItemCount();i++)
	{
		TCITEM cur_item;

		cur_item.mask = TCIF_PARAM;
		if (GetItem(i, &cur_item))
			delete (CChatItem*)cur_item.lParam;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::OnTimer(UINT_PTR nIDEvent)
{
	blinkstate = !blinkstate;
	bool globalnotify = false;

	for (int i = 0; i < GetItemCount();i++)
	{
		TCITEM cur_item;

		cur_item.mask = TCIF_PARAM;
		if (!GetItem(i, &cur_item))
			break;

		cur_item.mask = TCIF_IMAGE;
		if (((CChatItem*)cur_item.lParam)->m_bNotify)
		{
			cur_item.iImage = (blinkstate) ? 1 : 2;
			SetItem(i, &cur_item);
			globalnotify = true;
		}
		else if (cur_item.iImage != 0)
		{
			cur_item.iImage = 0;
			SetItem(i, &cur_item);
		}
	}
	if (globalnotify)
	{
		g_eMuleApp.m_pdlgEmule->ShowMessageState(((blinkstate) ? 1 : 2));
		lastemptyicon = false;
	}
	else if (!lastemptyicon)
	{
		g_eMuleApp.m_pdlgEmule->ShowMessageState(0);
		lastemptyicon = true;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CChatItem* CChatSelector::GetCurrentChatItem()
{
	int iCurSel = GetCurSel();

	if (iCurSel == -1)
		return NULL;

	TCITEM cur_item;

	cur_item.mask = TCIF_PARAM;
	if (!GetItem(iCurSel, &cur_item))
		return NULL;
	return (CChatItem*)cur_item.lParam;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::ShowChat()
{
	if (GetCurSel() == ( -1))
		return;
	/*
	TCITEM cur_item;
	cur_item.mask = TCIF_PARAM;
	GetItem(GetCurSel(),&cur_item);*/

	CChatItem* ci = GetCurrentChatItem();
	if (!ci)
		return;

	chatout.SetHyperText(ci->m_pLog);
	chatout.SetTitle(ci->m_pClient->GetUserName());
	ci->m_bNotify = false;
	m_pMessageBox->SetFocus();
	if ((chatout.m_bAutoScroll) /*&& (GetItemState(GetCurSel(), TCIS_HIGHLIGHTED) & TCIS_HIGHLIGHTED)*/)
		chatout.ScrollToLastLine();
	SetItemState(GetCurSel(), TCIS_HIGHLIGHTED, NULL);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::OnTcnSelchangeChatsel(NMHDR *pNMHDR, LRESULT *pResult)
{
	ShowChat();
	*pResult = 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
INT	CChatSelector::InsertItem(int nItem, TCITEM* pTabCtrlItem)
{
	if (!GetItemCount())
	{
		WINDOWPLACEMENT wp;

		chatout.GetWindowPlacement(&wp);
		wp.rcNormalPosition.top += 20;
		chatout.SetWindowPlacement(&wp);
	}
	int result = CClosableTabCtrl::InsertItem(nItem, pTabCtrlItem);

	RedrawWindow();
	return result;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CChatSelector::DeleteItem(int nItem)
{
	CClosableTabCtrl::DeleteItem(nItem);
	if (GetItemCount() == 0)
	{
		WINDOWPLACEMENT wp;

		chatout.GetWindowPlacement(&wp);
		wp.rcNormalPosition.top -= 20;
		chatout.SetWindowPlacement(&wp);
		chatout.Reset();
		chatout.AppendHyperLink(_T(""), _T(""), CLIENT_NAME, _T(""));
		chatout.AppendText(CString(" Version " CURRENT_VERSION_LONG " - ") + GetResString(IDS_CHAT_WELCOME));
		chatout.SetTitle(GetResString(IDS_CW_MESSAGES));
		m_pCloseBtn->EnableWindow(false);
		m_pSendBtn->EnableWindow(false);
	}
	RedrawWindow();
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::EndSession(CUpDownClient* client)
{
	int iCurSel;

	if (client)
		iCurSel = GetTabByClient(client);
	else
		iCurSel = GetCurSel();
	if (iCurSel == -1)
		return;

	TCITEM item;
	
	item.mask = TCIF_PARAM;
	if (!GetItem(iCurSel, &item) || item.lParam == 0)
		return;
	CChatItem* ci = (CChatItem*)item.lParam;
	ci->m_pClient->SetChatState(MS_NONE);

	DeleteItem(iCurSel);
	if (chatout.GetHyperText() == ci->m_pLog)
		chatout.SetHyperText(0);
	delete ci;

	int iTabItems = GetItemCount();

	if (iTabItems > 0)
	{
		// select next tab
		if (iCurSel == CB_ERR)
			iCurSel = 0;
		else if (iCurSel >= iTabItems)
			iCurSel = iTabItems - 1;
		(void)SetCurSel(iCurSel);				// returns CB_ERR if error or no prev. selection(!)
		iCurSel = GetCurSel();					// get the real current selection
		if (iCurSel == CB_ERR)					// if still error
			iCurSel = SetCurSel(0);
		ShowChat();
	}

}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::OnSize(UINT nType, int cx, int cy)
{
	CClosableTabCtrl::OnSize(nType, cx, cy);

	CRect rect, rClose, rSend, rMessage;
	GetClientRect(&rect);
	AdjustRect(FALSE, rect);

	m_pCloseBtn->GetWindowRect(&rClose);
	m_pCloseBtn->SetWindowPos( NULL, rect.right - 7 - rClose.Width(), rect.bottom - 7 - rClose.Height(),
	                           rClose.Width(), rClose.Height(), SWP_NOZORDER );
	m_pSendBtn->GetWindowRect(&rSend);
	m_pSendBtn->SetWindowPos( NULL, rect.right - 7 - rClose.Width() - 7 - rSend.Width(), rect.bottom - 7 - rSend.Height(),
	                          rSend.Width(), rSend.Height(), SWP_NOZORDER );

	m_pMessageBox->GetWindowRect(&rMessage);
	m_pMessageBox->SetWindowPos( NULL, rect.left + 7, rect.bottom - 9 - rMessage.Height(),
	                             rect.right - 7 - rClose.Width() - 7 - rSend.Width() - 21,
	                             rMessage.Height(), SWP_NOZORDER );

	int iTop = rClose.Height() > rSend.Height() ? rClose.Height() : rSend.Height();

	if (iTop < rMessage.Height())
		iTop = rMessage.Height();

	chatout.SetWindowPos(NULL, rect.left + 7, rect.top + 7, rect.right - 18, rect.Height() - 7 - iTop - 14, SWP_NOZORDER);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::OnBnClickedCclose()
{
	EndSession();

	int iTotalTabs = GetItemCount();

	if(iTotalTabs != -1 && iTotalTabs > 0)
	{
		SetCurSel(iTotalTabs-1);
		ShowChat();
	}
	else if (iTotalTabs != -1 && iTotalTabs == 0)
	{
		SetCurSel(iTotalTabs);
		ShowChat();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::OnBnClickedCsend()
{
	EMULE_TRY

	CString strNewMessage;

	m_pMessageBox->GetWindowText(strNewMessage);
	strNewMessage.Trim();
	if (!strNewMessage.IsEmpty() && SendMessage(strNewMessage))
		m_pMessageBox->SetWindowText(_T(""));

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CChatSelector::PreTranslateMessage(MSG* pMsg)
{
	if (pMsg->message == WM_KEYDOWN)
	{

		if (pMsg->wParam == VK_RETURN)
		{
			if (pMsg->hwnd == m_pMessageBox->m_hWnd)
				OnBnClickedCsend();
		}

		if ( (pMsg->hwnd == GetDlgItem(IDC_CMESSAGE)->m_hWnd) &&
		     (pMsg->wParam == VK_UP || pMsg->wParam == VK_DOWN) )
		{
			g_eMuleApp.m_pdlgEmule->m_wndChat.ScrollHistory(pMsg->wParam == VK_DOWN);
			return TRUE;
		}
	}
	return CClosableTabCtrl::PreTranslateMessage(pMsg);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::Localize(void)
{
	if (m_hWnd)
	{
		if (m_pSendBtn)
			m_pSendBtn->SetWindowText(GetResString(IDS_CW_SEND));
		else
			GetParent()->SetDlgItemText(IDC_CSEND, GetResString(IDS_CW_SEND));
		if (m_pCloseBtn)
			m_pCloseBtn->SetWindowText(GetResString(IDS_CW_CLOSE));
		else
			GetParent()->SetDlgItemText(IDC_CCLOSE, GetResString(IDS_CW_CLOSE));
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::OnEnLinkMessageBox(NMHDR *pNMHDR, LRESULT *pResult)
{
	*pResult = 0;
	ENLINK* pEnLink = reinterpret_cast<ENLINK *>(pNMHDR);

	if (pEnLink && pEnLink->msg == WM_LBUTTONDOWN)
	{
		CString strUrl;

		chatout.GetTextRange(pEnLink->chrg.cpMin, pEnLink->chrg.cpMax, strUrl);
		CString sLangUrl = strUrl; // GetTextRange returns '?' for non-english chars :(
		sLangUrl.Remove('?'); sLangUrl.Remove(' ');
		if ((strUrl == CLIENT_NAME) || sLangUrl.IsEmpty())
			ShellExecute(NULL, NULL, _T("http://emuleplus.info"), NULL, NULL, SW_SHOWDEFAULT);
		*pResult = 1;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::OnDestroy()
{
	if (m_Timer)
		KillTimer(m_Timer);
	CClosableTabCtrl::OnDestroy();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CChatSelector::SendAwayMessage(CChatItem* ci)
{
	if (ci == NULL  || ci->m_pClient == NULL)
		return false;

	if (ci->m_pClient->GetChatState() == MS_CONNECTING)
		return false; //send AwayMessage only if connected

#ifdef OLD_SOCKETS_ENABLED
	if (ci->m_pClient->m_pRequestSocket && ci->m_pClient->m_pRequestSocket->IsConnected())
	{
		CString		strAwayMessage = CString("[AUTO REPLY:AWAY] ") + CString(g_eMuleApp.m_pGlobPrefs->GetAwayStateMessage());
		uint16		uMessageLen = _tcslen(static_cast<LPCTSTR>(strAwayMessage));
		Packet		*pPacket = new Packet(OP_MESSAGE, uMessageLen + 2);

		memcpy2(pPacket->m_pcBuffer, &uMessageLen, 2);								// <messagelen 2>
		memcpy2(pPacket->m_pcBuffer + 2, strAwayMessage, uMessageLen);				// (<messagechar 1>)*messagelen
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pPacket->m_dwSize);
		ci->m_pClient->m_pRequestSocket->SendPacket(pPacket, true, true);

		COleDateTime		timelog(COleDateTime::GetCurrentTime());
		CString				strTmp;

		strTmp.Format(_T("%s (%s): "), g_eMuleApp.m_pGlobPrefs->GetUserNick(), timelog.Format());
		ci->m_pLog->AppendText(strTmp, RGB(1, 180, 20));
		ci->m_pLog->AppendText(CString(strAwayMessage) + _T('\n'));
	}
#endif //OLD_SOCKETS_ENABLED
	if (chatout.GetHyperText() == ci->m_pLog)
		chatout.UpdateSize();
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CChatSelector::SendMessagePacket(CChatItem &ci, const CString &strMessage)
{
	uint16		uMessageLen = strMessage.GetLength();
	Packet		*pPacket = new Packet(OP_MESSAGE, uMessageLen + 2);

	memcpy2(pPacket->m_pcBuffer, &uMessageLen, 2);							//<messagelen 2>
	memcpy2(pPacket->m_pcBuffer + 2, strMessage.GetString(), uMessageLen);	//(<messagechar 1>)*messagelen
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
	ci.m_pClient->m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
