
#include "stdafx.h"
#include "emule.h"
#include "CommentDialogLst.h"
#include "PastComment.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


#define DRAWTEXTFMT		(DT_LEFT | DT_BOTTOM | DT_SINGLELINE | DT_END_ELLIPSIS | DT_NOPREFIX | DT_NOCLIP)


IMPLEMENT_DYNAMIC(CCommentDialogLst, CDialog)
CCommentDialogLst::CCommentDialogLst(CPartFile* file)
	: CResizableDialog(CCommentDialogLst::IDD, 0)
{
	m_file = file;
}

CCommentDialogLst::~CCommentDialogLst()
{
}

void CCommentDialogLst::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_LST, pmyListCtrl);
}

BEGIN_MESSAGE_MAP(CCommentDialogLst, CResizableDialog)
	ON_BN_CLICKED(IDC_CMT_OK, OnBnClickedApply)
	ON_BN_CLICKED(IDC_CMT_REF, OnBnClickedRefresh)
	ON_WM_DESTROY()
	ON_NOTIFY(NM_RCLICK, IDC_LST, OnNMRclickList)
END_MESSAGE_MAP()



void CCommentDialogLst::OnBnClickedApply()
{
	CDialog::OnOK();
}

void CCommentDialogLst::OnBnClickedRefresh()
{
	CompleteList();
}

BOOL CCommentDialogLst::OnInitDialog()
{
	CResizableDialog::OnInitDialog();

	AddAnchor(IDC_LST,TOP_LEFT,BOTTOM_RIGHT);
	AddAnchor(IDC_CMT_OK,BOTTOM_RIGHT);
	AddAnchor(IDC_CMT_REF,BOTTOM_RIGHT);
	AddAnchor(IDC_CMSTATUS,BOTTOM_LEFT);

	pmyListCtrl.InsertColumn(0, GetResString(IDS_QL_USERNAME), LVCFMT_LEFT, 100, -1);
	pmyListCtrl.InsertColumn(1, GetResString(IDS_DL_FILENAME), LVCFMT_LEFT, 150, -1);
	pmyListCtrl.InsertColumn(2, GetResString(IDS_QL_RATING), LVCFMT_LEFT, 80, 1);
	pmyListCtrl.InsertColumn(3, GetResString(IDS_COMMENT), LVCFMT_LEFT, 315, 1);

	Localize();
	CompleteList();

	return TRUE;
}

void CCommentDialogLst::Localize(void)
{
	static const int s_aiResTbl[][2] =
	{
		{ IDC_CMT_OK, IDS_CW_CLOSE },
		{ IDC_CMT_REF, IDS_CMT_REFRESH }
	};
	CString	strBuff;

	for (unsigned ui = 0; ui < ELEMENT_COUNT(s_aiResTbl); ui++)
		SetDlgItemText(s_aiResTbl[ui][0], GetResString(static_cast<UINT>(s_aiResTbl[ui][1])));

	strBuff.Format(_T("%s (%s)"), GetResString(IDS_CMT_READALL), m_file->GetFileName());
	SetWindowText(strBuff);
}

void CCommentDialogLst::CompleteList()
{
	POSITION			pos;
	CUpDownClient	   *pClient;
	int					iCount = 0;

	pmyListCtrl.DeleteAllItems();

	for (int sl = 0; sl < SOURCESSLOTS; sl++)
	{
		if (!m_file->m_srcLists[sl].IsEmpty())
		{
			for (pos = m_file->m_srcLists[sl].GetHeadPosition(); pos != NULL; )
			{
				pClient = m_file->m_srcLists[sl].GetNext(pos);

				if (pClient->GetFileComment().GetLength() > 0 || pClient->GetFileRating() != PF_RATING_NONE)
				{
					pmyListCtrl.InsertItem(iCount, pClient->GetUserName());
					pmyListCtrl.SetItemText(iCount, 1, pClient->GetClientFilename());
					pmyListCtrl.SetItemText(iCount, 2, GetRatingString(pClient->GetFileRating()));
					pmyListCtrl.SetItemText(iCount, 3, pClient->GetFileComment());
					iCount++;
				}
			}
		}
	}

	CString				strTmp = m_file->GetFileComment();
	EnumPartFileRating	eRated = m_file->GetFileRating();

	if (!strTmp.IsEmpty() || (eRated != PF_RATING_NONE))
	{
		pmyListCtrl.InsertItem(iCount, g_eMuleApp.m_pGlobPrefs->GetUserNick());
		pmyListCtrl.SetItemText(iCount, 1, m_file->GetFileName());
		pmyListCtrl.SetItemText(iCount, 2, GetRatingString(eRated));
		pmyListCtrl.SetItemText(iCount, 3, strTmp);
		iCount++;
	}

	CPastCommentList	   &pclist = m_file->GetPastCommentList();

	for (pos = pclist.GetHeadPosition(); pos != NULL; )
	{
		CPastComment &pc = pclist.GetNext(pos);

		pmyListCtrl.InsertItem(iCount, pc.GetClientName());
		pmyListCtrl.SetItemState(iCount, LVIS_CUT, LVIS_CUT);
		pmyListCtrl.SetItemText(iCount, 1, pc.GetFileName());
		pmyListCtrl.SetItemText(iCount, 2, GetRatingString(pc.GetRating()));
		pmyListCtrl.SetItemText(iCount, 3, pc.GetComment());
		iCount++;
	}

	if (iCount == 0)
	{
		strTmp.Format(_T("(%s)"), GetResString(IDS_CMT_NONE));
	}
	else
	{
		strTmp.Empty();
	}
	SetDlgItemText(IDC_CMSTATUS, strTmp);
	m_file->UpdateFileRatingCommentAvail();
}

void CCommentDialogLst::OnDestroy()
{
	CResizableDialog::OnDestroy();
}

void CCommentDialogLst::OnNMRclickList(NMHDR *pNMHDR, LRESULT *pResult)
{
	if ((pmyListCtrl.GetSelectionMark() != (-1)) && (pmyListCtrl.GetSelectedCount() == 1))
	{
		CTitleMenu	menuCmt;
		POINT		point;

		::GetCursorPos(&point);
		menuCmt.CreatePopupMenu();
		menuCmt.AddMenuTitle(GetResString(IDS_COMMENT));
		menuCmt.AppendMenu(MF_STRING, MP_MESSAGE, GetResString(IDS_COPY));
		menuCmt.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);
//	Menu objects are destroyed in their destructor
	}
	*pResult = 0;
}

BOOL CCommentDialogLst::OnCommand(WPARAM wParam,LPARAM lParam )
{
	if (pmyListCtrl.GetNextItem(-1, LVIS_SELECTED | LVIS_FOCUSED) != -1 && wParam == MP_MESSAGE)
	{
		POSITION pos = pmyListCtrl.GetFirstSelectedItemPosition();

		g_eMuleApp.CopyTextToClipboard( pmyListCtrl.GetItemText(pmyListCtrl.GetNextSelectedItem(pos), 3) );
		return true;
	}
	return CDialog::OnCommand(wParam, lParam);
}
