#include "stdafx.h"
#include "eMule.h"
#include "HTRichEditCtrl.h"
#include "otherfunctions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

static const struct
{
	LPCTSTR pszScheme;
} _apszSchemes[] =
{
	{ _T("ed2k://") },
	{ _T("ftp.") },
	{ _T("ftp://") },
	{ _T("http://") },
	{ _T("https://") },
	{ _T("mailto:") },
	{ _T("www.") },
	{ _T("www1.") },
	{ _T("www2.") },
	{ _T("www3.") },
	{ _T("www4.") },
	{ _T("www5.") }
};

CString strRtfHeader = _T("{\\rtf\\ansi");

IMPLEMENT_DYNAMIC(CHTRichEditCtrl, CRichEditCtrl)

BEGIN_MESSAGE_MAP(CHTRichEditCtrl, CRichEditCtrl)
	ON_WM_CONTEXTMENU()
	ON_WM_SETCURSOR()
	ON_WM_KEYDOWN()
	ON_WM_PAINT()
	ON_NOTIFY_REFLECT_EX(EN_LINK, OnEnLink)
	ON_WM_CREATE()
END_MESSAGE_MAP()

CHTRichEditCtrl::CHTRichEditCtrl()
{
	m_bAutoScroll = true;
	m_bNoPaint = false;
	m_pPreparedText = NULL;
	m_bInMenu = false;
	m_strTitle.Empty();
	switch (get_cpu_type())
	{
	case 1:
		m_iMaxBufSize = (32*1024);
		break;
	case 2:
		m_iMaxBufSize = (64*1024);
		break;
	default:
		if (g_eMuleApp.m_pGlobPrefs->GetWindowsVersion() == _WINVER_95_)
			m_iMaxBufSize = (64*1024);
		else
			m_iMaxBufSize = (128*1024);
		break;
	}
	m_iTextEnd = 0;
	m_hCursor = ::LoadCursor(NULL, MAKEINTRESOURCE(32512));
}

CHTRichEditCtrl::~CHTRichEditCtrl()
{
	if (m_hCursor)
		::DestroyCursor(m_hCursor);
}

void CHTRichEditCtrl::SetTitle(LPCTSTR pszTitle)
{
	m_strTitle = pszTitle;
}

void CHTRichEditCtrl::Init()
{
	VERIFY( SendMessage(EM_SETUNDOLIMIT, 0, 0) == 0 );
	if (g_eMuleApp.m_pGlobPrefs->GetWindowsVersion() == _WINVER_95_)
		LimitText((64*1024)-1);
	else
		LimitText(128*1024);
	FlushBuffer();
}

int CHTRichEditCtrl::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CRichEditCtrl::OnCreate(lpCreateStruct) == -1)
		return -1;
	Init();
	return 0;
}

LRESULT CHTRichEditCtrl::WindowProc(UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_ERASEBKGND:
			if (m_bNoPaint)
				return TRUE;
		case WM_PAINT:
			if (m_bNoPaint)
				return TRUE;
	}
	return CRichEditCtrl::WindowProc(message, wParam, lParam);
}

void CHTRichEditCtrl::FlushBuffer()
{
	if (m_astrBuff.GetSize() > 0)
	{
		for (int i = 0; i < m_astrBuff.GetSize(); i++)
			AddLine(m_astrBuff[i]);
		m_astrBuff.RemoveAll();
	}
}

void CHTRichEditCtrl::AddEntry(LPCTSTR pszMsg)
{
	CString			strLine(pszMsg);

	strLine += _T("\n");
	if (m_hWnd == NULL)
		m_astrBuff.Add(strLine);
	else
	{
		FlushBuffer();
		AddLine(strLine);
	}
}

void CHTRichEditCtrl::AddLine(LPCTSTR pszMsg, bool bLink, COLORREF crColor)
{
	long		iMsgLen = _tcslen(pszMsg);

	if (iMsgLen == 0)
		return;

	// Get cursor position
	long		iStartChar, iEndChar;

	GetSel(iStartChar, iEndChar);

	long		iOldTextEnd = m_iTextEnd;

	// Reduce flicker by ignoring WM_PAINT
	m_bNoPaint = true;

	bool		bIsVisible = IsWindowVisible();

	if (bIsVisible)
		SetRedraw(false);

	if (m_iTextEnd >= m_iMaxBufSize)
	{
		CString			strLog;

		GetWindowText(strLog);

		long			len = strLog.Find(_T('\n'), m_iMaxBufSize >> 3);

		if (len == -1)
			len = m_iMaxBufSize >> 3;
		strLog.Truncate(len);
		strLog.Remove(_T('\n'));
		len = strLog.GetLength();
		SetSel(0, len);
		ReplaceSel(_T(""));
		m_iTextEnd -= len;
	}

	// Remember where we are
	//int nFirstLine = !m_bAutoScroll ? GetFirstVisibleLine() : 0;
	POINT ptOldPos;

	SendMessage(EM_GETSCROLLPOS, 0, (LPARAM)&ptOldPos);

	// Select at the end of text and replace the selection
	// This is a very fast way to add text to an edit control
	SetSel(m_iTextEnd, m_iTextEnd);

	CHARFORMAT2			cf;

	memzero(&cf, sizeof cf);
	GetSelectionCharFormat(cf);
	if (bLink)
	{
		cf.dwMask |= CFM_LINK;
		cf.dwEffects |= CFE_LINK;
	}
	if ((crColor != -1) && (crColor != GetSysColor(COLOR_WINDOWTEXT)))
	{
		cf.dwMask |= CFM_COLOR;
		cf.dwEffects &= ~CFE_AUTOCOLOR;
		cf.crTextColor = crColor;
	}
	else
		cf.dwEffects |= CFE_AUTOCOLOR;
	SetSelectionCharFormat(cf);
	ReplaceSel(pszMsg);

	m_iTextEnd += iMsgLen;
	if (iStartChar == iOldTextEnd && iOldTextEnd == iEndChar)
		SetSel(m_iTextEnd, m_iTextEnd);
	else
		SetSel(iStartChar, iEndChar);
	if (m_bAutoScroll)
	{
		ScrollToLastLine();

		POINT			ptNewPos;

		SendMessage(EM_GETSCROLLPOS, 0, (LPARAM)&ptNewPos);
		ptOldPos.y = ptNewPos.y;
	}
	SendMessage(EM_SETSCROLLPOS, 0, (LPARAM)&ptOldPos);

	m_bNoPaint = false;
	if (bIsVisible)
	{
		SetRedraw();
		Invalidate();
	}
}

void CHTRichEditCtrl::ScrollToLastLine()
{
	//LineScroll(GetLineCount());
	long iFirstVisible = GetFirstVisibleLine();
	if (iFirstVisible > 0)
		LineScroll(-iFirstVisible, 0);
	// WM_VSCROLL does not work correctly under Win98 (or older versions of comctl.dll)
	SendMessage(WM_VSCROLL, SB_BOTTOM);
	if (m_pPreparedText == NULL)
		SendMessage(WM_VSCROLL, SB_LINEUP);
	if (g_eMuleApp.m_pGlobPrefs->GetWindowsVersion() == _WINVER_95_)
	{
		// older versions of comctl.dll seem to need this to properly update the display
		int			iPos = GetScrollPos(SB_VERT);

		SendMessage(WM_VSCROLL, MAKELONG(SB_THUMBPOSITION, iPos));
		SendMessage(WM_VSCROLL, SB_ENDSCROLL);
	}
}

void CHTRichEditCtrl::Reset()
{
	if (m_pPreparedText)
	{
		m_pPreparedText->m_sRtfHeader = strRtfHeader;
		m_pPreparedText->m_sText.Empty();
		m_pPreparedText->m_iTextEnd = 0;
		m_pPreparedText->m_aiLinkStart.RemoveAll();
		m_pPreparedText->m_aiLinkSize.RemoveAll();
	}
	m_astrBuff.RemoveAll();
	m_iTextEnd = 0;
	SetRedraw(false);
	SetWindowText(_T(""));
	switch(g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctrlBoxSwitcher.GetCurSel())
	{
		case 0:
		case 2:
			g_eMuleApp.m_pdlgEmule->m_ctlStatusBar.SetText(_T(""),SB_MESSAGETEXT,0);
			break;
		default:
			break;
	}
	SetRedraw();
	Invalidate();
}

void CHTRichEditCtrl::OnContextMenu(CWnd *pWnd, CPoint point)
{
	long			lSelStart, lSelEnd;
	CTitleMenu		menuLog;

	GetSel(lSelStart, lSelEnd);
	m_bInMenu = true;
	menuLog.CreatePopupMenu();
	menuLog.AddMenuTitle(GetResString(IDS_LOGENTRY));
	menuLog.AppendMenu(MF_STRING | ((lSelEnd > lSelStart) ? MF_ENABLED : MF_GRAYED), MP_COPYSELECTED, GetResString(IDS_COPY));

	UINT			dwFlags = MF_STRING | ((GetWindowTextLength() > 0) ? MF_ENABLED : MF_GRAYED);
	CString			strBuffer = GetResString(IDS_SAVETOFILE);

	menuLog.AppendMenu(dwFlags, MP_SAVELOG, strBuffer + _T(" (.log)"));
	menuLog.AppendMenu(dwFlags, MP_SAVERTF, strBuffer + _T(" (.rtf)"));
	menuLog.AppendMenu(MF_SEPARATOR);
	menuLog.AppendMenu(dwFlags, MP_SELECTALL, GetResString(IDS_SELECTALL));
	menuLog.AppendMenu(dwFlags, MP_REMOVEALL, GetResString(IDS_PW_RESET));

	menuLog.AppendMenu(MF_SEPARATOR);
	menuLog.AppendMenu(MF_STRING | (m_bAutoScroll ? MF_CHECKED : MF_UNCHECKED), MP_AUTOSCROLL, GetResString(IDS_AUTOSCROLL));

	menuLog.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);
	m_bInMenu = false;
//	Menu objects are destroyed in their destructor
}

BOOL CHTRichEditCtrl::OnCommand(WPARAM wParam, LPARAM lParam)
{
	switch (wParam)
	{
	case MP_COPYSELECTED:
		CopySelectedItems();
		break;
	case MP_SAVELOG:
		SaveLogToDisk();
		break;
	case MP_SAVERTF:
		SaveRtfToDisk();
		break;
	case MP_SELECTALL:
		SelectAllItems();
		break;
	case MP_REMOVEALL:
		Reset();
		break;
	case MP_AUTOSCROLL:
		m_bAutoScroll = !m_bAutoScroll;
		break;
	}
	return TRUE;
}

BOOL CHTRichEditCtrl::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message)
{
	if (m_bInMenu)
	{
		::SetCursor(m_hCursor);
		return TRUE;
	}
	return CWnd::OnSetCursor(pWnd, nHitTest, message);
}

CString CHTRichEditCtrl::GetLastLogEntry()
{
	CString			strLog;
	long			iLastLine = GetLineCount() - 2;

	if (iLastLine >= 0)
	{
		GetLine(iLastLine, strLog.GetBuffer(1024), 1024);
		strLog.ReleaseBuffer();
	}
	return strLog;
}

CString CHTRichEditCtrl::GetAllLogEntries()
{
	CString			strLog;

	GetWindowText(strLog);
	return strLog;
}

void CHTRichEditCtrl::SelectAllItems()
{
	SetSel(0, -1);
}

void CHTRichEditCtrl::SaveLogToDisk()
{
	CString			strLog, strTemp;

	if (DialogBrowseFile(strTemp, _T("Log (*.log)|*.log||"), m_strTitle + _T(".log"), 0, FALSE))
	{
		HANDLE		hFile = CreateFile(strTemp, GENERIC_WRITE, 0,NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, NULL);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			GetWindowText(strLog);
			DWORD written;
			WriteFile(hFile, strLog, strLog.GetLength(), &written, NULL);
			CloseHandle(hFile);
		}
	}
}

void CHTRichEditCtrl::SaveRtfToDisk()
{
	CString			strLog, strTemp;

	if (DialogBrowseFile(strTemp, _T("Rich Text Format (RTF)|*.rtf||"), m_strTitle + _T(".rtf"), 0, FALSE))
	{
		HANDLE		hFile = CreateFile(strTemp, GENERIC_WRITE,0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE,NULL);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			strTemp.Empty();
			strLog.Empty();

			EDITSTREAM		es;

			es.dwCookie = (DWORD)&strTemp;
			es.pfnCallback = MEditStreamOutCallback;
			StreamOut(SF_RTF,es);

			LPCTSTR			psz = strTemp;
			LPCTSTR			pszStart = psz;
			CString			sColor;
			long			iLen;
			bool			bFoundScheme;

			while (*psz != _T('\0'))
			{
				bFoundScheme = false;

				for (int i = 0; i < ARRSIZE(_apszSchemes); i++)
				{
					if (CString(psz, _tcslen(_apszSchemes[i].pszScheme)).MakeLower() == CString(_apszSchemes[i].pszScheme))
					{
						if (psz - pszStart > 0)
							strLog += CString(pszStart, psz - pszStart);
						m_pPreparedText->CheckTable(sColor, strLog, RGB(0, 0, 255));
						iLen = _tcscspn(psz, _T(" \n\r\t\\"));
						while ((iLen > 0) && !((*(psz + iLen - 1) >= _T('/') && *(psz + iLen - 1) <= _T('9')) ||
							(*(psz + iLen - 1) >= _T('A') && *(psz + iLen - 1) <= _T('Z')) ||
							(*(psz + iLen - 1) >= _T('a') && *(psz + iLen - 1) <= _T('z')) ||
							(*(psz + iLen - 1) == _T('|'))))
							iLen --;
						if (iLen == 0)
						{
							strLog += _T("{\\field{\\*\\fldinst {HYPERLINK \"") + CString(psz) + _T("\"}}{\\fldrslt{{\\ul") + sColor + CString(psz) + _T("}}}}");
							psz += _tcslen(psz);
						}
						else
						{
							strLog += _T("{\\field{\\*\\fldinst {HYPERLINK \"") + CString(psz, iLen) + _T("\"}}{\\fldrslt{{\\ul") + sColor + CString(psz, iLen) + _T("}}}}");
							psz += iLen;
						}
						pszStart = psz;
						bFoundScheme = true;
						break;
					}
				}
				if (!bFoundScheme)
					psz = _tcsinc(psz);
			}

			if (pszStart != psz)
				strLog += pszStart;

			DWORD		written;

			WriteFile(hFile, strLog, strLog.GetLength(), &written, NULL);
			CloseHandle(hFile);
		}
	}
}

void CHTRichEditCtrl::CopySelectedItems()
{
	Copy();
}

void CHTRichEditCtrl::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if (nChar == 'A' && (GetKeyState(VK_CONTROL) & 0x8000))
	{
		//////////////////////////////////////////////////////////////////
		// Ctrl+A: Select all items
		SelectAllItems();
	}
	else if (nChar == 'C' && (GetKeyState(VK_CONTROL) & 0x8000))
	{
		//////////////////////////////////////////////////////////////////
		// Ctrl+C: Copy listview items to clipboard
		CopySelectedItems();
	}

	CRichEditCtrl::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CHTRichEditCtrl::AppendText(const CString& sText, COLORREF crColor)
{
	LPCTSTR			psz = sText;
	LPCTSTR			pszStart = psz;
	long			iLen;
	bool			bFoundScheme;

	while (*psz != _T('\0'))
	{
		bFoundScheme = false;

		for (int i = 0; i < ARRSIZE(_apszSchemes); i++)
		{
			if (CString(psz, _tcslen(_apszSchemes[i].pszScheme)).MakeLower() == CString(_apszSchemes[i].pszScheme))
			{
				// output everything before the URL
				if (psz - pszStart > 0)
					AddLine(CString(pszStart, psz - pszStart), false, crColor);

				// search next space or EOL
				iLen = _tcscspn(psz, _T(" \n\r\t"));

				while ((iLen > 0) && !((*(psz + iLen - 1) >= _T('/') && *(psz + iLen - 1) <= _T('9')) ||
						(*(psz + iLen - 1) >= _T('A') && *(psz + iLen - 1) <= _T('Z')) ||
						(*(psz + iLen - 1) >= _T('a') && *(psz + iLen - 1) <= _T('z')) ||
						(*(psz + iLen - 1) == _T('|'))))
					iLen --;
				if (iLen == 0)
				{
					AddLine(psz, true);
					psz += _tcslen(psz);
				}
				else
				{
					AddLine(CString(psz, iLen), true);
					psz += iLen;
				}
				pszStart = psz;
				bFoundScheme = true;
				break;
			}
		}
		if (!bFoundScheme)
			psz = _tcsinc(psz);
	}
	if (pszStart != psz)
		AddLine(pszStart, false, crColor);
}

void CHTRichEditCtrl::AppendHyperLink(const CString& sText, const CString& sTitle, const CString& sCommand, const CString& sDirectory)
{
	ASSERT( sText.IsEmpty() );
	ASSERT( sTitle.IsEmpty() );
	ASSERT( sDirectory.IsEmpty() );
	AddLine(sCommand, true);
}

BOOL CHTRichEditCtrl::OnEnLink(NMHDR *pNMHDR, LRESULT *pResult)
{
	bool		bMsgHandled = false;
	ENLINK		*pEnLink = reinterpret_cast<ENLINK *>(pNMHDR);

	*pResult = 0;

	if (pEnLink && pEnLink->msg == WM_LBUTTONDOWN)
	{
		CString			strUrl;

		GetTextRange(pEnLink->chrg.cpMin, pEnLink->chrg.cpMax, strUrl);

		// check if that "URL" has a valid URL scheme. if it does not have, pass that notification up to the
		// parent window which may interpret that "URL" in some other way.
		for (int i = 0; i < ARRSIZE(_apszSchemes); i++)
		{
			if (strUrl.Left(_tcslen(_apszSchemes[i].pszScheme)).MakeLower() == CString(_apszSchemes[i].pszScheme))
			{
				ShellExecute(NULL, NULL, strUrl, NULL, NULL, SW_SHOWDEFAULT);
				*pResult = 1;
				bMsgHandled = true; // do not route this message to any parent
				break;
			}
		}
	}
	return bMsgHandled;
}

CString CHTRichEditCtrl::GetText() const
{
	CString			strText;

	GetWindowText(strText);
	return strText;
}

CString CHTRichEditCtrl::GetHtml()
{
	CString			strTemp, strText;
	EDITSTREAM		es;

	es.dwCookie = (DWORD)&strTemp;
	es.pfnCallback = MEditStreamOutCallback;
	StreamOut(SF_RTF, es);
	strTemp.Replace(_T("&"), _T("&amp;"));
	strTemp.Replace(_T("<"), _T("&lt;"));
	strTemp.Replace(_T(">"), _T("&gt;"));
	strTemp.Replace(_T("\""), _T("&quot;"));
	strTemp.Replace(_T("\\\'"), _T(" \\\'"));

	int				iTableStart = strTemp.Find(_T("\n{\\colortbl ;"));
	int				iTableEnd = strTemp.Find(_T(";}"), iTableStart + 1);

	if ((iTableStart != -1) && (iTableEnd != -1))
	{
		CStringArray	astrColorsRtf;
		CStringArray	astrColorsHtml;
		int				iRtfColor = 0;
		long			iHtmlColor = 0;
		int				iColorEnd;
		CString			sColor;

		while ((iTableStart < iTableEnd) && (iTableStart != -1))
		{
			iTableStart = strTemp.Find(_T("\\red"), iTableStart + 1);
			sColor.Format(_T("\\cf%i"), iRtfColor);
			astrColorsRtf.Add(sColor);
			sColor.Format(_T(" </font><font color=\"#%06x\">"), iHtmlColor);
			astrColorsHtml.Add(sColor);
			iRtfColor ++;
			if (iTableStart != -1)
			{
				iColorEnd = strTemp.Find(_T('\\'), iTableStart + 1);
				if (iColorEnd != -1)
				{
					iHtmlColor = (_tstoi(strTemp.Mid(iTableStart + 4, iColorEnd - (iTableStart + 4))) << 16);
					iTableStart = iColorEnd + 6;
					iColorEnd = strTemp.Find(_T('\\'), iTableStart + 1);
					if (iColorEnd != -1)
					{
						iHtmlColor += (_tstoi(strTemp.Mid(iTableStart, iColorEnd - iTableStart)) << 8);
						iTableStart = iColorEnd + 5;
						iColorEnd = strTemp.Find(_T(';'), iTableStart + 1);
						if (iColorEnd != -1)
							iHtmlColor += _tstoi(strTemp.Mid(iTableStart, iColorEnd - iTableStart));
					}
				}
			}
		}
		if (astrColorsRtf.GetSize() > 0)
		{
			for (int i = 0; i < astrColorsRtf.GetSize(); i++)
			{
				strTemp.Replace(astrColorsRtf[i] + _T(" \\\'"), astrColorsHtml[i] + _T(" \\\'"));
				strTemp.Replace(astrColorsRtf[i] + _T(' '), astrColorsHtml[i]);
				strTemp.Replace(astrColorsRtf[i] + _T('\\'), astrColorsHtml[i] + _T('\\'));
			}
		}

		LPCTSTR			psz = strTemp;

		psz += strTemp.Find(_T('\\'), iTableEnd);

		LPCTSTR			pszStart = psz;
		TCHAR			strChar[2];
		long			iLen;

		while (*psz != _T('\0'))
		{
			if (*psz == _T('\\'))
			{
				if (*(psz + 1) == _T('\\'))
				{
					strText += CString(pszStart, 1 + psz - pszStart);
					psz += 2;
				}
				else if (*(psz + 1) == _T('\''))
				{
					if (psz - pszStart > 0)
						strText += CString(pszStart, psz - pszStart - 1);

					strChar[0] = *(psz + 2);
					strChar[1] = *(psz + 3);
					strText += (TCHAR)_tcstol(strChar, NULL, 16);
					psz += 4;
				}
				else
				{
					if (psz - pszStart > 0)
						strText += CString(pszStart, psz - pszStart);

					iLen = _tcscspn(psz, _T(" \n"));

					if (iLen == 0)
						psz += _tcslen(psz);
					else
					{
						psz += iLen;
						if (*psz == _T(' '))
							psz++;
					}
				}
				pszStart = psz;
			}
			else
				psz++;
		}
		if (pszStart != psz)
			strText += pszStart;
	}

	int			iStrEnd = strText.ReverseFind(_T('}'));

	if (iStrEnd != -1)
		strText.Truncate(iStrEnd);
	strText.TrimRight('\n');
	strText.Replace(_T(" </font><font color"), _T("</font><font color"));

	return strText;
}

CString CHTRichEditCtrl::GetToolTip()
{
	CString			strTemp, strText;
	EDITSTREAM		es;

	es.dwCookie = (DWORD)&strTemp;
	es.pfnCallback = MEditStreamOutCallback;
	StreamOut(SF_RTF, es);
	strTemp.Replace(_T("<"), _T("<<"));
	strTemp.Replace(_T("\\\'"), _T(" \\\'"));

	int				iTableStart = strTemp.Find(_T("\n{\\colortbl ;"));
	int				iTableEnd = strTemp.Find(_T(";}"), iTableStart + 1);

	if ((iTableStart != -1) && (iTableEnd != -1))
	{
		CStringArray	astrColorsRtf;
		CStringArray	astrColorsTT;
		int				iRtfColor = 0;
		long			iHtmlColor = 0;
		int				iColorEnd;
		CString			sColor;

		while ((iTableStart < iTableEnd) && (iTableStart != -1))
		{
			iTableStart = strTemp.Find(_T("\\red"), iTableStart + 1);
			sColor.Format(_T("\\cf%i"), iRtfColor);
			astrColorsRtf.Add(sColor);
			
			if (iRtfColor == 0)
				sColor = _T(" <ct>");
			else
				sColor.Format(_T(" <ct=0x%06x>"), iHtmlColor);

			astrColorsTT.Add(sColor);
			iRtfColor ++;
			if (iTableStart != -1)
			{
				iColorEnd = strTemp.Find(_T('\\'), iTableStart + 1);
				if (iColorEnd != -1)
				{
					iHtmlColor = _tstoi(strTemp.Mid(iTableStart + 4, iColorEnd - (iTableStart + 4)));
					iTableStart = iColorEnd + 6;
					iColorEnd = strTemp.Find(_T('\\'), iTableStart + 1);
					if (iColorEnd != -1)
					{
						iHtmlColor += (_tstoi(strTemp.Mid(iTableStart, iColorEnd - iTableStart)) << 8);
						iTableStart = iColorEnd + 5;
						iColorEnd = strTemp.Find(_T(';'), iTableStart + 1);
						if (iColorEnd != -1)
							iHtmlColor += (_tstoi(strTemp.Mid(iTableStart, iColorEnd - iTableStart)) << 16);
					}
				}
			}
		}
		if (astrColorsRtf.GetSize() > 0)
		{
			for (int i = 0; i < astrColorsRtf.GetSize(); i++)
			{
				strTemp.Replace(astrColorsRtf[i] + _T(" \\\'"), astrColorsTT[i] + _T(" \\\'"));
				strTemp.Replace(astrColorsRtf[i] + _T(' '), astrColorsTT[i]);
				strTemp.Replace(astrColorsRtf[i] + _T('\\'), astrColorsTT[i] + _T('\\'));
			}
		}

		int			iLines = 0, iStrEnd = strTemp.ReverseFind(_T('}'));

		if (iStrEnd != -1)
			strTemp.Truncate(iStrEnd);
		strTemp.TrimRight('\n');

		LPCTSTR		pszStart = strTemp;
		LPCTSTR		psz = pszStart;

		pszStart += strTemp.Find(_T('\\'), iTableEnd);
		psz += strTemp.GetLength() - 1;

		CString		strLog, strStatus = g_eMuleApp.m_pdlgEmule->m_ctlStatusBar.GetText(SB_MESSAGETEXT, 0);
		int			iLastFound, iNextFound, iCount = 11;

		GetWindowText(strLog);
		iLastFound = strLog.Find(strStatus);
		iNextFound = iLastFound;
		while (iNextFound != -1)
		{
			iLastFound = iNextFound;
			iNextFound = strLog.Find(strStatus, iLastFound + 1);
		}
		
		if (iLastFound != -1)
		{
			iCount = 0;
			iNextFound = strLog.Find(_T("\n"), iLastFound + 1);
			while (iNextFound != -1)
			{
				iCount++;
				iNextFound = strLog.Find(_T("\n"), iNextFound + 1);
			}
			
			while (iCount > 0 && psz > pszStart)
			{
				if (*psz == _T('\n'))
					iCount--;
				psz--;
			}
			iCount = 10 - iCount;
			strTemp.Truncate(psz - strTemp);
		}

		while (iLines < iCount && psz > pszStart)
		{
			if (*psz == _T('\n'))
				iLines++;
			psz--;
		}

		if (psz > pszStart)
			psz++;

		pszStart = psz;

		TCHAR		strChar[2];
		long		iLen;

		while (*psz != _T('\0'))
		{
			if (*psz == _T('\\'))
			{
				if (*(psz + 1) == _T('\\'))
				{
					strText += CString(pszStart, 1 + psz - pszStart);
					psz += 2;
				}
				else if (*(psz + 1) == _T('\''))
				{
					if (psz - pszStart > 0)
						strText += CString(pszStart, psz - pszStart - 1);

					strChar[0] = *(psz + 2);
					strChar[1] = *(psz + 3);
					strText += (TCHAR)_tcstol(strChar, NULL, 16);
					psz += 4;
				}
				else
				{
					if (psz - pszStart > 0)
						strText += CString(pszStart, psz - pszStart);

					iLen = _tcscspn(psz, _T(" \n"));

					if (iLen == 0)
						psz += _tcslen(psz);
					else
					{
						psz += iLen;
						if (*psz == _T(' '))
							psz++;
					}
				}
				pszStart = psz;
			}
			else
				psz++;
		}
		if (pszStart != psz)
			strText += pszStart;
	}

	strText.Trim(_T('\n'));
	strText.Replace(_T(" <ct"), _T("<ct"));
	while (strText.Right(4) == _T("<ct>"))
		strText.Truncate(strText.GetLength() - 4);
	strText.TrimRight(_T('\n'));
	strText.Replace(_T("\n"), _T("<br>"));
	if (strText.GetAt(0) != _T('<'))
		strText.Insert(0, _T("<"));

	return strText;
}

CPreparedRTFText* CHTRichEditCtrl::GetHyperText()
{
	return m_pPreparedText;
}

void CHTRichEditCtrl::SetHyperText(CPreparedRTFText* pPreparedText, bool bInvalidate)
{
	m_pPreparedText = pPreparedText;
	if (bInvalidate)
		UpdateSize();
}

void CHTRichEditCtrl::UpdateSize()
{
	if (m_pPreparedText == NULL)
		return;

	POINT		ptOldPos;
	long		iStartChar, iEndChar;

	SendMessage(EM_GETSCROLLPOS, 0, (LPARAM)&ptOldPos);
	GetSel(iStartChar, iEndChar);

	bool		bIsVisible = IsWindowVisible();

	if (bIsVisible)
		SetRedraw(FALSE);

	long		iLastEnd = GetTextLength() - GetLineCount() + 1;
	CString		m_sOutText = m_pPreparedText->m_sRtfHeader + m_pPreparedText->m_sText + _T("}");

	SETTEXTEX st = {0};
	st.flags = ST_DEFAULT;
	st.codepage = CP_ACP;
	SendMessage(EM_SETTEXTEX, (WPARAM)&st, (LPARAM)(LPCTSTR)m_sOutText);
	if (m_pPreparedText->m_aiLinkStart.GetSize() > 0)
	{
		CHARFORMAT2		cf;

		for (int i = 0; i < m_pPreparedText->m_aiLinkStart.GetSize(); i++)
		{
			SetSel(m_pPreparedText->m_aiLinkStart[i], m_pPreparedText->m_aiLinkStart[i] + m_pPreparedText->m_aiLinkSize[i]);
			memzero(&cf, sizeof cf);
			GetSelectionCharFormat(cf);
			cf.dwMask |= CFM_LINK;
			cf.dwEffects |= CFE_LINK;
			SetSelectionCharFormat(cf);
		}
	}
	if (iStartChar == iLastEnd && iLastEnd == iEndChar)
		SetSel(m_pPreparedText->m_iTextEnd, m_pPreparedText->m_iTextEnd);
	else
		SetSel(iStartChar, iEndChar);
	if (m_bAutoScroll)
	{
		ScrollToLastLine();

		POINT			ptNewPos;

		SendMessage(EM_GETSCROLLPOS, 0, (LPARAM)&ptNewPos);
		ptOldPos.y = ptNewPos.y;
	}
	SendMessage(EM_SETSCROLLPOS, 0, (LPARAM)&ptOldPos);

	if (bIsVisible)
	{
		SetRedraw();
		Invalidate();
	}
}

void CHTRichEditCtrl::SetFont(CFont* pFont, BOOL bRedraw)
{
	LOGFONT		lf = {0};

	pFont->GetLogFont(&lf);

	CHARFORMAT	cf = {0};

	cf.cbSize = sizeof cf;

	cf.dwMask |= CFM_BOLD;
	cf.dwEffects |= (lf.lfWeight == FW_BOLD) ? CFE_BOLD : 0;

	cf.dwMask |= CFM_ITALIC;
	cf.dwEffects |= (lf.lfItalic) ? CFE_ITALIC : 0;

	cf.dwMask |= CFM_UNDERLINE;
	cf.dwEffects |= (lf.lfUnderline) ? CFE_UNDERLINE : 0;

	cf.dwMask |= CFM_STRIKEOUT;
	cf.dwEffects |= (lf.lfStrikeOut) ? CFE_STRIKEOUT : 0;

	cf.dwMask |= CFM_SIZE;

	HDC			hDC = ::GetDC(NULL);
	int			iPointSize = -MulDiv(lf.lfHeight, 72, GetDeviceCaps(hDC, LOGPIXELSY));

	cf.yHeight = iPointSize * 20;
	::ReleaseDC(NULL, hDC);

	cf.dwMask |= CFM_FACE;
	cf.bPitchAndFamily = lf.lfPitchAndFamily;
	_tcsncpy(cf.szFaceName, lf.lfFaceName, ARRSIZE(cf.szFaceName));
	cf.szFaceName[ARRSIZE(cf.szFaceName) - 1] = _T('\0');

	// although this should work correctly (according SDK) it may give false results (e.g. the "click here..." text
	// which is shown in the server info window may not be entirely used as a hyperlink???)
	//cf.dwMask |= CFM_CHARSET;
	//cf.bCharSet = lf.lfCharSet;

	cf.yOffset = 0;
	VERIFY( SetDefaultCharFormat(cf) );

	if (bRedraw)
	{
		Invalidate();
		UpdateWindow();
	}
}

CFont* CHTRichEditCtrl::GetFont() const
{
	ASSERT(0);
	return NULL;
}


///////////////////////////////////////////////////////////////////////////////
// CPreparedRTFText

CPreparedRTFText::CPreparedRTFText()
{
	m_sRtfHeader = strRtfHeader;
	m_sText.Empty();
	m_iTextEnd = 0;
	m_aiLinkStart.RemoveAll();
	m_aiLinkSize.RemoveAll();
	switch (get_cpu_type())
	{
	case 1:
		m_iMaxBufSize = (16*1024);
		break;
	case 2:
		m_iMaxBufSize = (24*1024);
		break;
	default:
		m_iMaxBufSize = (32*1024);
		break;
	}
}

CPreparedRTFText::~CPreparedRTFText()
{
}

void CPreparedRTFText::CheckTable(CString& sColor, CString& sTable, COLORREF crColor) const
{
	int			iColor = 0;

	if (((crColor & 0x1000000) == 0) && ((crColor & 0xFFFFFF) != GetSysColor(COLOR_WINDOWTEXT)))
	{
		sColor.Format(_T("\\red%i\\green%i\\blue%i;"), GetRValue(crColor), GetGValue(crColor), GetBValue(crColor));

		int iTableStart = sTable.Find(_T("\n{\\colortbl ;"));
		if (iTableStart == -1)
		{
			sTable += _T("\n{\\colortbl ;") + sColor + _T("}\n");
			iColor ++;
		}
		else
		{
			int			iTableEnd = sTable.Find(_T(";}"), iTableStart + 1);

			while ((iTableStart < iTableEnd) && (iTableStart != -1))
			{
				iColor ++;
				iTableStart = sTable.Find(_T("\\red"), iTableStart + 1);
				if ((iTableStart >= 0) && (sTable.Mid(iTableStart, sColor.GetLength()) == sColor))
				{
					sColor.Format(_T("\\cf%i "), iColor);
					return;
				}
			}
			sTable.Insert(iTableEnd + 1, sColor);
		}
	}
	sColor.Format(_T("\\cf%i "), iColor);
}

void CPreparedRTFText::AppendText(const CString& strLine, COLORREF crColor)
{
	long		len = m_sText.GetLength();

	if (len >= m_iMaxBufSize)
	{
		len	= m_sText.Find(_T("\n\\par"), (m_iMaxBufSize >> 2)) + 5;

		CString			strCut = m_sText.Left(len);

		m_sText.Delete(0, len);
		strCut.Replace(_T("\n\\par"), _T("\n"));
		strCut.Replace(_T("\\b0 "), _T(""));
		strCut.Replace(_T("\\b "), _T(""));
		strCut.Replace(_T("\\i0 "), _T(""));
		strCut.Replace(_T("\\i "), _T(""));
		strCut.Replace(_T("\\ul0 "), _T(""));
		strCut.Replace(_T("\\ul "), _T(""));

		int			iTableStart = m_sRtfHeader.Find(_T("\n{\\colortbl ;"));
		int			iTableEnd = m_sRtfHeader.Find(_T(";}"), iTableStart + 1);
		CString		sColor;
		int			iColor = 0;

		while ((iTableStart < iTableEnd) && (iTableStart != -1))
		{
			iTableStart = m_sRtfHeader.Find(_T("\\red"), iTableStart + 1);
			sColor.Format(_T("\\cf%i "), iColor);
			strCut.Replace(sColor, _T(""));
			iColor ++;
		}
		strCut.Replace(_T("\\}"),_T("}"));
		strCut.Replace(_T("\\{"),_T("{"));
		strCut.Replace(_T("\\\\"),_T("\\"));
		len = strCut.GetLength();
		m_iTextEnd -= len;

		if (m_aiLinkStart.GetSize() > 0)
		{
			int			i = 0;

			while (i < m_aiLinkStart.GetSize())
			{
				if (m_aiLinkStart[i] >= (unsigned int)len)
				{
					m_aiLinkStart[i] -= len;
					i++;
				}
				else
				{
					m_aiLinkStart.RemoveAt(i);
					m_aiLinkSize.RemoveAt(i);
				}
			}
		}
	}
	CString			sColor, sColor2, sText = strLine;

	CheckTable(sColor, m_sRtfHeader, crColor);
	if ((crColor & 0x2000000) == 0)
		sText.Replace(_T("\\"),_T("\\\\"));
	sText.Replace(_T("{"),_T("\\{"));
	sText.Replace(_T("}"),_T("\\}"));
	sText.Replace(_T("\n"),_T("\n\\par"));
	sText.Insert(0, sColor);

	LPCTSTR			psz = sText;
	LPCTSTR			pszStart = psz;
	bool			bFoundScheme;
	long			iLen;

	while (*psz != _T('\0'))
	{
		bFoundScheme = false;

		for (int i = 0; i < ARRSIZE(_apszSchemes); i++)
		{
			if (CString(psz, _tcslen(_apszSchemes[i].pszScheme)).MakeLower() == CString(_apszSchemes[i].pszScheme))
			{
				if (psz - pszStart > 0)
				{
					CString			strClean = CString(pszStart, psz - pszStart);

					strClean.Replace(_T("\\b0 "),_T(""));
					strClean.Replace(_T("\\b "),_T(""));
					strClean.Replace(_T("\\i0 "),_T(""));
					strClean.Replace(_T("\\i "),_T(""));
					strClean.Replace(_T("\\ul0 "),_T(""));
					strClean.Replace(_T("\\ul "),_T(""));
					strClean.Replace(_T("\n\\par"), _T("\n"));
					strClean.Replace(_T("\\}"), _T("}"));
					strClean.Replace(_T("\\{"), _T("{"));
					strClean.Replace(_T("\\\\"), _T("\\"));
					strClean.Replace(sColor, _T(""));
					m_iTextEnd += strClean.GetLength();
					m_sText += CString(pszStart, psz - pszStart);
				}
				CheckTable(sColor2, m_sRtfHeader, RGB(0, 0, 255));

				iLen = _tcscspn(psz, _T(" \n\r\t\\"));
				while ((iLen > 0) && !((*(psz + iLen - 1) >= _T('/') && *(psz + iLen - 1) <= _T('9')) ||
						(*(psz + iLen - 1) >= _T('A') && *(psz + iLen - 1) <= _T('Z')) ||
						(*(psz + iLen - 1) >= _T('a') && *(psz + iLen - 1) <= _T('z')) ||
						(*(psz + iLen - 1) == _T('|'))))
					iLen --;
				if (iLen == 0)
				{
					m_aiLinkStart.Add(m_iTextEnd);
					m_aiLinkSize.Add(_tcslen(psz));
					m_iTextEnd += _tcslen(psz);
					m_sText += _T("\\ul ") + sColor2 + CString(psz) + _T("\\ul0 ") + sColor;
					psz += _tcslen(psz);
				}
				else
				{
					m_aiLinkStart.Add(m_iTextEnd);
					m_aiLinkSize.Add(iLen);
					m_iTextEnd += iLen;
					m_sText += _T("\\ul ") + sColor2 + CString(psz, iLen) + _T("\\ul0 ") + sColor;
					psz += iLen;
				}
				pszStart = psz;
				bFoundScheme = true;
				break;
			}
		}
		if (!bFoundScheme)
			psz = _tcsinc(psz);
	}
	if (pszStart != psz)
	{
		CString			strClean = CString(pszStart);

		strClean.Replace(_T("\\b0 "), _T(""));
		strClean.Replace(_T("\\b "), _T(""));
		strClean.Replace(_T("\\i0 "), _T(""));
		strClean.Replace(_T("\\i "), _T(""));
		strClean.Replace(_T("\\ul0 "), _T(""));
		strClean.Replace(_T("\\ul "), _T(""));
		strClean.Replace(_T("\n\\par"), _T("\n"));
		strClean.Replace(_T("\\}"), _T("}"));
		strClean.Replace(_T("\\{"), _T("{"));
		strClean.Replace(_T("\\\\"), _T("\\"));
		strClean.Replace(sColor, _T(""));
		m_iTextEnd += strClean.GetLength();
		m_sText += pszStart;
	}
}

const CString& CPreparedRTFText::GetText()
{
	return m_sText;
}
