//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//	IrcWnd.cpp : implementation file
//

#include "stdafx.h"
#include "IrcWnd.h"
#include "emule.h"
#include "otherfunctions.h"
#include "opcodes.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

static const COLORREF _acrColorPal[16] = {
	RGB(255, 255, 255), RGB(0, 0, 0), RGB(0, 0, 127), RGB(0, 127, 0),
	RGB(255, 0, 0), RGB(127, 0, 0), RGB(156, 0, 156), RGB(252, 127, 0),
	RGB(255, 255, 0), RGB(0, 252, 0), RGB(0, 147, 147), RGB(0, 255, 255),
	RGB(0, 0, 252), RGB(255, 0, 255), RGB(127, 127, 127), RGB(210, 210, 210)
};

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	CIrcWnd dialog
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
IMPLEMENT_DYNAMIC(CIrcWnd, CDialog)
CIrcWnd::CIrcWnd(CWnd* pParent /*=NULL*/)
		: CResizableDialog(CIrcWnd::IDD, pParent)
{
	m_pIrcMain = NULL;
	m_bConnected = false;
	m_bLoggedIn = false;
	m_pCurrentChannel = NULL;
	memset(&m_bSortAscendingNickList, true, sizeof(m_bSortAscendingNickList));
	memset(&m_bSortAscendingChanList, true, sizeof(m_bSortAscendingChanList));
}

CIrcWnd::~CIrcWnd()
{
	if (m_bConnected)
		m_pIrcMain->Disconnect(true);
	POSITION pos1, pos2;
	for (pos1 = channelLPtrList.GetHeadPosition(); (pos2 = pos1) != NULL; )
	{
		ChannelList* cur_channel = (ChannelList*)channelLPtrList.GetNext(pos1);
		channelLPtrList.RemoveAt(pos2);
		delete cur_channel;
	}
	DeleteAllChannel();
	delete m_pIrcMain;
}

void CIrcWnd::UpdateNickCount()
{
	CHeaderCtrl * pHeaderCtrl;
	HDITEM hdi;
	hdi.mask = HDI_TEXT;
	CString strRes;
	pHeaderCtrl = nickList.GetHeaderCtrl();
	if (nickList.GetItemCount())
		strRes.Format(_T("%s[%i]"), GetResString(IDS_IRC_NICK), nickList.GetItemCount());
	else
		strRes.Format(_T("%s"), GetResString(IDS_IRC_NICK));
	hdi.pszText = strRes.GetBuffer();
	pHeaderCtrl->SetItem(0, &hdi);
	strRes.ReleaseBuffer();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CIrcWnd::Localize()
{
	SetDlgItemText(IDC_BN_IRCCONNECT, GetResString((m_bConnected) ? IDS_IRC_DISCONNECT : IDS_IRC_CONNECT));
	SetDlgItemText(IDC_CHATSEND, GetResString(IDS_IRC_SEND));
	SetDlgItemText(IDC_CLOSECHAT, GetResString(IDS_FD_CLOSE));

	for (int i = 0; i < channelselect.GetItemCount();i++)
	{
		TCITEM item;
		item.mask = TCIF_PARAM;
		item.lParam = -1;
		channelselect.GetItem(i,&item);
		Channel* cur_chan = (Channel*)item.lParam;
		if (cur_chan != NULL)
		{
			if( cur_chan->type == 1 )
			{
				cur_chan->title = GetResString(IDS_STATUS);
				item.mask = TCIF_TEXT;
				item.pszText = cur_chan->title.GetBuffer();
				channelselect.SetItem(i,&item);
				cur_chan->title.ReleaseBuffer();
			}
			if( cur_chan->type == 2 )
			{
				cur_chan->title = GetResString(IDS_IRC_CHANNELLIST);
				item.mask = TCIF_TEXT;
				item.pszText = cur_chan->title.GetBuffer();
				channelselect.SetItem(i,&item);
				cur_chan->title.ReleaseBuffer();
			}
		}
	}
	if (m_pCurrentChannel)
	{
		if( m_pCurrentChannel->type == 1 )
		titleWindow.SetWindowText(GetResString(IDS_STATUS));
		if( m_pCurrentChannel->type == 2 )
		titleWindow.SetWindowText(GetResString(IDS_IRC_CHANNELLIST));
	}

	CHeaderCtrl	   *pHeaderCtrl;
	HDITEM			hdi;
	hdi.mask = HDI_TEXT;
	CString			strRes;

	pHeaderCtrl = nickList.GetHeaderCtrl();

	strRes = GetResString(IDS_STATUS);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(1, &hdi);

	UpdateNickCount();

	pHeaderCtrl = serverChannelList.GetHeaderCtrl();

	GetResString(&strRes, IDS_UUSERS);
	hdi.pszText = strRes.GetBuffer();
	pHeaderCtrl->SetItem(1, &hdi);
	strRes.ReleaseBuffer();

	GetResString(&strRes, IDS_DESCRIPTION);
	hdi.pszText = strRes.GetBuffer();
	pHeaderCtrl->SetItem(2, &hdi);
	strRes.ReleaseBuffer();

	GetResString(&strRes, IDS_IRC_NAME);
	hdi.pszText = strRes.GetBuffer();
	pHeaderCtrl->SetItem(0, &hdi);
	strRes.ReleaseBuffer();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CIrcWnd::OnInitDialog()
{
	static const uint16 s_auIconResID[] =
	{
		IDI_CHAT,
		IDI_MESSAGE,
		IDI_MPENDING
	};

	CResizableDialog::OnInitDialog();

	m_imageList.Create(16, 16, ILC_COLOR32, 0, 10);
	m_imageList.SetBkColor(RGB(255, 0, 255));
	FillImgLstWith16x16Icons(&m_imageList, s_auIconResID, ELEMENT_COUNT(s_auIconResID));

	CImageList imageList;
	CBitmap Bitmap, *pOldBitmap;
	CDC *pCtrlDC = GetDC(), TempDC;

	TempDC.CreateCompatibleDC(pCtrlDC);
	Bitmap.CreateCompatibleBitmap(pCtrlDC, 16 * 3, 16);
	pOldBitmap = TempDC.SelectObject(&Bitmap);
	m_imageList.Draw(&TempDC, 0, CPoint(0, 0), ILD_NORMAL);
	m_imageList.Draw(&TempDC, 1, CPoint(16, 0), ILD_NORMAL);
	m_imageList.Draw(&TempDC, 2, CPoint(32, 0), ILD_NORMAL);
	TempDC.SelectObject(pOldBitmap);
	imageList.Create(16, 16, ILC_COLOR32 | ILC_MASK, 0, 10);
	imageList.Add(&Bitmap, RGB(255, 0, 255));
	channelselect.SetImageList(&imageList);
	imageList.Detach();
	Bitmap.DeleteObject();
	ReleaseDC(pCtrlDC);

	m_bConnected = false;
	m_bLoggedIn = false;
	Localize();
	m_pIrcMain = new CIrcMain();
	m_pIrcMain->SetIRCWnd(this);

	nickList.InsertColumn(IRC1COL_NICK, GetResString(IDS_IRC_NICK), LVCFMT_LEFT, 90);
	nickList.InsertColumn(IRC1COL_STATUS, GetResString(IDS_STATUS), LVCFMT_LEFT, 70);

	serverChannelList.InsertColumn(IRC2COL_NAME, GetResString(IDS_IRC_NAME), LVCFMT_LEFT, 203);
	serverChannelList.InsertColumn(IRC2COL_USERS, GetResString(IDS_UUSERS), LVCFMT_RIGHT, 50);
	serverChannelList.InsertColumn(IRC2COL_DESCRIPTION, GetResString(IDS_DESCRIPTION), LVCFMT_LEFT, 350);
	serverChannelList.LoadSettings(CPreferences::TABLE_IRC);

	CRect rect;
	GetDlgItem(IDC_STATUSWINDOW)->GetWindowRect(rect);
	::MapWindowPoints(NULL, m_hWnd, (LPPOINT)&rect, 2);
	statusWindow.Create(WS_VISIBLE | WS_CHILD | WS_HSCROLL | WS_VSCROLL | ES_MULTILINE | ES_READONLY | ES_NOHIDESEL, rect, this, 125);
	statusWindow.ModifyStyleEx(0, WS_EX_STATICEDGE, SWP_FRAMECHANGED);
	statusWindow.SetEventMask(statusWindow.GetEventMask() | ENM_LINK);

	NewChannel(GetResString(IDS_STATUS), 1);
	NewChannel(GetResString(IDS_IRC_CHANNELLIST), 2);

	InitFont();
	m_pCurrentChannel = (Channel*)channelPtrList.GetTail();
	AddStatus(GetResString(IDS_IRC_STATUSLOG));
	titleWindow.SetWindowText(GetResString(IDS_IRC_CHANNELLIST));
	channelselect.SetCurSel(2);
	channelselect.SetCurFocus(2);
	OnTcnSelchangeTab2(NULL, NULL);
	channelselect.SetItemState(1, TCIS_HIGHLIGHTED, NULL);

	AddAnchor(IDC_BN_IRCCONNECT, BOTTOM_LEFT);
	AddAnchor(IDC_CLOSECHAT, BOTTOM_LEFT);
	AddAnchor(IDC_CHATSEND, BOTTOM_RIGHT);
	AddAnchor(IDC_INPUTWINDOW, BOTTOM_LEFT, BOTTOM_RIGHT);
	AddAnchor(IDC_NICKLIST, TOP_LEFT, BOTTOM_LEFT);
	AddAnchor(IDC_TITLEWINDOW, TOP_LEFT, TOP_RIGHT);
	AddAnchor(IDC_SERVERCHANNELLIST, TOP_LEFT, BOTTOM_RIGHT);
	AddAnchor(statusWindow, TOP_LEFT, BOTTOM_RIGHT);
	AddAnchor(IDC_TAB2, TOP_LEFT, TOP_RIGHT);

	statusWindow.ShowWindow(SW_SHOW);
	serverChannelList.ShowWindow(SW_HIDE);
	m_bSortAscendingChanList[IRC2COL_USERS] = false;

	nickList.SetSortArrow(IRC1COL_STATUS, m_bSortAscendingNickList[IRC1COL_STATUS]);
	nickList.SortItems(SortProcNick, IRC1COL_STATUS);

	if (g_eMuleApp.m_pGlobPrefs->DoUseSort())
		SortInit(g_eMuleApp.m_pGlobPrefs->GetIrcSortCol());
	else
	{
		int		iSortCode = g_eMuleApp.m_pGlobPrefs->GetColumnSortItem(CPreferences::TABLE_IRC);

		iSortCode |= g_eMuleApp.m_pGlobPrefs->GetColumnSortAscending(CPreferences::TABLE_IRC) ? MLC_SORTASC : MLC_SORTDESC;
		SortInit(iSortCode);
	}

	GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
	GetDlgItem(IDC_CHATSEND)->EnableWindow(false);

	return true;
}

void CIrcWnd::OnDestroy()
{
	CResizableDialog::OnDestroy();
	m_imageList.DeleteImageList();
}

void CIrcWnd::InitFont()
{
	statusWindow.SetFont(&g_eMuleApp.m_pdlgEmule->m_fontDefault);
	titleWindow.SetFont(&g_eMuleApp.m_pdlgEmule->m_fontDefault);
	inputWindow.SetFont(&g_eMuleApp.m_pdlgEmule->m_fontDefault);
}

void CIrcWnd::OnSize(UINT nType, int cx, int cy)
{
	CResizableDialog::OnSize(nType, cx, cy);
}

int CIrcWnd::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	return CResizableDialog::OnCreate(lpCreateStruct);
}

void CIrcWnd::DoDataExchange(CDataExchange* pDX)
{
	CResizableDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_NICKLIST, nickList);
	DDX_Control(pDX, IDC_INPUTWINDOW, inputWindow);
	DDX_Control(pDX, IDC_TITLEWINDOW, titleWindow);
	DDX_Control(pDX, IDC_SERVERCHANNELLIST, serverChannelList);
	DDX_Control(pDX, IDC_TAB2, channelselect);
}

BOOL CIrcWnd::OnCommand(WPARAM wParam, LPARAM lParam)
{
	int nickItem = nickList.GetNextItem(-1, LVIS_SELECTED | LVIS_FOCUSED);
	int chanItem = channelselect.GetCurSel();
//	int chanLItem= serverChannelList.GetNextItem(-1, LVIS_SELECTED | LVIS_FOCUSED); 
	switch (wParam)
	{
		case IDC_BN_IRCCONNECT:
		{
			OnBnClickedBnIrcconnect();
			return true;
		}
		case IDC_CHATSEND:
		{
			OnBnClickedChatsend();
			return true;
		}
		case IDC_CLOSECHAT:
		{
			OnBnClickedClosechat();
			return true;
		}
		case Irc_Priv:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			if (nick)
				AddMessage(nick->nick, _T(""), GetResString(IDS_IRC_PRIVATECHANSTART));
			return true;
		}
		case Irc_Owner:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("PRIVMSG chanserv owner %s %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_DeOwner:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("PRIVMSG chanserv deowner %s %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_Op:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("MODE %s +o %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_DeOp:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("MODE %s -o %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_HalfOp:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("MODE %s +h %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_DeHalfOp:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("MODE %s -h %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_Voice:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("MODE %s +v %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_DeVoice:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("MODE %s -v %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_Protect:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("PRIVMSG chanserv protect %s %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_DeProtect:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("PRIVMSG chanserv deprotect %s %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_Kick:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("KICK %s %s"), chan->name, nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_Slap:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(GetResString(IDS_IRC_SLAPMSGSEND), chan->name, nick->nick);
				AddMessage(chan->name, _T(""), GetResString(IDS_IRC_SLAPMSG), m_pIrcMain->GetNick(), nick->nick);
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_AddFriend:
		{
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("PRIVMSG %s :\001RQSFRIEND |%i|\001"), nick->nick, m_pIrcMain->SetVerify());
				m_pIrcMain->SendString(send);
			}
			return true;
		}
		case Irc_SendLink:
		{
		   	if(!GetSendFileString())
				   return true;
			Nick* nick = (Nick*)nickList.GetItemData(nickItem);
			TCITEM item;
			item.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &item);
			Channel* chan = (Channel*)item.lParam;
			if (nick && chan)
			{
				CString send;
				send.Format(_T("PRIVMSG %s :\001SENDLINK|%s|%s\001"), nick->nick, HashToString(g_eMuleApp.m_pGlobPrefs->GetUserHash()), GetSendFileString() );
				m_pIrcMain->SendString(send);
			}
			return true;
		}

		case Irc_Close:
		{
			OnBnClickedClosechat();
			return true;
		}
		case Irc_Join:
		{
			JoinChannels();
			return true;
		}
	}
	return true;
}

BOOL CIrcWnd::PreTranslateMessage(MSG* pMsg)
{
	if (pMsg->message == WM_KEYDOWN && (pMsg->hwnd == GetDlgItem(IDC_INPUTWINDOW)->m_hWnd))
	{
		if (pMsg->wParam == VK_RETURN)
		{
			OnBnClickedChatsend();
			return TRUE;
		}

		if (pMsg->wParam == VK_UP || pMsg->wParam == VK_DOWN)
		{
			ScrollHistory(pMsg->wParam == VK_DOWN);
			return TRUE;
		}
	}

	return CResizableDialog::PreTranslateMessage(pMsg);
}

void CIrcWnd::OnBnClickedBnIrcconnect()
{
	if (!m_bConnected)
		m_pIrcMain->Connect();
	else
		m_pIrcMain->Disconnect();
}

void CIrcWnd::OnBnClickedClosechat(int nItem)
{
	TCITEM item;

	item.mask = TCIF_PARAM;
	if (nItem == -1)
		nItem = channelselect.GetCurSel();

	if (nItem == -1)
		return;

	channelselect.GetItem(nItem, &item);
	Channel* partChannel = (Channel*)item.lParam;

	if (partChannel)
	{
		if (partChannel->type == 4 && m_bConnected)
		{
			CString part;
			part = _T("PART ") + partChannel->name;
			m_pIrcMain->SendString(part);
			return;
		}
		else if (partChannel->type == 5 || partChannel->type == 4)
		{
			RemoveChannel(partChannel->name);
			return;
		}
	}
}

void CIrcWnd::OnEnLinkIrcBox(NMHDR *pNMHDR, LRESULT *pResult)
{
	*pResult = 0;
	ENLINK* pEnLink = reinterpret_cast<ENLINK *>(pNMHDR);
	if (pEnLink && pEnLink->msg == WM_LBUTTONDOWN)
	{
		CString strUrl;
		statusWindow.GetTextRange(pEnLink->chrg.cpMin, pEnLink->chrg.cpMax, strUrl);
		ShellExecute(NULL, NULL, strUrl, NULL, NULL, SW_SHOWDEFAULT);
		*pResult = 1;
	}
}

void CIrcWnd::OnTcnSelchangeTab2(NMHDR *pNMHDR, LRESULT *pResult)
{
	nickList.DeleteAllItems();

	TCITEM item;
	item.mask = TCIF_PARAM;
	int cur_sel = channelselect.GetCurSel();
	if (cur_sel == -1)
		return;
	if (!channelselect.GetItem(cur_sel, &item))
		return;
	Channel* update = (Channel*)item.lParam;

	m_pCurrentChannel = update;
	UpdateNickCount();
	// bool bWasChanged = (channelselect.GetItemState(cur_sel, TCIS_HIGHLIGHTED) & TCIS_HIGHLIGHTED);
	channelselect.SetItemState(cur_sel, TCIS_HIGHLIGHTED, NULL);
	if (m_pCurrentChannel->type == 1)
	{
		titleWindow.SetWindowText(GetResString(IDS_STATUS));
		statusWindow.ShowWindow(SW_SHOW);
		serverChannelList.ShowWindow(SW_HIDE);
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
	}
	else if (m_pCurrentChannel->type == 2)
	{
		titleWindow.SetWindowText(GetResString(IDS_IRC_CHANNELLIST));
		statusWindow.ShowWindow(SW_HIDE);
		serverChannelList.ShowWindow(SW_SHOW);
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
		return;
	}
	else
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(true);
	SetActivity(m_pCurrentChannel->name, false);
	statusWindow.SetHyperText(m_pCurrentChannel->log);
	statusWindow.ShowWindow(SW_SHOW);
	serverChannelList.ShowWindow(SW_HIDE);
	RefreshNickList(update->name);
	SetTitle(update->name, update->title);
	statusWindow.SetTitle(update->name);
	if ((statusWindow.m_bAutoScroll) /*&& (bWasChanged)*/)
		statusWindow.ScrollToLastLine();
	if (pResult)
		* pResult = 0;
}

BEGIN_MESSAGE_MAP(CIrcWnd, CDialog)
	ON_NOTIFY(EN_LINK, 125, OnEnLinkIrcBox)
	ON_NOTIFY(NM_DBLCLK, IDC_SERVERCHANNELLIST, OnNMDblclkserverChannelList)
	ON_NOTIFY(NM_DBLCLK, IDC_NICKLIST, OnNMDblclkNickList)
	ON_NOTIFY(LVN_COLUMNCLICK, IDC_NICKLIST, OnColumnClickNick)
	ON_NOTIFY(LVN_COLUMNCLICK, IDC_SERVERCHANNELLIST, OnColumnClickChanL)
	ON_NOTIFY(NM_RCLICK, IDC_NICKLIST, OnNMRclickNick)
	ON_NOTIFY(NM_RCLICK, IDC_SERVERCHANNELLIST, OnNMRclickChanL)
	ON_NOTIFY(TCN_SELCHANGE, IDC_TAB2, OnTcnSelchangeTab2)
	ON_MESSAGE(WM_CLOSETAB, OnCloseTab)
	ON_WM_SIZE()
	ON_WM_CREATE()
	ON_WM_DESTROY()
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	Channel List
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

void CIrcWnd::ResetServerChannelList()
{
	POSITION pos1, pos2;
	for (pos1 = channelLPtrList.GetHeadPosition(); (pos2 = pos1) != NULL; )
	{
		ChannelList* cur_channel = (ChannelList*)channelLPtrList.GetNext(pos1);
		channelLPtrList.RemoveAt(pos2);
		delete cur_channel;
	}
	serverChannelList.DeleteAllItems();
}

void CIrcWnd::AddChannelToList(CString name, CString user, CString description)
{
	CString ntemp = name;
	CString dtemp = description;
	int usertest = _tstoi(user);
	if ((g_eMuleApp.m_pGlobPrefs->GetIRCChanNameFilter() || g_eMuleApp.m_pGlobPrefs->GetIRCChannelUserFilter()) && g_eMuleApp.m_pGlobPrefs->GetIRCUseChanFilter())
	{
		if (usertest < g_eMuleApp.m_pGlobPrefs->GetIRCChannelUserFilter())
			return;
		if (dtemp.MakeLower().Find(g_eMuleApp.m_pGlobPrefs->GetIRCChanNameFilter().MakeLower()) == -1 && ntemp.MakeLower().Find(g_eMuleApp.m_pGlobPrefs->GetIRCChanNameFilter().MakeLower()) == -1)
			return;
	}
	ChannelList* toadd = new ChannelList;
	toadd->name = name;
	toadd->users = user;
	toadd->desc = StripMessageOfFontCodes(description);
	channelLPtrList.AddTail(toadd);
	uint16 itemnr = serverChannelList.GetItemCount();

	itemnr = serverChannelList.InsertItem(LVIF_PARAM, itemnr, 0, 0, 0, 0, (LPARAM)toadd);
	serverChannelList.SetItemText(itemnr, 0, toadd->name);
	serverChannelList.SetItemText(itemnr, 1, toadd->users);
	serverChannelList.SetItemText(itemnr, 2, toadd->desc);
	if(channelselect.GetCurSel() != 1)
		channelselect.SetItemState(1, TCIS_HIGHLIGHTED, TCIS_HIGHLIGHTED);
}

int CIrcWnd::SortProcChanL(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	ChannelList *item1 = (ChannelList*)lParam1;
	ChannelList *item2 = (ChannelList*)lParam2;

	switch (lParamSort)
	{
		case IRC2COL_NAME:
			return CString(item1->name).CompareNoCase(item2->name);
		case IRC2COL_NAME + MLC_SORTDESC:
			return CString(item2->name).CompareNoCase(item1->name);
		case IRC2COL_USERS:
			return _tstoi(item1->users) - _tstoi(item2->users);
		case IRC2COL_USERS + MLC_SORTDESC:
			return _tstoi(item2->users) - _tstoi(item1->users);
		case IRC2COL_DESCRIPTION:
			return CString(item1->desc).CompareNoCase(item2->desc);
		case IRC2COL_DESCRIPTION + MLC_SORTDESC:
			return CString(item2->desc).CompareNoCase(item1->desc);
		default:
			return 0;
	}
}

void CIrcWnd::OnColumnClickChanL(NMHDR* pNMHDR, LRESULT* pResult)
{
	NM_LISTVIEW	*pNMListView = reinterpret_cast<NM_LISTVIEW*>(pNMHDR);
	int			iSubItem = pNMListView->iSubItem;
	bool		bSortOrder = m_bSortAscendingChanList[iSubItem];

// Reverse sorting direction for the same column and keep the same if column was changed
	if ((serverChannelList.GetSortParam() & MLC_COLUMNMASK) == iSubItem)
		m_bSortAscendingChanList[iSubItem] = bSortOrder = !bSortOrder;

	serverChannelList.SetSortArrow(iSubItem, bSortOrder);
	serverChannelList.SortItems(SortProcChanL, iSubItem + ((bSortOrder) ? MLC_SORTASC : MLC_SORTDESC));
	g_eMuleApp.m_pGlobPrefs->SetColumnSortItem(CPreferences::TABLE_IRC, iSubItem);
	g_eMuleApp.m_pGlobPrefs->SetColumnSortAscending(CPreferences::TABLE_IRC, bSortOrder);
	*pResult = 0;
}

void CIrcWnd::SortInit(int iSortCode)
{
//	Get the sort column
	int	iSortColumn = (iSortCode & MLC_COLUMNMASK);
//	Get the sort order
	bool	bSortAscending = (iSortCode & MLC_SORTDESC) == 0;

	serverChannelList.SetSortArrow(iSortColumn, bSortAscending);
	serverChannelList.SortItems(SortProcChanL, iSortCode);
	m_bSortAscendingChanList[iSortColumn] = bSortAscending;
}

void CIrcWnd::OnNMRclickChanL(NMHDR *pNMHDR, LRESULT *pResult)
{
	CTitleMenu	menuChanL;
	POINT		point;

	::GetCursorPos(&point);
	
	menuChanL.CreatePopupMenu();
	menuChanL.AddMenuTitle(GetResString(IDS_IRC_CHANNEL));
	menuChanL.AppendMenu(MF_STRING, Irc_Join, GetResString(IDS_IRC_JOIN));
	menuChanL.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);

//	Menu objects are destroyed in their destructor
	*pResult = 0;
}

void CIrcWnd::OnNMDblclkserverChannelList(NMHDR *pNMHDR, LRESULT *pResult)
{
	JoinChannels();
	*pResult = 0;
}

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	Nick List
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

Nick* CIrcWnd::FindNickByName(CString channel, CString name)
{
	Channel * curr_channel = FindChannelByName(channel);
	if (!curr_channel)
		return 0;

	for (POSITION pos = curr_channel->nicks.GetHeadPosition(); pos != NULL; )
	{
		Nick* cur_nick = (Nick*)curr_channel->nicks.GetNext(pos);
		if (cur_nick->nick == name)
			return cur_nick;
	}
	return 0;
}

Nick* CIrcWnd::NewNick(CString channel, CString nick)
{
	Channel * toaddchan = FindChannelByName(channel);
	if (!toaddchan)
		return NULL;
	if (FindNickByName(channel, nick))
		return NULL;
	Nick* toaddnick = NULL;
	if (toaddchan)
	{
		toaddnick = new Nick;

		if (nick.Left(1) == _T("!"))
		{
			nick = nick.Mid(1);
			toaddnick->owner = _T("!");
		}
		else
			toaddnick->owner.Empty();

		if (nick.Left(1) == _T("*"))
		{
			nick = nick.Mid(1);
			toaddnick->protect = _T("*");
		}
		else
			toaddnick->protect.Empty();

		if (nick.Left(1) == _T("@"))
		{
			toaddnick->op = _T("@");
			toaddnick->hop.Empty();
			toaddnick->voice.Empty();
			toaddnick->uop.Empty();
			toaddnick->nick = nick.Mid(1);
		}
		else if (nick.Left(1) == _T("%"))
		{
			toaddnick->op.Empty();
			toaddnick->hop = _T("%");
			toaddnick->voice.Empty();
			toaddnick->uop.Empty();
			toaddnick->nick = nick.Mid(1);
		}
		else if (nick.Left(1) == _T("+"))
		{
			toaddnick->op.Empty();
			toaddnick->hop.Empty();
			toaddnick->voice = _T("+");
			toaddnick->uop.Empty();
			toaddnick->nick = nick.Mid(1);
		}
		else if (nick.Left(1) == _T("-"))
		{
			toaddnick->op.Empty();
			toaddnick->hop.Empty();
			toaddnick->voice.Empty();
			toaddnick->uop = _T("-");
			toaddnick->nick = nick.Mid(1);
		}
		else
		{
			toaddnick->op.Empty();
			toaddnick->hop.Empty();
			toaddnick->voice.Empty();
			toaddnick->nick = nick;
		}
		toaddchan->nicks.AddTail(toaddnick);
		if (toaddchan == m_pCurrentChannel)
		{
			uint16 itemnr = nickList.GetItemCount();
			itemnr = nickList.InsertItem(LVIF_PARAM, itemnr, 0, 0, 0, 0, (LPARAM)toaddnick);
			nickList.SetItemText(itemnr, 0, (LPCTSTR)toaddnick->nick);
			CString mode;
			mode.Format(_T("%s%s%s%s%s%s"), toaddnick->owner, toaddnick->protect, toaddnick->op, toaddnick->hop, toaddnick->voice, toaddnick->uop);
			nickList.SetItemText(itemnr, 1, (LPCTSTR)mode);
			UpdateNickCount();
		}
	}
	return toaddnick;
}

void CIrcWnd::RefreshNickList(CString channel)
{
	nickList.DeleteAllItems();
	Channel* refresh = FindChannelByName(channel);
	if (!refresh)
		return;
	for (POSITION pos = refresh->nicks.GetHeadPosition(); pos != NULL; )
	{
		Nick* curr_nick = (Nick*)refresh->nicks.GetNext(pos);
		uint16 itemnr = nickList.GetItemCount();
		itemnr = nickList.InsertItem(LVIF_PARAM, itemnr, 0, 0, 0, 0, (LPARAM)curr_nick);
		nickList.SetItemText(itemnr, 0, (LPCTSTR)curr_nick->nick);
		CString mode;
		mode.Format(_T("%s%s%s%s%s%s"), curr_nick->owner, curr_nick->protect, curr_nick->op, curr_nick->hop, curr_nick->voice, curr_nick->uop);
		nickList.SetItemText(itemnr, 1, (LPCTSTR)mode);
	}
	UpdateNickCount();
}

bool CIrcWnd::RemoveNick(CString channel, CString nick)
{
	Channel * update = FindChannelByName(channel);
	if (!update)
		return false;
	POSITION pos1, pos2;
	for (pos1 = update->nicks.GetHeadPosition(); (pos2 = pos1) != NULL; )
	{
		Nick* curr_nick = (Nick*)update->nicks.GetNext(pos1);
		if (curr_nick->nick == nick)
		{
			if (update == m_pCurrentChannel)
			{
				LVFINDINFO find;
				find.flags = LVFI_PARAM;
				find.lParam = (LPARAM)curr_nick;
				sint32 result = nickList.FindItem(&find);
				nickList.DeleteItem(result);
				UpdateNickCount();
			}
			update->nicks.RemoveAt(pos2);
			delete curr_nick;
			return true;
		}
	}
	return false;
}

void CIrcWnd::DeleteAllNick(CString channel)
{
	Channel * curr_channel = FindChannelByName(channel);
	if (!curr_channel)
		return;
	POSITION pos1, pos2;
	for (pos1 = curr_channel->nicks.GetHeadPosition(); (pos2 = pos1) != NULL; )
	{
		Nick* cur_nick = (Nick*)curr_channel->nicks.GetNext(pos1);
		curr_channel->nicks.RemoveAt(pos2);
		delete cur_nick;
	}
}

void CIrcWnd::DeleteNickInAll(CString nick, CString message)
{
	for (POSITION pos = channelPtrList.GetHeadPosition(); pos != NULL; )
	{
		Channel* cur_channel = (Channel*)channelPtrList.GetNext(pos);
		if (RemoveNick(cur_channel->name, nick))
		{
			if (!g_eMuleApp.m_pGlobPrefs->GetIrcIgnoreInfoMessage())
			{
				CString strMessage = message;
				strMessage.Trim();
				CString strLine;
				strLine.Format(GetResString(IDS_IRC_HASQUIT), nick, strMessage);
				if (strMessage.GetLength() == 0)
					strLine.Truncate(strLine.GetLength() - 3);
				AddMessage(cur_channel->name, _T(""), strLine);
			}
		}
	}
}

bool CIrcWnd::ChangeNick(CString channel, CString oldnick, CString newnick)
{
	Channel * update = FindChannelByName(channel);
	if (!update)
		return false;
	for (POSITION pos = update->nicks.GetHeadPosition(); pos != NULL;)
	{
		Nick* curr_nick = (Nick*)update->nicks.GetNext(pos);
		if (curr_nick->nick == oldnick)
		{
			if ((update = m_pCurrentChannel) != NULL)
			{
				LVFINDINFO find;
				find.flags = LVFI_PARAM;
				find.lParam = (LPARAM)curr_nick;
				sint32 itemnr = nickList.FindItem(&find);
				if (itemnr != ( -1))
					nickList.SetItemText(itemnr, 0, (LPCTSTR)newnick);
			}
			curr_nick->nick = newnick;
			return true;
		}
	}
	return false;
}

bool CIrcWnd::ChangeMode(CString channel, CString nick, CString mode)
{
	Channel * update = FindChannelByName(channel);
	if (!update)
		return false;
	for (POSITION pos = update->nicks.GetHeadPosition(); pos != NULL; )
	{
		Nick* curr_nick = (Nick*)update->nicks.GetNext(pos);
		if (curr_nick->nick == nick)
		{
			sint32 itemnr = -1;
			if (( update = m_pCurrentChannel) != NULL )
			{
				LVFINDINFO find;
				find.flags = LVFI_PARAM;
				find.lParam = (LPARAM)curr_nick;
				itemnr = nickList.FindItem(&find);
			}
			if (mode == _T("+a"))
				curr_nick->protect = _T("*");
			if (mode == _T("-a"))
				curr_nick->protect.Empty();
			if (mode == _T("+h"))
				curr_nick->hop = _T("%");
			if (mode == _T("-h"))
				curr_nick->hop.Empty();
			if (mode == _T("+o"))
				curr_nick->op = _T("@");
			if (mode == _T("-o"))
				curr_nick->op.Empty();
			if (mode == _T("+q"))
				curr_nick->owner = _T("!");
			if (mode == _T("-q"))
				curr_nick->owner.Empty();
			if (mode == _T("+u"))
				curr_nick->uop = _T("-");
			if (mode == _T("-u"))
				curr_nick->uop.Empty();
			if (mode == _T("+v"))
				curr_nick->voice = _T("+");
			if (mode == _T("-v"))
				curr_nick->voice.Empty();
			if (itemnr != ( -1))
			{
				CString mode;
				mode.Format(_T("%s%s%s%s%s%s"), curr_nick->owner, curr_nick->protect, curr_nick->op, curr_nick->hop, curr_nick->voice, curr_nick->uop);
				nickList.SetItemText(itemnr, 1, (LPCTSTR)mode);
			}
		}
	}
	return true;
}

void CIrcWnd::ParseChangeMode(CString channel, CString changer, CString commands, CString names)
{
	try
	{
		if (commands.GetLength() == 2)
		{
			if (ChangeMode(channel, names, commands))
				if (!g_eMuleApp.m_pGlobPrefs->GetIrcIgnoreInfoMessage())
					AddMessage(channel, _T(""), GetResString(IDS_IRC_SETSMODE), changer, commands, names);
			return;
		}
		else
		{
			CString dir;
			dir = commands[0];
			if (dir == _T("+") || dir == _T("-"))
			{
				int currMode = 1;
				int currName = 0;
				int currNameBack = names.Find(_T(' '), currName);
				while (currMode < commands.GetLength())
				{
					CString test;
					if( currNameBack > currName )
					{
						test = names.Mid(currName, currNameBack - currName);
						currName = currNameBack + 1;
					}
					if (ChangeMode(channel, test, dir + commands[currMode]))
						if (!g_eMuleApp.m_pGlobPrefs->GetIrcIgnoreInfoMessage())
							AddMessage(channel, _T(""), GetResString(IDS_IRC_SETSMODE), changer, dir + commands[currMode] , test);
					currNameBack = names.Find(_T(' '), currName + 1);
					if (currNameBack == -1)
						currNameBack = names.GetLength();
					currMode++;
				}
			}
		}
	}
	catch (...)
	{
		AddMessage(channel, _T(""), GetResString(IDS_IRC_NOTSUPPORTED));
	}
}

void CIrcWnd::ChangeAllNick(CString oldnick, CString newnick)
{
	Channel * currchannel = FindChannelByName(oldnick);
	if (currchannel)
	{
		currchannel->name = newnick;
		TCITEM item;
		item.mask = TCIF_PARAM;
		item.lParam = -1;
		int i;
		for (i = 0; i < channelselect.GetItemCount();i++)
		{
			channelselect.GetItem(i, &item);
			if (((Channel*)item.lParam) == currchannel)
				break;
		}
		if (((Channel*)item.lParam) != currchannel)
			return;
		item.mask = TCIF_TEXT;
		item.pszText = newnick.GetBuffer();
		item.cchTextMax = (int)newnick.GetLength() + 1;
		channelselect.SetItem(i, &item);
	}
	for (POSITION pos = channelPtrList.GetHeadPosition(); pos != NULL;)
	{
		Channel* cur_channel = (Channel*)channelPtrList.GetNext(pos);
		if (ChangeNick(cur_channel->name, oldnick, newnick))
		{
			if (!g_eMuleApp.m_pGlobPrefs->GetIrcIgnoreInfoMessage())
				AddMessage(cur_channel->name, _T(""), GetResString(IDS_IRC_NOWKNOWNAS), oldnick, newnick);
		}
	}
}

/*
void CIrcWnd::SetNick( CString in_nick )
{
	g_eMuleApp.m_pGlobPrefs->SetIRCNick( in_nick.GetBuffer() );
	//Need to also update the preference window for this to work right..
}
*/

int CIrcWnd::SortProcNick(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	Nick* item1 = (Nick*)lParam1;
	Nick* item2 = (Nick*)lParam2;

	switch (lParamSort)
	{
	case IRC1COL_NICK:
		return item1->nick.CompareNoCase(item2->nick);
	case IRC1COL_NICK + MLC_SORTDESC:
		return item2->nick.CompareNoCase(item1->nick);
	case IRC1COL_STATUS:
		if( item1->owner == _T("!") )
		{
			if( item2->owner != _T("!") )
				return -1;
		}
		else if( item2->owner == _T("!") )
		{
			return 1;
		}
		if( item1->protect == _T("*") )
		{
			if( item2->protect != _T("*") )
				return -1;
		}
		else if( item2->protect == _T("*") )
		{
			return 1;
		}
		if( item1->op == _T("@") )
		{
			if( item2->op != _T("@") )
				return -1;
		}
		else if( item2->op == _T("@") )
		{
			return 1;
		}
		if( item1->hop == _T("%") )
		{
			if( item2->hop != _T("%") )
				return -1;
		}
		else if( item2->hop == _T("%") )
		{
			return 1;
		}
		if( item1->voice == _T("+") )
		{
			if( item2->voice != _T("+") )
				return -1;
		}
		else if( item2->voice == _T("+") )
		{
			return 1;
		}
		if( item1->uop == _T("-") )
		{
			if( item2->uop != _T("-") )
				return -1;
		}
		else if( item2->uop == _T("-") )
		{
			return 1;
		}

	case IRC1COL_STATUS + MLC_SORTDESC:
		if( item2->owner == _T("!") )
		{
			if( item1->owner != _T("!") )
				return -1;
		}
		else if( item1->owner == _T("!") )
		{
			return 1;
		}
		if( item2->protect == _T("*") )
		{
			if( item1->protect != _T("*") )
				return -1;
		}
		else if( item1->protect == _T("*") )
		{
			return 1;
		}
		if( item2->op == _T("@") )
		{
			if( item1->op != _T("@") )
				return -1;
		}
		else if( item1->op == _T("@") )
		{
			return 1;
		}
		if( item2->hop == _T("%") )
		{
			if( item1->hop != _T("%") )
				return -1;
		}
		else if( item1->hop == _T("%") )
		{
			return 1;
		}
		if( item2->voice == _T("+") )
		{
			if( item1->voice != _T("+") )
				return -1;
		}
		else if( item1->voice == _T("+") )
		{
			return 1;
		}
		if( item2->uop == _T("-") )
		{
			if( item1->uop != _T("-") )
				return -1;
		}
		else if( item1->uop == _T("-") )
		{
			return 1;
		}
	default:
		return 0;
	}
}

void CIrcWnd::OnColumnClickNick(NMHDR* pNMHDR, LRESULT* pResult)
{
	NM_LISTVIEW	*pNMListView = reinterpret_cast<NM_LISTVIEW*>(pNMHDR);
	int			iSubItem = pNMListView->iSubItem;
	bool		bSortOrder = m_bSortAscendingNickList[iSubItem];

// Reverse sorting direction for the same column and keep the same if column was changed
	if ((nickList.GetSortParam() & MLC_COLUMNMASK) == iSubItem)
		m_bSortAscendingNickList[iSubItem] = bSortOrder = !bSortOrder;

	nickList.SetSortArrow(iSubItem, bSortOrder);
	nickList.SortItems(SortProcNick, iSubItem + ((bSortOrder) ? MLC_SORTASC : MLC_SORTDESC));
	*pResult = 0;
}

void CIrcWnd::OnNMRclickNick(NMHDR *pNMHDR, LRESULT *pResult)
{
	if (nickList.GetSelectionMark() != ( -1))
	{
		CTitleMenu	menuNick;
		POINT		point;

		::GetCursorPos(&point);

		menuNick.CreatePopupMenu();
		menuNick.AddMenuTitle(GetResString(IDS_IRC_NICK));
		menuNick.AppendMenu(MF_STRING, Irc_Priv, GetResString(IDS_IRC_PRIVMESSAGE));
	//	menuNick.AppendMenu(MF_STRING,Irc_AddFriend,GetResString(IDS_IRC_ADDTOFRIENDLIST));
	//	if( !GetSendFileString().IsEmpty() )
	//	menuNick.AppendMenu(MF_STRING,Irc_SendLink, (CString)GetResString(IDS_IRC_SENDLINK) + GetSendFileString() );
		menuNick.AppendMenu(MF_STRING, Irc_Slap, GetResString(IDS_IRC_SLAP));
		menuNick.AppendMenu(MF_STRING, Irc_Owner, _T("Owner"));
		menuNick.AppendMenu(MF_STRING, Irc_DeOwner, _T("DeOwner"));
		menuNick.AppendMenu(MF_STRING, Irc_Op, GetResString(IDS_IRC_OP));
		menuNick.AppendMenu(MF_STRING, Irc_DeOp, GetResString(IDS_IRC_DEOP));
		menuNick.AppendMenu(MF_STRING, Irc_HalfOp, GetResString(IDS_IRC_HALFOP));
		menuNick.AppendMenu(MF_STRING, Irc_DeHalfOp, GetResString(IDS_IRC_DEHALFOP));
		menuNick.AppendMenu(MF_STRING, Irc_Voice, GetResString(IDS_IRC_VOICE));
		menuNick.AppendMenu(MF_STRING, Irc_DeVoice, GetResString(IDS_IRC_DEVOICE));
		menuNick.AppendMenu(MF_STRING, Irc_Protect, _T("Protect"));
		menuNick.AppendMenu(MF_STRING, Irc_DeProtect, _T("DeProtect"));
		menuNick.AppendMenu(MF_STRING, Irc_Kick, GetResString(IDS_IRC_KICK));
		menuNick.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);

//	Menu objects are destroyed in their destructor
	}
	*pResult = 0;
}

void CIrcWnd::OnNMDblclkNickList(NMHDR *pNMHDR, LRESULT *pResult)
{
	int nickItem = nickList.GetSelectionMark();
	if (nickItem != -1)
	{
		Nick * nick = (Nick*)nickList.GetItemData(nickItem);
		if (nick)
			AddMessage(nick->nick, _T(""), GetResString(IDS_IRC_PRIVATECHANSTART));
	}
	*pResult = 0;
}

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	Messages
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

void CIrcWnd::AddStatus(CString line, ...)
{
	va_list		argptr;
	CString		strTemp;
	Channel		*pUpdateChannel = (Channel*)channelPtrList.GetHead();

	va_start(argptr, line);
	strTemp.FormatV(line, argptr);
	va_end(argptr);

	if (pUpdateChannel == NULL)
		return;

	line = StripMessageOfFontCodes(strTemp);
	line += _T('\n');
	line.Replace(_T('\004'), _T('%'));

	COLORREF crLineColor = 0x1000000;

	strTemp = m_pIrcMain->GetNick();

	int iLen = strTemp.GetLength();

	if (iLen > 0 && line.Left(iLen) == strTemp)
	{
		if (line.Mid(iLen + 1, 1) == _T(":"))
			iLen ++;
		line = line.Right(line.GetLength() - (iLen + 1));
		crLineColor = RGB(72, 61, 139);
	}
	else if (line.Mid(0, 1) == _T("*"))
		crLineColor = RGB(153, 51, 0);
	else if (line.Mid(0, 1) == _T("-"))
		crLineColor = RGB(153, 51, 102);

	if (g_eMuleApp.m_pGlobPrefs->GetIRCAddTimestamp())
	{
		strTemp = CTime::GetCurrentTime().Format(_T("%X: "));
		pUpdateChannel->log->AppendText(strTemp, RGB(0,0,128));
	}

	pUpdateChannel->log->AppendText(line, crLineColor);

	if (m_pCurrentChannel == pUpdateChannel)
	{
		statusWindow.SetHyperText(pUpdateChannel->log);
		return;
	}
	SetActivity(pUpdateChannel->name, true);
}

void CIrcWnd::AddMessage(CString channelName, CString targetname, CString line, ...)
{
	if (channelName.IsEmpty())
		return;

	va_list		argptr;
	CString		strTemp;
	Channel		*pUpdateChannel = FindChannelByName(channelName);

	va_start(argptr, line);
	strTemp.FormatV(line, argptr);
	line = strTemp;
	va_end(argptr);

	if (pUpdateChannel == NULL)
		pUpdateChannel = NewChannel(channelName, (channelName.Left(1) == _T("#")) ? 4 : 5);

	line.Remove(_T('\017')); // Fixed pitch
	line.Remove(_T('\022')); // Reverse
	line.Remove(_T('\026'));
	line.Remove(_T('\b'));
	line.Remove(_T('\t'));
	line.Replace(_T('\004'), _T('%'));
	line.Replace(_T("\\"),_T("\\\\"));

	if (line.IsEmpty())
		return;

	COLORREF	crDefaultColor = GetSysColor(COLOR_WINDOWTEXT);

	if (targetname.IsEmpty())
	{
		if (line.Mid(0, 1) == _T("*"))
			crDefaultColor = RGB(153, 51, 0);
		else if (line.Mid(0, 1) == _T("-"))
			crDefaultColor = RGB(153, 51, 102);
	}

	COLORREF	crColor = crDefaultColor;
	int			iIndex = 0, iColorIndex;
	bool		bBold = false, bItalic = false, bUnderline = false;

	if (g_eMuleApp.m_pGlobPrefs->GetIRCAddTimestamp())
	{
		strTemp = CTime::GetCurrentTime().Format(_T("%X: "));
		pUpdateChannel->log->AppendText(strTemp, RGB(0,0,128));
	}

	if (!targetname.IsEmpty())
		pUpdateChannel->log->AppendText(_T("<") + targetname + _T("> "), (m_pIrcMain->GetNick() == targetname) ? RGB(128, 0, 128) : RGB(51, 51, 153));

	while (!line.IsEmpty())
	{
		switch (line.GetAt(iIndex))
		{
			case 2:
			{
				line.Delete(iIndex, 1);
				if (bBold)
				{
					line.Insert(iIndex, _T("\\b0 "));
					iIndex += 4;
				}
				else
				{
					line.Insert(iIndex, _T("\\b "));
					iIndex += 3;
				}
				bBold = !bBold;
				break;
			}
			case 3:
			{
				line.Delete(iIndex, 1);
				if (iIndex > 0)
				{
					pUpdateChannel->log->AppendText(line.Left(iIndex), crColor | 0x2000000);
					line = line.Mid(iIndex);
					iIndex = 0;
				}

				if (!line.IsEmpty())
				{
					if (line.GetAt(0) >= 48 && line.GetAt(0) <= 58)
					{
						if (line.GetLength() > 1)
						{
							iIndex = 1;
							if (line.GetAt(1) >= 48 && line.GetAt(1) <= 58)
								iIndex++;
							iColorIndex = _tstoi(line.Left(iIndex));
							if (iColorIndex < 16)
							{
								crColor = _acrColorPal[iColorIndex];
								if (crColor == GetSysColor(COLOR_WINDOW))
									crColor = crDefaultColor;
							}
							if (line.GetLength() > iIndex && line.GetAt(iIndex) == 44 && line.GetAt(iIndex + 1) >= 48 && line.GetAt(iIndex + 1) <= 58)
							{
								iIndex += 2;
								if (line.GetLength() > iIndex && line.GetAt(iIndex) >= 48 && line.GetAt(iIndex) <= 58)
									iIndex++;
							}
							line = line.Mid(iIndex);
							iIndex = 0;
						}
						else
							line.Empty();
					}
					else
						crColor = crDefaultColor;
				}
				break;
			}
			case 15:
			{
				line.Delete(iIndex, 1);
				crColor = crDefaultColor;
				if (bBold)
				{
					line.Insert(iIndex, _T("\\b0 "));
					iIndex += 4;
					bBold = false;
				}
				if (bItalic)
				{
					line.Insert(iIndex, _T("\\i0 "));
					iIndex += 4;
					bItalic = false;
				}
				if (bUnderline)
				{
					line.Insert(iIndex, _T("\\ul0 "));
					iIndex += 5;
					bUnderline = false;
				}
				break;
			}
			case 29:
			{
				line.Delete(iIndex, 1);
				if (bItalic)
				{
					line.Insert(iIndex, _T("\\i0 "));
					iIndex += 4;
				}
				else
				{
					line.Insert(iIndex, _T("\\i "));
					iIndex += 3;
				}
				bItalic = !bItalic;
				break;
			}
			case 31:
			{
				line.Delete(iIndex, 1);
				if (bUnderline)
				{
					line.Insert(iIndex, _T("\\ul0 "));
					iIndex += 5;
				}
				else
				{
					line.Insert(iIndex, _T("\\ul "));
					iIndex += 4;
				}
				bUnderline = !bUnderline;
				break;
			}
			default:
			{
				iIndex++;
				break;
			}
		}
		if (iIndex >= line.GetLength())
		{
			if (bBold)
				line += _T("\\b0 ");

			if (bItalic)
				line += _T("\\i0 ");

			if (bUnderline)
				line += _T("\\ul0 ");

			line += _T('\n');

			pUpdateChannel->log->AppendText(line, crColor | 0x2000000);
			line.Empty();
		}
	}

	if (m_pCurrentChannel == pUpdateChannel)
	{
		statusWindow.SetHyperText(pUpdateChannel->log);
		return;
	}
	SetActivity(pUpdateChannel->name, true);
}

void CIrcWnd::SetConnectStatus(bool flag)
{
	if (flag)
	{
		SetDlgItemText(IDC_BN_IRCCONNECT, GetResString(IDS_IRC_DISCONNECT));
		AddStatus(GetResString(IDS_CONNECTED));
		m_bConnected = true;
	}
	else
	{
		SetDlgItemText(IDC_BN_IRCCONNECT, GetResString(IDS_IRC_CONNECT));
		AddStatus(GetResString(IDS_DISCONNECTED));
		m_bConnected = false;
		m_bLoggedIn = false;
		while (channelPtrList.GetCount() > 2)
		{
			Channel * todel = (Channel*)channelPtrList.GetTail();
			RemoveChannel(todel->name);
		}
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
		GetDlgItem(IDC_CHATSEND)->EnableWindow(false);
	}
}

void CIrcWnd::NoticeMessage(CString source, CString message)
{
	bool flag = false;
	Channel* curr_channel = FindChannelByName(source);
	if (curr_channel)
	{
		AddMessage(source, _T(""), _T("-%s- %s"), source, message);
		flag = true;
	}
	for (POSITION pos = channelPtrList.GetHeadPosition(); pos != NULL; )
	{
		curr_channel = (Channel*)channelPtrList.GetNext(pos);
		Nick* curr_nick = FindNickByName(curr_channel->name, source);
		if (curr_nick)
		{
			AddMessage(curr_channel->name, _T(""), _T("-%s- %s"), source, message);
			flag = true;
		}
	}
	if (flag == false)
	{
		if (m_pCurrentChannel->type == 4)
			AddMessage(m_pCurrentChannel->name, _T(""), _T("-%s- %s"), source, message);
		else
			AddStatus(_T("-%s- %s"), source, message);
	}
}

//We cannot support color within the text since HyperTextCtrl does not detect hyperlinks with color. So I will filter it.
CString CIrcWnd::StripMessageOfFontCodes(CString temp)
{
	temp = StripMessageOfColorCodes(temp);
	temp.Remove(_T('\002'));
	temp.Remove(_T('\003'));
	temp.Remove(_T('\017'));
	temp.Remove(_T('\026'));
	temp.Remove(_T('\037'));
	temp.Remove(_T('\b'));
	temp.Remove(_T('\t'));
	return temp;
}

CString CIrcWnd::StripMessageOfColorCodes(CString temp)
{
	if (!temp.IsEmpty())
	{
		CString temp1, temp2;
		int test = temp.Find(3);
		if (test != -1)
		{
			int testlength = temp.GetLength() - test;
			if (testlength < 2)
				return temp;
			temp1 = temp.Left(test);
			temp2 = temp.Mid(test + 2);
			if (testlength < 4)
				return temp1 + temp2;
			if (temp2[0] == 44 && temp2.GetLength() > 2)
			{
				temp2 = temp2.Mid(2);
				for (int I = 48; I < 58; I++)
				{
					if (temp2[0] == I)
						temp2 = temp2.Mid(1);
				}
			}
			else
			{
				for (int I = 48; I < 58; I++)
				{
					if (temp2[0] == I)
					{
						temp2 = temp2.Mid(1);
						if (temp2[0] == 44 && temp2.GetLength() > 2)
						{
							temp2 = temp2.Mid(2);
							for (int I = 48; I < 58; I++)
							{
								if (temp2[0] == I)
									temp2 = temp2.Mid(1);
							}
						}
					}
				}
			}
			temp = temp1 + temp2;
			temp = StripMessageOfColorCodes(temp);
		}
	}
	return temp;
}

void CIrcWnd::SetTitle(CString channel, CString title)
{
	Channel * curr_channel = FindChannelByName(channel);
	if (!curr_channel)
		return;
	curr_channel->title = StripMessageOfFontCodes(title);
	if (curr_channel == m_pCurrentChannel)
		titleWindow.SetWindowText(curr_channel->title);
}

void CIrcWnd::SetActivity(CString channel, bool flag)
{
	Channel * refresh = FindChannelByName(channel);
	if (!refresh)
	{
		refresh = (Channel*)channelPtrList.GetHead();
		if (!refresh)
			return;
	}
	TCITEM item;
	item.mask = TCIF_PARAM;
	item.lParam = -1;
	int i;
	for (i = 0; i < channelselect.GetItemCount();i++)
	{
		channelselect.GetItem(i, &item);
		if (((Channel*)item.lParam) == refresh)
			break;
	}
	if (((Channel*)item.lParam) != refresh)
		return;
	/*if (flag)
	{
		item.mask = TCIF_IMAGE;
		item.iImage = 2;
		channelselect.SetItem(i, &item);
	}
	else
	{
		item.mask = TCIF_IMAGE;
		item.iImage = 1;
		channelselect.SetItem(i, &item);
	}*/
	if(channelselect.GetCurSel() != i)
		channelselect.SetItemState(i, TCIS_HIGHLIGHTED, TCIS_HIGHLIGHTED);
}

void CIrcWnd::OnBnClickedChatsend()
{
	CString send;
	GetDlgItemText(IDC_INPUTWINDOW, send);
	SetDlgItemText(IDC_INPUTWINDOW, _T(""));

	if (m_pCurrentChannel->history.GetCount() == g_eMuleApp.m_pGlobPrefs->GetMaxChatHistoryLines())
		m_pCurrentChannel->history.RemoveAt(0);
	m_pCurrentChannel->history.Add(send);
	m_pCurrentChannel->history_pos = m_pCurrentChannel->history.GetCount();

	if (send.IsEmpty())
		return;

	if (!this->m_bConnected)
		return;
	if( send.Left(4) == _T("/hop") )
	{
		if( m_pCurrentChannel->name.Left(1) == _T("#") )
		{
			CString channel = m_pCurrentChannel->name;
			m_pIrcMain->SendString(_T("PART ") + channel );
			m_pIrcMain->SendString(_T("JOIN ") + channel );
			return;
		}
		return;
	}
	if (send.Left(1) == _T("/") && send.Left(3).MakeLower() != _T("/me") && send.Left(6).MakeLower() != _T("/sound"))
	{
		if (send.Left(4) == _T("/msg"))
		{
			if (m_pCurrentChannel->type == 4 || m_pCurrentChannel->type == 5)
			{
				send.Replace(_T('%'), _T('\004'));
				AddMessage(m_pCurrentChannel->name, _T(""), _T("* >> ") + send.Mid(5));
				send.Replace(_T('\004'), _T('%'));
			}
			else
			{
				send.Replace(_T('%'), _T('\004'));
				AddStatus(_T("* >> ") + send.Mid(5));
				send.Replace(_T('\004'), _T('%'));
			}
			send = _T("/PRIVMSG") + send.Mid(4);
		}
		if( ((CString)send.Left(17)).CompareNoCase(_T("/PRIVMSG nickserv") )== 0)
		{
			send = _T("/ns") + send.Mid(17);
		}
		else if( ((CString)send.Left(17)).CompareNoCase(_T("/PRIVMSG chanserv") )== 0)
		{
			send = _T("/cs") + send.Mid(17);
		}
		else if( ((CString)send.Left(8)).CompareNoCase(_T("/PRIVMSG") )== 0)
		{
			int index = send.Find(_T(" "), send.Find(_T(" "))+1);
			send.Insert(index+1, _T(":"));
		}
		else if( ((CString)send.Left(6)).CompareNoCase(_T("/TOPIC") )== 0)
		{
			int index = send.Find(_T(" "), send.Find(_T(" "))+1);
			send.Insert(index+1, _T(":"));
		}
		m_pIrcMain->SendString(send.Mid(1));
		return;
	}
	if (m_pCurrentChannel->type < 4)
	{
		m_pIrcMain->SendString(send);
		return;
	}
	if (send.Left(3) == _T("/me"))
	{
		CString build;
		build.Format(_T("PRIVMSG %s :\001ACTION %s\001"), m_pCurrentChannel->name, send.Mid(4));
		send.Replace(_T('%'), _T('\004'));
		AddMessage(m_pCurrentChannel->name, _T(""), _T("* %s %s"), m_pIrcMain->GetNick(), send.Mid(4));
		m_pIrcMain->SendString(build);
		return;
	}
	CString build = _T("PRIVMSG ") + m_pCurrentChannel->name + _T(" :") + send;
	m_pIrcMain->SendString(build);
	send.Replace(_T('%'), _T('\004'));
	AddMessage(m_pCurrentChannel->name, m_pIrcMain->GetNick(), send);
	send.Replace(_T('\004'), _T('%'));
}

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	Channels
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

Channel* CIrcWnd::FindChannelByName(CString name)
{
	for (POSITION pos = channelPtrList.GetHeadPosition(); pos != NULL; )
	{
		Channel* cur_channel = (Channel*)channelPtrList.GetNext(pos);
		if (cur_channel->name.CompareNoCase(name.Trim()) == 0 && (cur_channel->type == 4 || cur_channel->type == 5))
			return cur_channel;
	}
	return 0;
}

Channel* CIrcWnd::NewChannel(CString name, byte type)
{
	Channel * toadd = new Channel;

	toadd->name = name;
	toadd->log = new CPreparedRTFText();
	toadd->title = name;
	toadd->type = type;
	toadd->history_pos = 0;
	channelPtrList.AddTail(toadd);
	TCITEM newitem;

	newitem.mask = TCIF_PARAM | TCIF_TEXT;
	if (type >= 4)
		newitem.mask |= TCIF_IMAGE;
	newitem.lParam = (LPARAM)toadd;
	newitem.pszText = name.GetBuffer();
	newitem.cchTextMax = (int)name.GetLength() + 1;
	newitem.iImage = 1;
	uint32 pos = channelselect.GetItemCount();

	channelselect.InsertItem(pos, &newitem);
	if (type >= 4)
	{
		channelselect.SetCurSel(pos);
		channelselect.SetCurFocus(pos);
		OnTcnSelchangeTab2( NULL, NULL );
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(true);
		GetDlgItem(IDC_CHATSEND)->EnableWindow(true);
	}
	return toadd;
}

void CIrcWnd::RemoveChannel(CString channel)
{
	Channel * todel = FindChannelByName(channel);

	if (!todel)
		return;

	TCITEM item;

	item.mask = TCIF_PARAM;
	item.lParam = -1;
	
	int i;

	for (i = 0; i < channelselect.GetItemCount();i++)
	{
		channelselect.GetItem(i, &item);
		if (((Channel*)item.lParam) == todel)
			break;
	}
	if (((Channel*)item.lParam) != todel)
		return;
	channelselect.DeleteItem(i);

	if (todel == m_pCurrentChannel)
	{
		nickList.DeleteAllItems();
		statusWindow.SetWindowText(_T(""));
		if (channelselect.GetItemCount() > 2 && i > 1)
		{
			if (i == 2)
				i++;
			channelselect.SetCurSel(i - 1);
			channelselect.SetCurFocus(i - 1);
			OnTcnSelchangeTab2(NULL, NULL);
		}
		else
		{
			channelselect.SetCurSel(0);
			channelselect.SetCurFocus(0);
			OnTcnSelchangeTab2(NULL, NULL);
			GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
			GetDlgItem(IDC_CHATSEND)->EnableWindow(false);
		}
	}
	DeleteAllNick(todel->name);
	channelPtrList.RemoveAt(channelPtrList.Find(todel));
	delete todel->log;
	delete todel;
}

void CIrcWnd::DeleteAllChannel()
{
	POSITION	pos1, pos2;
	Channel	   *pChannel;

	for (pos1 = channelPtrList.GetHeadPosition();(pos2 = pos1) != NULL;)
	{
		pChannel = (Channel*)channelPtrList.GetNext(pos1);
		DeleteAllNick(pChannel->name);
		channelPtrList.RemoveAt(pos2);
		delete pChannel->log;
		delete pChannel;
	}
}

void CIrcWnd::JoinChannels()
{
	if (!this->m_bConnected)
		return;

	int			index = -1;
	POSITION	pos = serverChannelList.GetFirstSelectedItemPosition();

	while (pos != NULL)
	{
		index = serverChannelList.GetNextSelectedItem(pos);
		if (index > -1)
		{
			CString join;
			join = _T("JOIN ") + serverChannelList.GetItemText(index, 0);
			m_pIrcMain->SendString(join);
		}
	}
}

LRESULT CIrcWnd::OnCloseTab(WPARAM wparam, LPARAM lparam)
{
	OnBnClickedClosechat((int)wparam);
	return true;
}

void CIrcWnd::SendString(CString send)
{
	if (this->m_bConnected)
		m_pIrcMain->SendString(send);
}

void CIrcWnd::ScrollHistory(bool down)
{
	CString buffer;

	if ( (m_pCurrentChannel->history_pos == 0 && !down) ||
	     (m_pCurrentChannel->history_pos == m_pCurrentChannel->history.GetCount() && down) )
		return;

	if (down)
		++m_pCurrentChannel->history_pos;
	else
		--m_pCurrentChannel->history_pos;

	buffer = (m_pCurrentChannel->history_pos == m_pCurrentChannel->history.GetCount()) ? _T("") : m_pCurrentChannel->history.GetAt(m_pCurrentChannel->history_pos);

	SetDlgItemText(IDC_INPUTWINDOW, buffer);
	inputWindow.SetSel(buffer.GetLength(), buffer.GetLength());
}
