//////////////////////////////////
//	MuleListCtrl.cpp
//	: implementation file
//

#include "stdafx.h"
#include "emule.h"
#include "memdc.h"
#include "MuleListCtrl.h"
#include "otherfunctions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif


#define MLC_BLEND(A, B, X) ((A + B * (X-1) + ((X+1)/2)) / X)

#define MLC_RGBBLEND(A, B, X) (                   \
	RGB(MLC_BLEND(GetRValue(A), GetRValue(B), X), \
	MLC_BLEND(GetGValue(A), GetGValue(B), X),     \
	MLC_BLEND(GetBValue(A), GetBValue(B), X))     \
)

#define MLC_DT_TEXT (DT_SINGLELINE | DT_NOPREFIX | DT_VCENTER | DT_END_ELLIPSIS)

#define MLC_IDC_MENU	4875
#define MLC_IDC_UPDATE	(MLC_IDC_MENU - 1)

//a value that's not a multiple of 4 and uncommon
#define MLC_MAGIC 0xFEEBDEEF

//used for very slow assertions
//#define MLC_ASSERT(f)	ASSERT(f)
#define MLC_ASSERT(f)	((void)0)

#ifndef HDM_SETBITMAPMARGIN
#define HDM_SETBITMAPMARGIN	(HDM_FIRST + 20)
#define HDM_GETBITMAPMARGIN	(HDM_FIRST + 21)
#endif

BEGIN_MESSAGE_MAP(CMuleListCtrl, CListCtrl)
	ON_WM_DRAWITEM()
	ON_WM_KEYDOWN()
	ON_WM_ERASEBKGND()
	ON_WM_SYSCOLORCHANGE()
END_MESSAGE_MAP()

//////////////////////////////////
//	CMuleListCtrl

IMPLEMENT_DYNAMIC(CMuleListCtrl, CListCtrl)
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CMuleListCtrl::CMuleListCtrl(PFNLVCOMPARE pfnCompare, DWORD dwParamSort)
{
	m_SortProc = pfnCompare;
	m_dwParamSort = dwParamSort;

	m_bCustomDraw = false;
	m_iCurrentSortItem[0] = -1;
	m_iCurrentSortItem[1] = -1;
	m_iCurrentSortItem[2] = -1;
	m_iColumnsTracked = 0;
	m_aColumns = NULL;
	m_lRedrawCount = 0;
	::InitializeCriticalSection(&m_csRedraw);

//	just in case
	m_crWindow = 0;
	m_crWindowText = 0;
	m_crWindowTextBk = 0;
	m_crHighlight = 0;
	m_crGlow = 0;
	m_crGlowText = 0;
	m_crDimmedText = 0;
	m_crFocusLine = 0;
	m_crNoHighlight = 0;
	m_crNoFocusLine = 0;
	m_bGeneralPurposeFind = false;
	m_bFindMatchCase = false;
	m_iFindDirection = 1;
	m_iFindColumn = 0;
	m_bGeneralPurposeFind = false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CMuleListCtrl::~CMuleListCtrl()
{
	if (m_aColumns != NULL)
		delete[] m_aColumns;

	::DeleteCriticalSection(&m_csRedraw);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CMuleListCtrl::SortProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::SetName(LPCTSTR lpszName)
{
	m_Name = lpszName;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::PreSubclassWindow()
{
	SetColors();
	CListCtrl::PreSubclassWindow();
	ModifyStyle(LVS_LIST | LVS_ICON | LVS_SMALLICON, LVS_REPORT | LVS_SINGLESEL);
	SetExtendedStyle(LVS_EX_HEADERDRAGDROP);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CMuleListCtrl::IndexToOrder(CHeaderCtrl* pHeader, int iIndex)
{
	int			iCount = pHeader->GetItemCount();
	int			*piArray = new int[iCount];

	Header_GetOrderArray(pHeader->m_hWnd, iCount, piArray);
	for (int i = 0; i < iCount; i++)
	{
		if (piArray[i] == iIndex)
		{
			delete[] piArray;
			return i;
		}
	}
	delete[] piArray;
	return -1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::HideColumn(int iColumn)
{
	CHeaderCtrl			*pHeaderCtrl = GetHeaderCtrl();
	int					iCount = pHeaderCtrl->GetItemCount();

	if (iColumn < 1 || iColumn >= iCount || m_aColumns[iColumn].bHidden)
		return;

//	stop it from redrawing
	SetRedraw(FALSE);

//	shrink width to 0
	HDITEM		item;

	item.mask = HDI_WIDTH;
	pHeaderCtrl->GetItem(iColumn, &item);
	m_aColumns[iColumn].iWidth = item.cxy;
	item.cxy = 0;
	pHeaderCtrl->SetItem(iColumn, &item);

//	move to front of list
	int			*piArray = new INT[m_iColumnsTracked];

	pHeaderCtrl->GetOrderArray(piArray, m_iColumnsTracked);

	int			iFrom = m_aColumns[iColumn].iLocation;

	for (int i = 0; i < m_iColumnsTracked; i++)
		if (m_aColumns[i].iLocation > m_aColumns[iColumn].iLocation && m_aColumns[i].bHidden)
			iFrom++;

	for (int i = iFrom; i > 0; i--)
		piArray[i] = piArray[i - 1];
	piArray[0] = iColumn;
	pHeaderCtrl->SetOrderArray(m_iColumnsTracked, piArray);
	delete[] piArray;

//	update entry
	m_aColumns[iColumn].bHidden = true;

//	redraw
	SetRedraw(TRUE);
	Invalidate(FALSE);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::ShowColumn(int iColumn)
{
	CHeaderCtrl		*pHeaderCtrl = GetHeaderCtrl();
	int				iCount = pHeaderCtrl->GetItemCount();

	if (iColumn < 1 || iColumn >= iCount || !m_aColumns[iColumn].bHidden)
		return;

//	stop it from redrawing
	SetRedraw(FALSE);

//	restore position in list
	int				*piArray = new INT[m_iColumnsTracked];

	pHeaderCtrl->GetOrderArray(piArray, m_iColumnsTracked);

	int				iCurrent = IndexToOrder(pHeaderCtrl, iColumn);

	for (; iCurrent < IndexToOrder(pHeaderCtrl, 0) && iCurrent < m_iColumnsTracked - 1; iCurrent++)
		piArray[iCurrent] = piArray[iCurrent + 1];
	for ( ; m_aColumns[iColumn].iLocation > m_aColumns[pHeaderCtrl->OrderToIndex(iCurrent + 1)].iLocation &&
	      iCurrent < m_iColumnsTracked - 1; iCurrent++ )
		piArray[iCurrent] = piArray[iCurrent + 1];
	piArray[iCurrent] = iColumn;
	pHeaderCtrl->SetOrderArray(m_iColumnsTracked, piArray);
	delete[] piArray;

//	and THEN restore original width
	HDITEM			item;

	item.mask = HDI_WIDTH;
	item.cxy = m_aColumns[iColumn].iWidth;
	pHeaderCtrl->SetItem(iColumn, &item);

//	update entry
	m_aColumns[iColumn].bHidden = false;

//	redraw
	SetRedraw(TRUE);
	Invalidate(FALSE);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::SaveSettings(CPreferences::EnumTable tID)
{
	int			*piArray = new INT[m_iColumnsTracked];

	for (int i = 0; i < m_iColumnsTracked; i++)
	{
		g_eMuleApp.m_pGlobPrefs->SetColumnWidth(tID, i, GetColumnWidth(i));
		g_eMuleApp.m_pGlobPrefs->SetColumnHidden(tID, i, IsColumnHidden(i));
		piArray[i] = m_aColumns[i].iLocation;
	}

	g_eMuleApp.m_pGlobPrefs->SetColumnOrder(tID, piArray);
	delete[] piArray;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::LoadSettings(CPreferences::EnumTable tID)
{
	CHeaderCtrl			*pHeaderCtrl = GetHeaderCtrl();
	int					*piArray = new INT[m_iColumnsTracked];

	for (int i = 0; i < m_iColumnsTracked; i++)
		piArray[i] = i;

	for (int i = 0; i < m_iColumnsTracked; i++)
	{
		int iWidth = g_eMuleApp.m_pGlobPrefs->GetColumnWidth(tID, i);

		if (iWidth != 0)
			SetColumnWidth(i, iWidth);
		if (i == 0)
		{
			piArray[0] = 0;
		}
		else
		{
			int iOrder = g_eMuleApp.m_pGlobPrefs->GetColumnOrder(tID, i);

			if (iOrder > 0 && iOrder < m_iColumnsTracked && iOrder != i)
				piArray[iOrder] = i;
		}
	}

	pHeaderCtrl->SetOrderArray(m_iColumnsTracked, piArray);
	pHeaderCtrl->GetOrderArray(piArray, m_iColumnsTracked);
	for (int i = 0; i < m_iColumnsTracked; i++)
		m_aColumns[piArray[i]].iLocation = i;

	delete[] piArray;

	for (int i = 1; i < m_iColumnsTracked; i++)
	{
		if (g_eMuleApp.m_pGlobPrefs->GetColumnHidden(tID, i))
			HideColumn(i);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::SetColors()
{
	m_crWindow = ::GetSysColor(COLOR_WINDOW);
	m_crWindowText = ::GetSysColor(COLOR_WINDOWTEXT);
	m_crWindowTextBk = m_crWindow;

	COLORREF		crHighlight = ::GetSysColor(COLOR_HIGHLIGHT);

	m_crFocusLine	= crHighlight;
	m_crNoHighlight = MLC_RGBBLEND(crHighlight, m_crWindow, 8);
	m_crNoFocusLine = MLC_RGBBLEND(crHighlight, m_crWindow, 2);
	m_crHighlight	= MLC_RGBBLEND(crHighlight, m_crWindow, 4);
	m_crGlow		= ::GetSysColor(COLOR_INFOBK);
	m_crGlowText	= ::GetSysColor(COLOR_INFOTEXT);
	m_crDimmedText	= ::GetSysColor(COLOR_GRAYTEXT);
}
void CMuleListCtrl::SetSortArrow(int iColumn, bool bAscending, int iColumnIndex)
{
	ArrowType		arrowType;

	if (bAscending)
	{
		switch (iColumnIndex)
		{
			case 0:
				arrowType = arrowUp;
				break;
			case 1:
				arrowType = arrowUp1;
				break;
			case 2:
				arrowType = arrowUp2;
				break;
			case 3:
				arrowType = arrowUp3;
				break;
			default:
				arrowType = arrowUp;
		}
	}
	else
	{
		switch (iColumnIndex)
		{
			case 1:
				arrowType = arrowDown1;
				break;
			case 2:
				arrowType = arrowDown2;
				break;
			case 3:
				arrowType = arrowDown3;
				break;
			default:
				arrowType = arrowDown;
		}
	}
	SetSortArrow(iColumn, arrowType);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::SetSortArrow(int iColumn, ArrowType atType)
{
	HDITEM		headerItem;

	headerItem.mask = HDI_FORMAT;

	CHeaderCtrl		*pHeaderCtrl = GetHeaderCtrl();
	int				iSortIndex = 0;
	bool			bMultiColumn = false;

	if (atType == arrowUp1 || atType == arrowDown1)
	{
		bMultiColumn = true;
	}
	if (atType == arrowUp2 || atType == arrowDown2)
	{
		iSortIndex = 2 - 1;
		bMultiColumn = true;
	}
	else if (atType == arrowUp3 || atType == arrowDown3)
	{
		iSortIndex = 3 - 1;
		bMultiColumn = true;
	}
//
//	Delete old image if column has changed
	if (iColumn != m_iCurrentSortItem[iSortIndex])
	{
		if (m_iCurrentSortItem[iSortIndex] >= 0)
		{
			pHeaderCtrl->GetItem(m_iCurrentSortItem[iSortIndex], &headerItem);
			headerItem.fmt &= ~(HDF_IMAGE | HDF_BITMAP_ON_RIGHT);
		}
		else
			headerItem.fmt = HDF_STRING;
		pHeaderCtrl->SetItem(m_iCurrentSortItem[iSortIndex], &headerItem);
		m_iCurrentSortItem[iSortIndex] = iColumn;
		m_imlHeaderCtrl.DeleteImageList();
	}
	if (!bMultiColumn)
	{
		headerItem.fmt = HDF_STRING;
		pHeaderCtrl->SetItem(m_iCurrentSortItem[1], &headerItem);
		m_iCurrentSortItem[1] = -1;
		pHeaderCtrl->SetItem(m_iCurrentSortItem[2], &headerItem);
		m_iCurrentSortItem[2] = -1;
	}
//place new arrow unless we were given an invalid column
	if (iColumn >= 0 && pHeaderCtrl->GetItem(iColumn, &headerItem))
	{
		m_atSortArrow[iSortIndex] = atType;

		HINSTANCE		hInstRes = AfxFindResourceHandle(MAKEINTRESOURCE(m_atSortArrow[0]), RT_BITMAP);
		HINSTANCE		hInstRes2 = NULL;
		HINSTANCE		hInstRes3 = NULL;

		if (hInstRes != NULL)
		{
			HBITMAP			hbmSortStates = (HBITMAP)::LoadImage(hInstRes, MAKEINTRESOURCE(m_atSortArrow[0]), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION | LR_LOADMAP3DCOLORS);

			if (hbmSortStates != NULL)
			{
				CBitmap			bmSortStates;

				bmSortStates.Attach(hbmSortStates);

				CImageList		imlSortStates;

				if (imlSortStates.Create(14, 14, ILC_COLOR | ILC_MASK, 1, 0))
				{
					VERIFY(imlSortStates.Add(&bmSortStates, RGB(255, 0, 255)) != -1);
					if (m_iCurrentSortItem[1] >= 0)
					{
						hInstRes2 = AfxFindResourceHandle(MAKEINTRESOURCE(m_atSortArrow[1]), RT_BITMAP);

						if (hInstRes2 != NULL)
						{
							HBITMAP		hbmSortStates2 = (HBITMAP)::LoadImage(hInstRes, MAKEINTRESOURCE(m_atSortArrow[1]), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION | LR_LOADMAP3DCOLORS);

							if (hbmSortStates2 != NULL)
							{
								CBitmap		bmSortStates2;

								bmSortStates2.Attach(hbmSortStates2);
								VERIFY(imlSortStates.Add(&bmSortStates2, RGB(255, 0, 255)) != -1);
							}
						}
					}
					if (m_iCurrentSortItem[2] >= 0)
					{
						hInstRes3 = AfxFindResourceHandle(MAKEINTRESOURCE(m_atSortArrow[2]), RT_BITMAP);

						if (hInstRes3 != NULL)
						{
							HBITMAP		hbmSortStates3 = (HBITMAP)::LoadImage(hInstRes, MAKEINTRESOURCE(m_atSortArrow[2]), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION | LR_LOADMAP3DCOLORS);

							if (hbmSortStates3 != NULL)
							{
								CBitmap		bmSortStates3;

								bmSortStates3.Attach(hbmSortStates3);
								VERIFY(imlSortStates.Add(&bmSortStates3, RGB(255, 0, 255)) != -1);
							}
						}
					}

				//	To avoid drawing problems (which occure only with an image list *with* a mask) while
				//	resizing list view columns which have the header control bitmap right aligned, set
				//	the background color of the image list.
					if (g_eMuleApp.m_qwComCtrlVer < MAKEDLLVERULL(6, 0, 0, 0))
						imlSortStates.SetBkColor(GetSysColor(COLOR_BTNFACE));

				//	When setting the image list for the header control for the first time we'll get
				//	the image list of the listview control!! So, better store the header control imagelist separate.
					(void)pHeaderCtrl->SetImageList(&imlSortStates);
					m_imlHeaderCtrl.DeleteImageList();
					m_imlHeaderCtrl.Attach(imlSortStates.Detach());

				//	Use smaller bitmap margins -- this saves some pixels which may be required for
				//	rather small column titles.
					if (g_eMuleApp.m_qwComCtrlVer >= MAKEDLLVERULL(5, 8, 0, 0))
					{
						int iBmpMargin = pHeaderCtrl->SendMessage(HDM_GETBITMAPMARGIN);
						int iNewBmpMargin = GetSystemMetrics(SM_CXEDGE) + GetSystemMetrics(SM_CXEDGE) / 2;
						if (iNewBmpMargin < iBmpMargin)
							pHeaderCtrl->SendMessage(HDM_SETBITMAPMARGIN, iNewBmpMargin);
					}
				}
			}
		}
		headerItem.mask |= HDI_IMAGE;
		headerItem.fmt |= HDF_IMAGE | HDF_BITMAP_ON_RIGHT;
		headerItem.iImage = iSortIndex;
		pHeaderCtrl->SetItem(iColumn, &headerItem);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//move item in list, returns index of new item
int CMuleListCtrl::MoveItem(int iOldIndex, int iNewIndex)
{

//eklmn: prevent redraw during moving (delete & inserts)
	SetRedraw(FALSE);

	if (iNewIndex > iOldIndex)
		iNewIndex--;

//save substrings
	CSimpleArray<void*>		aSubItems;
	DWORD					Style = GetStyle();

	if ((Style & LVS_OWNERDATA) == 0)
	{
		TCHAR			szText[256];
		LVITEM			lvi;

		lvi.mask = LVIF_TEXT | LVIF_NORECOMPUTE;
		lvi.iItem = iOldIndex;
		for (int i = 1; i < m_iColumnsTracked; i++)
		{
			lvi.iSubItem = i;
			lvi.cchTextMax = ARRSIZE(szText);
			lvi.pszText = szText;
			void* pstrSubItem = NULL;
			if (GetItem(&lvi))
			{
				if (lvi.pszText == LPSTR_TEXTCALLBACK)
					pstrSubItem = LPSTR_TEXTCALLBACK;
				else
					pstrSubItem = new CString(lvi.pszText);
			}
			aSubItems.Add(pstrSubItem);
		}
	}

//copy item
	LVITEM			lvi_copy;
	TCHAR			szText[256];

	lvi_copy.mask = LVIF_TEXT | LVIF_STATE | LVIF_PARAM | LVIF_INDENT | LVIF_IMAGE | LVIF_NORECOMPUTE;
	lvi_copy.stateMask = (UINT) - 1;
	lvi_copy.iItem = iOldIndex;
	lvi_copy.iSubItem = 0;	//set subitem pointer on Item
	lvi_copy.pszText = szText;
	lvi_copy.cchTextMax = sizeof(szText) / sizeof(szText[0]);
	lvi_copy.iIndent = 0;
	if (GetItem(&lvi_copy) == 0)
		return -1;

//do the move
	VERIFY (0 != DeleteItem(iOldIndex));

	lvi_copy.iItem = iNewIndex;

	iNewIndex = InsertItem(&lvi_copy);

//restore substrings
	if ((Style & LVS_OWNERDATA) == 0)
	{
		LVITEM			lvi;

		lvi.iItem = iNewIndex;
		for (int i = 1; i < m_iColumnsTracked; i++)
		{
			lvi.iSubItem = i;
			void* pstrSubItem = aSubItems[i - 1];
			if (pstrSubItem != NULL)
			{
				if (pstrSubItem == LPSTR_TEXTCALLBACK)
					lvi.pszText = LPSTR_TEXTCALLBACK;
				else
					lvi.pszText = (LPTSTR)((LPCTSTR) * ((CString*)pstrSubItem));
				DefWindowProc(LVM_SETITEMTEXT, iNewIndex, (LPARAM) & lvi);
				if (pstrSubItem != LPSTR_TEXTCALLBACK)
					delete (CString*)pstrSubItem;
			}
		}
	}

	SetRedraw(TRUE);

	return iNewIndex;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CMuleListCtrl::UpdateLocation(int iItem)
{
	int				iItemCount = GetItemCount();

	if (iItem >= iItemCount || iItem < 0)
		return iItem;

	DWORD_PTR		dwpItemData = GetItemData(iItem);

	if (dwpItemData == NULL)
		return iItem;

	bool			notLast = iItem < (iItemCount -1);
	bool			notFirst = iItem > 0;
	int				iLowerBound = 0;
	int				iUpperBound = iItemCount - 1;
	POSITION		posItem, posPrevItem, posNextItem;

//	if it isn't  first element, we gonna compare it with previous element
	if (notFirst)
	{
	//	get Index of previous element (Item-1)
		int			iNewIndex = iItem - 1;

		posItem = m_Params.FindIndex(iNewIndex);

	//	check if we need to move an Item
		int			iResult = m_SortProc(dwpItemData, GetParamAt(posItem, iNewIndex), m_dwParamSort);

	//	we need to move it
		if (iResult < 0)
		{
			posPrevItem = posItem;

			int			iDist = iNewIndex;

			while (iDist >= 1)
			{
				for (int i = 0; i < iDist; i++)
					m_Params.GetPrev(posPrevItem);

				if (m_SortProc(dwpItemData, GetParamAt(posPrevItem, iNewIndex - iDist), m_dwParamSort) < 0)
				{
				//	update an Index & postion
					iNewIndex = iNewIndex - iDist;
					posItem = posPrevItem;
				//	calculate new step based on new Index
					iDist = ceil((iNewIndex - iLowerBound) / 2.0);
				}
				else
				{
				//	set postion to initial one, if we don't need to move
					posPrevItem = posItem;
				//	set new boundary
					iLowerBound = iNewIndex - iDist;
				//	just decrease a step
					iDist = iDist / 2;
				}
			}
			return MoveItem(iItem, iNewIndex);
		}
	}

	if (notLast)
	{
	//	get Index of next element (Item+1)
		int			iNewIndex = iItem + 1;

		posItem = m_Params.FindIndex(iNewIndex);

	//	check if we need to move an Item
		int			iResult = m_SortProc(dwpItemData, GetParamAt(posItem, iNewIndex), m_dwParamSort);

		if (iResult > 0)
		{
			posNextItem = posItem;

			int			iDist = (iItemCount-1) - iNewIndex ;

			while (iDist >= 1)
			{
				for (int i = 0; i < iDist; i++)
					m_Params.GetNext(posNextItem);

				if (m_SortProc(dwpItemData, GetParamAt(posNextItem, iNewIndex + iDist), m_dwParamSort) > 0)
				{
				//	update an Index & postion
					iNewIndex = iNewIndex + iDist;
					posItem = posNextItem;
				//	calculate new step based on new Index
					iDist = ceil((iUpperBound - iNewIndex) / 2.0);
				}
				else
				{
				//	set postion to initial one, if we don't need to move
					posNextItem = posItem;
				//	set new boundary
					iUpperBound = iNewIndex + iDist;
				//	just decrease a step
					iDist = iDist / 2;
				}
			}
			iNewIndex++;
			return MoveItem(iItem, iNewIndex);
		}
	}

	return iItem;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DWORD_PTR CMuleListCtrl::GetItemData(int iItem)
{
	POSITION		pos = m_Params.FindIndex(iItem);
	LPARAM			lParam = 0;

	if (pos != NULL && iItem != -1)
	{
		lParam = GetParamAt(pos, iItem);
		MLC_ASSERT(lParam == CListCtrl::GetItemData(iItem));
	}

	return lParam;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CMuleListCtrl::OnNMDividerDoubleClick(NMHEADER* pHeader, LRESULT* pResult)
{
//	As long as we do not handle the HDN_DIVIDERDBLCLICK according the actual
//	listview item contents it's better to resize to the header width instead of
//	resizing to zero width. The complete solution for this would require a lot
//	of rewriting in the owner drawn listview controls...
	SetColumnWidth(pHeader->iItem, LVSCW_AUTOSIZE_USEHEADER);

	return *pResult = TRUE;		//netwolf: resize width to header (official 0.28a)
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	lower level than everything else so poorly overriden functions don't break us
BOOL CMuleListCtrl::OnWndMsg(UINT message, WPARAM wParam, LPARAM lParam, LRESULT* pResult)
{
//	lets look for the important messages that are essential to handle
	switch (message)
	{
		case WM_NOTIFY:
		{
			if (wParam == 0)
			{
				if (((NMHDR*)lParam)->code == NM_RCLICK)
				{
				//handle right click on headers and show column menu

					POINT		point;

					GetCursorPos (&point);

					CTitleMenu tmColumnMenu;

					tmColumnMenu.CreatePopupMenu();
					if (m_Name.GetLength() != 0)
						tmColumnMenu.AddMenuTitle(m_Name);

					CHeaderCtrl			*pHeaderCtrl = GetHeaderCtrl();
					int					iCount = pHeaderCtrl->GetItemCount();

					for (int iCurrent = 1; iCurrent < iCount; iCurrent++)
					{
						HDITEM		item;
						TCHAR		text[255];

						item.pszText = text;
						item.mask = HDI_TEXT;
						item.cchTextMax = 255;
						pHeaderCtrl->GetItem(iCurrent, &item);

						tmColumnMenu.AppendMenu( MF_STRING | ((m_aColumns[iCurrent].bHidden) ? 0 : MF_CHECKED),
												MLC_IDC_MENU + iCurrent, item.pszText );
					}
					tmColumnMenu.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);

				//	Menu objects are destroyed in their destructor
					return *pResult = TRUE;

				}
				else if (((NMHDR*)lParam)->code == HDN_BEGINTRACKA || ((NMHDR*)lParam)->code == HDN_BEGINTRACKW)
				{
				//stop them from changeing the size of anything "before" first column
					if (m_aColumns[((HD_NOTIFY*)lParam)->iItem].bHidden)
					{
						return *pResult = TRUE;
					}
				}
				else if (((NMHDR*)lParam)->code == HDN_ENDDRAG)
				{
				//stop them from moving first column

					NMHEADER		*pHeader = (NMHEADER*)lParam;

					if (pHeader->iItem != 0 && pHeader->pitem->iOrder != 0)
					{

						int			iNewLoc = pHeader->pitem->iOrder - GetHiddenColumnCount();

						if (iNewLoc > 0)
						{

							if (m_aColumns[pHeader->iItem].iLocation != iNewLoc)
							{

								if (m_aColumns[pHeader->iItem].iLocation > iNewLoc)
								{
									int			iMax = m_aColumns[pHeader->iItem].iLocation;
									int			iMin = iNewLoc;

									for (int i = 0; i < m_iColumnsTracked; i++)
									{
										if (m_aColumns[i].iLocation >= iMin && m_aColumns[i].iLocation < iMax)
											m_aColumns[i].iLocation++;
									}
								}

								else if (m_aColumns[pHeader->iItem].iLocation < iNewLoc)
								{
									int			iMin = m_aColumns[pHeader->iItem].iLocation;
									int			iMax = iNewLoc;

									for (int i = 0; i < m_iColumnsTracked; i++)
									{
										if (m_aColumns[i].iLocation > iMin && m_aColumns[i].iLocation <= iMax)
											m_aColumns[i].iLocation--;
									}
								}

								m_aColumns[pHeader->iItem].iLocation = iNewLoc;

								Invalidate(FALSE);
								break;
							}
						}
					}

					return *pResult = 1;
				}
				else if (((NMHDR*)lParam)->code == HDN_DIVIDERDBLCLICKA || ((NMHDR*)lParam)->code == HDN_DIVIDERDBLCLICKW)
				{
					if (GetStyle() & LVS_OWNERDRAWFIXED)
					{
						return OnNMDividerDoubleClick(((NMHEADER*)lParam), pResult);
					}
				}
			}
			break;
		}
		case WM_COMMAND:
		{
		//	deal with menu clicks

			if (wParam == MLC_IDC_UPDATE)
			{
				UpdateLocation(lParam);
				return *pResult = 1;

			}
			else if (wParam >= MLC_IDC_MENU)
			{
				CHeaderCtrl			*pHeaderCtrl = GetHeaderCtrl();
				int					iCount = pHeaderCtrl->GetItemCount();
				int					iToggle = wParam - MLC_IDC_MENU;

				if (iToggle >= iCount)
					break;

				if (m_aColumns[iToggle].bHidden)
					ShowColumn(iToggle);
				else
					HideColumn(iToggle);

				return *pResult = 1;
			}
			break;
		}
		case LVM_DELETECOLUMN:
		{
		//book keeping!
			if (m_aColumns != NULL)
			{
				for (int i = 0; i < m_iColumnsTracked; i++)
					if (m_aColumns[i].bHidden)
						ShowColumn(i);

				delete[] m_aColumns;
				m_aColumns = NULL; // 'new' may throw an exception
			}
			m_aColumns = new MULE_COLUMN[--m_iColumnsTracked];
			for (int i = 0; i < m_iColumnsTracked; i++)
			{
				m_aColumns[i].iLocation = i;
				m_aColumns[i].bHidden = false;
			}
			break;
		}
	//	case LVM_INSERTCOLUMN:
		case LVM_INSERTCOLUMNA:
		case LVM_INSERTCOLUMNW:
		{
		//book keeping!
			if (m_aColumns != NULL)
			{
				for (int i = 0; i < m_iColumnsTracked; i++)
					if (m_aColumns[i].bHidden)
						ShowColumn(i);

				delete[] m_aColumns;
				m_aColumns = NULL; // 'new' may throw an exception
			}
			m_aColumns = new MULE_COLUMN[++m_iColumnsTracked];
			for (int i = 0; i < m_iColumnsTracked; i++)
			{
				m_aColumns[i].iLocation = i;
				m_aColumns[i].bHidden = false;
			}
			break;
		}
		case LVM_SETITEM:
		{
		//	book keeping
			{
				POSITION pos = m_Params.FindIndex(((LPLVITEM)lParam)->iItem);
				if (pos)
				{
					m_Params.SetAt(pos, MLC_MAGIC);
					PostMessage(LVM_UPDATE, ((LPLVITEM)lParam)->iItem);
				}
			}
			break;
		}
		case LVN_KEYDOWN:
		{
			break;
		}
		case LVM_SETITEMTEXT:
		{
		//	need to check for movement
			*pResult = DefWindowProc(message, wParam, lParam);
			if (*pResult)
				PostMessage(WM_COMMAND, MLC_IDC_UPDATE, wParam);
			return *pResult;
		}
		case LVM_SORTITEMS:
		{
		//	book keeping...
			m_dwParamSort = (LPARAM)wParam;
			m_SortProc = (PFNLVCOMPARE)lParam;
			for (POSITION pos = m_Params.GetHeadPosition(); pos != NULL; m_Params.GetNext(pos))
				m_Params.SetAt(pos, MLC_MAGIC);
			break;
		}
		case LVM_DELETEALLITEMS:
		{
		//	book keeping...
			if (!CListCtrl::OnWndMsg(message, wParam, lParam, pResult) && DefWindowProc(message, wParam, lParam))
				m_Params.RemoveAll();
			return *pResult = TRUE;
		}
		case LVM_DELETEITEM:
		{
		//	book keeping.....
			MLC_ASSERT(m_Params.GetAt(m_Params.FindIndex(wParam)) == CListCtrl::GetItemData(wParam));
			if (!CListCtrl::OnWndMsg(message, wParam, lParam, pResult) && DefWindowProc(message, wParam, lParam))
				m_Params.RemoveAt(m_Params.FindIndex(wParam));
			return *pResult = TRUE;
		}
	//	case LVM_INSERTITEM:
		case LVM_INSERTITEMA:
		case LVM_INSERTITEMW:
	//	try to fix position of inserted items
		{
			LPLVITEM		pItem = (LPLVITEM)lParam;
			int				iItem = pItem->iItem;
			int				iItemCount = GetItemCount();
			bool			notLast = iItem < iItemCount;
			bool			notFirst = iItem > 0;
			int				iLowerBound = 0;
			int				iUpperBound =  iItemCount -1;

			if (notFirst)
			{
				// get Index of previous element (Item-1)
				int				iNewIndex = iItem - 1;
				POSITION		pos = m_Params.FindIndex(iNewIndex);
				// check if we need to move an Item
				int				iResult = m_SortProc(pItem->lParam, GetParamAt(pos, iNewIndex), m_dwParamSort);

				if (iResult < 0)
				{
					POSITION		posPrev = pos;
					int				iDist = iNewIndex;

					while(iDist >= 1)
					{
						for (int i = 0; i < iDist; i++)
							m_Params.GetPrev(posPrev);

						if (m_SortProc(pItem->lParam, GetParamAt(posPrev, iNewIndex - iDist), m_dwParamSort) < 0)
						{
							// update an Index & postion
							iNewIndex = iNewIndex - iDist;
							pos = posPrev;
							// calculate new step based on new Index
							iDist = ceil((iNewIndex - iLowerBound)/2.0);
						}
						else
						{
							// set postion to initial one, if we don't need to move
							posPrev = pos;
							// set new boundary
							iLowerBound = iNewIndex - iDist;
							// just decrease a step
							iDist = iDist/2;
						}
					}
					pItem->iItem = iNewIndex;
					notLast = false;
				}
			}

			if (notLast)
			{
				// get Index of next element
				int				iNewIndex = iItem;
				POSITION		pos = m_Params.FindIndex(iNewIndex);
				// check if we need to move an Item
				int				iResult = m_SortProc(pItem->lParam, GetParamAt(pos, iNewIndex), m_dwParamSort);

				if (iResult > 0)
				{
					POSITION		posNext = pos;
					int				iDist = (iItemCount - 1) - iItem;

					while(iDist >= 1)
					{
						for (int i = 0; i < iDist; i++)
							m_Params.GetNext(posNext);

						if (m_SortProc(pItem->lParam, GetParamAt(posNext, iNewIndex + iDist), m_dwParamSort) > 0)
						{
							// update an Index & postion
							iNewIndex = iNewIndex + iDist;
							pos = posNext;
							// calculate new step based on new Index
							iDist = ceil((iUpperBound - iNewIndex) / 2.0);
						}
						else
						{
							// set postion to initial one, if we don't need to move
							posNext = pos;
							// set new boundary
							iUpperBound = iNewIndex + iDist;
							// just decrease a step
							iDist = iDist/2;
						}
					}
					iNewIndex++;
					pItem->iItem = iNewIndex;
				}
			}

			if (pItem->iItem == 0)
			{
				m_Params.AddHead(pItem->lParam);
				return FALSE;
			}

			LRESULT lResult = DefWindowProc(message, wParam, lParam);
			if (lResult != -1)
			{
				if (lResult >= GetItemCount())
					m_Params.AddTail(pItem->lParam);
				else if (lResult == 0)
					m_Params.AddHead(pItem->lParam);
				else
					m_Params.InsertAfter(m_Params.FindIndex(lResult - 1), pItem->lParam);
			}
			return *pResult = lResult;
			break;
		}
		case LVM_UPDATE:
		{
		//	better fix for old problem... normally Update(int) causes entire list to redraw

			if (wParam == UpdateLocation(wParam))
			{ //no need to invalidate rect if item moved
				RECT		rcItem;
				bool		bResult = GetItemRect(wParam, &rcItem, LVIR_BOUNDS);

				if (bResult)
					InvalidateRect(&rcItem, FALSE);
				return *pResult = bResult;
			}
			return *pResult = TRUE;
		}
	}

	return CListCtrl::OnWndMsg(message, wParam, lParam, pResult);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if (nChar == 'A' && ::GetAsyncKeyState(VK_CONTROL) < 0)
	{
	//	Ctrl+A: Select all items
		LV_ITEM			theItem;

		theItem.mask = LVIF_STATE;
		theItem.iItem = -1;
		theItem.iSubItem = 0;
		theItem.state = LVIS_SELECTED;
		theItem.stateMask = 2;
		SetItemState( -1, &theItem);
	}
	else if (nChar == VK_DELETE)
		PostMessage(WM_COMMAND, MP_DELETE, 0);
	else if (m_bGeneralPurposeFind)
	{
		if (nChar == 'F' && (GetKeyState(VK_CONTROL) & 0x8000))
		{
		//	Ctrl+F: Search item
			OnFindStart();
		}
		else if (nChar == VK_F3)
		{
			if (GetKeyState(VK_SHIFT) & 0x8000)
			{
			//	Shift+F3: Search previous
				OnFindPrev();
			}
			else
			{
			//	F3: Search next
				OnFindNext();
			}
		}
	}

	return CListCtrl::OnKeyDown(nChar, nRepCnt, nFlags);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CMuleListCtrl::OnChildNotify(UINT message, WPARAM wParam, LPARAM lParam, LRESULT* pResult)
{
	if (message != WM_DRAWITEM)
	{
	//	Catch the prepaint and copy struct
		if ( message == WM_NOTIFY && ((NMHDR*)lParam)->code == NM_CUSTOMDRAW
		     && ((LPNMLVCUSTOMDRAW)lParam)->nmcd.dwDrawStage == CDDS_ITEMPREPAINT )
		{
			m_bCustomDraw = CListCtrl::OnChildNotify(message, wParam, lParam, pResult);

			if (m_bCustomDraw)
				memcpy2(&m_lvcd, (void*)lParam, sizeof(NMLVCUSTOMDRAW));

			return m_bCustomDraw;
		}

		return CListCtrl::OnChildNotify(message, wParam, lParam, pResult);
	}

	ASSERT(pResult == NULL); // no return value expected
	UNUSED(pResult);         // unused in release builds

	DrawItem((LPDRAWITEMSTRUCT)lParam);

	return TRUE;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	int					iItem = lpDrawItemStruct->itemID;
	CImageList			*pImageList;
	CHeaderCtrl			*pHeaderCtrl = GetHeaderCtrl();

//	gets the item image and state info
	LV_ITEM				lvi;

	lvi.mask = LVIF_IMAGE | LVIF_STATE;
	lvi.iItem = iItem;
	lvi.iSubItem = 0;
	lvi.stateMask = LVIS_DROPHILITED | LVIS_FOCUSED | LVIS_SELECTED | LVIS_GLOW | LVIS_CUT;
	GetItem(&lvi);

//	see if the item be highlighted
	bool			bHighlight = ((lvi.state & LVIS_DROPHILITED) || (lvi.state & LVIS_SELECTED));
	bool			bCtrlFocused = ((GetFocus() == this) || (GetStyle() & LVS_SHOWSELALWAYS));
	bool			bGlowing = (lvi.state & LVIS_GLOW);
	bool			bDimmed = (lvi.state & LVIS_CUT);

//	set up our ficker free drawing
	CRect			rcItem(lpDrawItemStruct->rcItem);
	int				iState = SaveDC(lpDrawItemStruct->hDC);
	CDC				*oDC = CDC::FromHandle(lpDrawItemStruct->hDC);
	COLORREF		crOldDCBkColor = oDC->SetBkColor(m_crWindow);

	CMemDC pDC(oDC, &rcItem);

	CFont			*pOldFont = pDC->SelectObject(GetFont());
	COLORREF		crOldTextColor;

	if (m_bCustomDraw)
		crOldTextColor = pDC->SetTextColor(m_lvcd.clrText);
	else if (bDimmed)
		crOldTextColor = pDC->SetTextColor(m_crDimmedText);
	else if (bGlowing)
		crOldTextColor = pDC->SetTextColor(m_crGlowText);
	else
		crOldTextColor = pDC->SetTextColor(m_crWindowText);

	if (m_crWindowTextBk == CLR_NONE)
		DefWindowProc(WM_ERASEBKGND, (WPARAM)pDC->m_hDC, 0);

	int iOffset = pDC->GetTextExtent(_T(" "), 1).cx * 2;

//	get rectangles for drawing
	CRect			rcBounds, rcLabel, rcIcon;

	GetItemRect(iItem, rcBounds, LVIR_BOUNDS);
	GetItemRect(iItem, rcLabel, LVIR_LABEL);
	GetItemRect(iItem, rcIcon, LVIR_ICON);

	CRect			rcCol(rcBounds);

//	the label!
	CString			sLabel = GetItemText(iItem, 0);
//	labels are offset by a certain amount
//	this offset is related to the width of a space character
	CRect			rcHighlight;
	CRect			rcWnd;

//	should I check (GetExtendedStyle() & LVS_EX_FULLROWSELECT) ?
	rcHighlight.top = rcBounds.top;
	rcHighlight.bottom = rcBounds.bottom;
	rcHighlight.left = rcBounds.left + 1;
	rcHighlight.right = rcBounds.right - 1;

	COLORREF		crOldBckColor;
	CBrush			*pOldBrush;
//	draw the background color
	if (bHighlight)
	{
		if (bCtrlFocused)
		{
			CBrush			hBrush(m_crHighlight);

			pOldBrush = pDC->SelectObject(&hBrush);
			pDC->FillRect(rcHighlight, &hBrush);
			pDC->SelectObject(pOldBrush);
			crOldBckColor = pDC->SetBkColor(m_crHighlight);
		}
		else if (bGlowing)
		{
			CBrush			nhBrush(m_crGlow);

			pOldBrush = pDC->SelectObject(&nhBrush);
			pDC->FillRect(rcHighlight, &nhBrush);
			pDC->SelectObject(pOldBrush);
			crOldBckColor = pDC->SetBkColor(m_crGlow);
		}
		else
		{
			CBrush			nhBrush(m_crNoHighlight);

			pOldBrush = pDC->SelectObject(&nhBrush);
			pDC->FillRect(rcHighlight, &nhBrush);
			pDC->SelectObject(pOldBrush);
			crOldBckColor = pDC->SetBkColor(m_crNoHighlight);
		}
	}
	else
	{
		if (bGlowing)
		{
			CBrush			nhBrush(m_crGlow);

			pOldBrush = pDC->SelectObject(&nhBrush);
			pDC->FillRect(rcHighlight, &nhBrush);
			pDC->SelectObject(pOldBrush);
			crOldBckColor = pDC->SetBkColor(m_crGlow);
		}
		else
		{
			if (m_crWindowTextBk != CLR_NONE)
			{
				CBrush			wBrush(m_crWindow);

				pOldBrush = pDC->SelectObject(&wBrush);
				pDC->FillRect(rcHighlight, &wBrush);
				pDC->SelectObject(pOldBrush);
			}
			crOldBckColor = pDC->SetBkColor(m_crWindow);
		}
	}

//	update column
	rcCol.right = rcCol.left + GetColumnWidth(0);

//	draw state icon
	if (lvi.state & LVIS_STATEIMAGEMASK)
	{
		int				nImage = ((lvi.state & LVIS_STATEIMAGEMASK) >> 12) - 1;

		pImageList = GetImageList(LVSIL_STATE);
		if (pImageList)
		{
			COLORREF			crOld = pImageList->SetBkColor(CLR_NONE);

			pImageList->Draw(pDC, nImage, rcCol.TopLeft(), ILD_NORMAL);
			pImageList->SetBkColor(crOld);
		}
	}

//	draw the item's icon
	pImageList = GetImageList(LVSIL_SMALL);
	if (pImageList)
	{
		COLORREF			crOld = pImageList->SetBkColor(CLR_NONE);

		pImageList->Draw(pDC, lvi.iImage, rcIcon.TopLeft(), ILD_NORMAL);
		pImageList->SetBkColor(crOld);
	//	moved to fix problem with serachlist when FileTypeIcon disabled
		rcLabel.left += iOffset / 2;
	}
	else
		rcLabel.left = rcBounds.left + 6;

	int				iOldBkMode = (m_crWindowTextBk == CLR_NONE) ? pDC->SetBkMode(TRANSPARENT) : OPAQUE;

//	draw item label (column 0)
	rcLabel.right -= iOffset;
	pDC->DrawText(sLabel, -1, rcLabel, MLC_DT_TEXT | DT_LEFT | DT_NOCLIP);

//	draw labels for remaining columns
	LV_COLUMN			lvc;

	lvc.mask = LVCF_FMT | LVCF_WIDTH;
	rcBounds.right = rcHighlight.right > rcBounds.right ? rcHighlight.right : rcBounds.right;

	int				iCount = pHeaderCtrl->GetItemCount();

	for (int iCurrent = 1; iCurrent < iCount; iCurrent++)
	{
		int				iColumn = pHeaderCtrl->OrderToIndex(iCurrent);

	//	don't draw column 0 again
		if (iColumn == 0)
			continue;

		GetColumn(iColumn, &lvc);
	//	don't draw anything with 0 width
		if (lvc.cx == 0)
			continue;

		rcCol.left = rcCol.right;
		rcCol.right += lvc.cx;

		sLabel = GetItemText(iItem, iColumn);
		if (sLabel.GetLength() == 0)
			continue;

	//	get the text justification
		UINT			nJustify = DT_LEFT;

		switch (lvc.fmt & LVCFMT_JUSTIFYMASK)
		{
			case LVCFMT_RIGHT:
				nJustify = DT_RIGHT;
				break;
			case LVCFMT_CENTER:
				nJustify = DT_CENTER;
				break;
			default:
				break;
		}

		if (IsRightToLeftLanguage())
			nJustify |= DT_RTLREADING;

		rcLabel = rcCol;
		rcLabel.left += iOffset;
		rcLabel.right -= iOffset;

		pDC->DrawText(sLabel, -1, rcLabel, MLC_DT_TEXT | nJustify);
	}

//	draw focus rectangle if item has focus
	if ((lvi.state & LVIS_FOCUSED) && (bCtrlFocused || (lvi.state & LVIS_SELECTED)))
	{
		if (!bCtrlFocused || !(lvi.state & LVIS_SELECTED))
		{
			CBrush			nfBrush(m_crNoFocusLine);

			pOldBrush = pDC->SelectObject(&nfBrush);
			pDC->FrameRect(rcHighlight, &nfBrush);
			pDC->SelectObject(pOldBrush);
		}
		else
		{
			CBrush			fBrush(m_crFocusLine);

			pOldBrush = pDC->SelectObject(&fBrush);
			pDC->FrameRect(rcHighlight, &fBrush);
			pDC->SelectObject(pOldBrush);
		}
	}

	pDC->Flush();
	if (m_crWindowTextBk == CLR_NONE)
		pDC->SetBkMode(iOldBkMode);
	pDC->SelectObject(pOldFont);
	pDC->SetTextColor(crOldTextColor);
	pDC->SetBkColor(crOldBckColor);
	RestoreDC(lpDrawItemStruct->hDC, iState);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CMuleListCtrl::OnEraseBkgnd(CDC* pDC)
{
	int				itemCount = GetItemCount();

	if (!itemCount)
		return CListCtrl::OnEraseBkgnd(pDC);

	RECT			clientRect;
	RECT			itemRect;
	int				topIndex = GetTopIndex();
	int				maxItems = GetCountPerPage();
	int				drawnItems = itemCount < maxItems ? itemCount : maxItems;
	CRect			rcClip;

//	draw top portion
	GetClientRect(&clientRect);
	rcClip = clientRect;
	GetItemRect(topIndex, &itemRect, LVIR_BOUNDS);
	clientRect.bottom = itemRect.top;
	if (m_crWindowTextBk != CLR_NONE)
		pDC->FillSolidRect(&clientRect, GetBkColor());
	else
		rcClip.top = itemRect.top;

//	draw bottom portion if we have to
	if (topIndex + maxItems >= itemCount)
	{
		GetClientRect(&clientRect);
		GetItemRect(topIndex + drawnItems - 1, &itemRect, LVIR_BOUNDS);
		clientRect.top = itemRect.bottom;
		rcClip.bottom = itemRect.bottom;
		if (m_crWindowTextBk != CLR_NONE)
			pDC->FillSolidRect(&clientRect, GetBkColor());
	}

//	draw right half if we need to
	if (itemRect.right < clientRect.right)
	{
		GetClientRect(&clientRect);
		clientRect.left = itemRect.right;
		rcClip.right = itemRect.right;
		if (m_crWindowTextBk != CLR_NONE)
			pDC->FillSolidRect(&clientRect, GetBkColor());
	}

	if (m_crWindowTextBk == CLR_NONE)
	{
		CRect			rcClipBox;

		pDC->GetClipBox(&rcClipBox);
		rcClipBox.SubtractRect(&rcClipBox, &rcClip);
		if (!rcClipBox.IsRectEmpty())
		{
			pDC->ExcludeClipRect(&rcClip);
			CListCtrl::OnEraseBkgnd(pDC);
			InvalidateRect(&rcClip, FALSE);
		}
	}
	return TRUE;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::OnSysColorChange()
{
//	adjust colors
	CListCtrl::OnSysColorChange();
	SetColors();

//	redraw the up/down sort arrow (if it's there)
	if (m_iCurrentSortItem[0] >= 0)
		SetSortArrow(m_iCurrentSortItem[0], (ArrowType)m_atSortArrow[0]);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::ApplyImageList(HIMAGELIST himl)
{
	SendMessage(LVM_SETIMAGELIST, LVSIL_SMALL, (LPARAM)himl);
	if (m_imlHeaderCtrl.m_hImageList != NULL)
	{
	//	Must *again* set the image list for the header control, because LVM_SETIMAGELIST
	//	always resets any already specified header control image lists!
		GetHeaderCtrl()->SetImageList(&m_imlHeaderCtrl);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//	CDlgListSearchListSearch

class CDlgListSearchListSearch : public CDialog
{
	DECLARE_DYNAMIC(CDlgListSearchListSearch)

public:
	CDlgListSearchListSearch(CWnd* pParent = NULL);   // standard constructor

// Dialog Data
	enum { IDD = IDD_LISTVIEW_SEARCH };

	CMuleListCtrl	*m_pListView;
	CString			m_strFindText;
	int				m_iSearchColumn;

protected:
	CComboBox m_ctlSearchCol;

	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support

public:
	afx_msg void OnEnChangeSearchText();
	virtual BOOL OnInitDialog();
};


IMPLEMENT_DYNAMIC(CDlgListSearchListSearch, CDialog)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CDlgListSearchListSearch::CDlgListSearchListSearch(CWnd* pParent /*=NULL*/)
		: CDialog(CDlgListSearchListSearch::IDD, pParent)
		, m_strFindText(_T(""))
{
	m_pListView = NULL;
	m_iSearchColumn = 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDlgListSearchListSearch::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_LISTVIEW_SEARCH_COLUMN, m_ctlSearchCol);
	DDX_CBIndex(pDX, IDC_LISTVIEW_SEARCH_COLUMN, m_iSearchColumn);
	DDX_Text(pDX, IDC_LISTVIEW_SEARCH_TEXT, m_strFindText);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CDlgListSearchListSearch::OnInitDialog()
{
	CDialog::OnInitDialog();

	SetDlgItemText(IDC_LISTVIEW_SEARCH_TEXT_LBL, GetResString(IDS_SEARCH_TEXT) + _T(':'));
	SetDlgItemText(IDC_LISTVIEW_SEARCH_COLUMN_LBL, GetResString(IDS_SEARCH_COLUMN) + _T(':'));

	SetDlgItemText(IDCANCEL, GetResString(IDS_CANCEL));

	SetIcon(g_eMuleApp.m_pdlgEmule->m_hiconSourceTray, FALSE);
	SetWindowText(GetResString(IDS_SW_SEARCHBOX));

	if (m_pListView != NULL)
	{
		TCHAR			szColTitle[256];
		LVCOLUMN		lvc;

		lvc.mask = LVCF_TEXT;
		lvc.cchTextMax = sizeof(szColTitle) / sizeof(szColTitle[0]);
		lvc.pszText = szColTitle;

		int				iCol = 0;

		while (m_pListView->GetColumn(iCol++, &lvc))
			m_ctlSearchCol.AddString(lvc.pszText);
		if ((UINT)m_iSearchColumn >= (UINT)m_ctlSearchCol.GetCount())
			m_iSearchColumn = 0;
	}
	else
	{
		m_ctlSearchCol.EnableWindow(FALSE);
		m_ctlSearchCol.ShowWindow(SW_HIDE);

		m_iSearchColumn = 0;
	}
	m_ctlSearchCol.SetCurSel(m_iSearchColumn);

	return TRUE;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::DoFind(int iStartItem, int iDirection /*1=down, 0 = up*/, BOOL bShowError)
{
	CWaitCursor curHourglass;

	if (iStartItem < 0)
	{
		MessageBeep((UINT) - 1);
		return;
	}

	int			iNumItems = iDirection ? GetItemCount() : 0;
	int			iItem = iStartItem;

	while (iDirection ? iItem < iNumItems : iItem >= 0)
	{
		CString				strItemText(GetItemText(iItem, m_iFindColumn));

	//	TRACE("iItem=%d  szText=%s\n", iItem, strItemText);
		if (!strItemText.IsEmpty())
		{
			if ( m_bFindMatchCase
			     ? _tcsstr(strItemText, m_strFindText) != NULL
			     : stristr(strItemText, m_strFindText) != NULL )
			{
			//	Deselect all listview entries
				SetItemState( -1, 0, LVIS_SELECTED);

			//	Select the found listview entry
				SetItemState(iItem, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
				SetSelectionMark(iItem);
				EnsureVisible(iItem, FALSE /*bPartialOK*/);
				SetFocus();

				return;
			}
		}

		if (iDirection)
			iItem++;
		else
			iItem--;
	}

	if (bShowError)
		AfxMessageBox(_T(GetResString(IDS_SEARCH_NORESULT)), MB_ICONINFORMATION);
	else
		MessageBeep((UINT) - 1);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::OnFindStart()
{
	CDlgListSearchListSearch		dlg;

	dlg.m_pListView = this;
	dlg.m_strFindText = m_strFindText;
//	dlg.m_bMatchCase = m_bFindMatchCase;
//	dlg.m_iSearchDirection = m_iFindDirection;
	dlg.m_iSearchColumn = m_iFindColumn;
	if (dlg.DoModal() != IDOK)
		return;
	m_strFindText = dlg.m_strFindText;
//	m_bFindMatchCase = dlg.m_bMatchCase;
//	m_iFindDirection = dlg.m_iSearchDirection;
	m_iFindColumn = dlg.m_iSearchColumn;

	if (!m_strFindText.IsEmpty())
	{
		DoFindNext(TRUE /*bShowError*/);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::OnFindNext()
{
	if (!m_strFindText.IsEmpty())
		DoFindNext(FALSE /*bShowError*/);
	else
		OnFindStart();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::DoFindNext(BOOL bShowError)
{
	int			iStartItem = GetNextItem( -1, LVNI_SELECTED | LVNI_FOCUSED);

	if (iStartItem == -1)
		iStartItem = 0;
	else
		iStartItem = iStartItem + (m_iFindDirection ? 1 : -1);
	DoFind(iStartItem, m_iFindDirection, bShowError);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CMuleListCtrl::OnFindPrev()
{
	if (!m_strFindText.IsEmpty())
	{
		int			iStartItem = GetNextItem( -1, LVNI_SELECTED | LVNI_FOCUSED);

		if (iStartItem == -1)
			iStartItem = 0;
		else
			iStartItem = iStartItem + (!m_iFindDirection ? 1 : -1);

		DoFind(iStartItem, !m_iFindDirection, FALSE /*bShowError*/);
	}
	else
		OnFindStart();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
