//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


//	ServerListCtrl.cpp : implementation file
//

#include "stdafx.h"
#include "emule.h"
#include "ServerListCtrl.h"
//#include "Sockets/EmEngine.h"
#include "otherfunctions.h"
#include "IP2Country.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

//	CServerListCtrl

IMPLEMENT_DYNAMIC(CServerListCtrl, CMuleListCtrl /*CTreeCtrl*/)

//#region message map
BEGIN_MESSAGE_MAP(CServerListCtrl, CMuleListCtrl /*CTreeCtrl*/)
//	ON_NOTIFY_REFLECT(NM_RCLICK, OnNMRclickServlist)
	ON_NOTIFY_REFLECT(LVN_COLUMNCLICK, OnLvnColumnclickServlist)
	ON_NOTIFY_REFLECT (NM_DBLCLK, OnNMLdblclk)
	ON_WM_CONTEXTMENU()
	ON_WM_DESTROY()
END_MESSAGE_MAP()
//#endregion

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CServerListCtrl::CServerListCtrl()
{
	memset(&m_bSortAscending, true, sizeof(m_bSortAscending));
	m_pServerList = 0;
	SetGeneralPurposeFind(true);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CServerListCtrl::Init(CServerList *pServerList)
{
	static const uint16 s_auIconResID[] =
	{
		IDI_SMALLSERVER,
		IDI_SMALLSERVERSTATIC,
		IDI_SERVERCONNECTED,
		IDI_SERVERCONNECTEDSTATIC,
		IDI_SERVERCONNECTEDFAILED,
		IDI_SERVERCONNECTEDFAILEDSTATIC,
		IDI_SERVERFAILED,
		IDI_SERVERFAILEDSTATIC
	};

	ModifyStyle(LVS_SINGLESEL,0);

//	Set Imagelist
	m_pServerList = pServerList;
	m_imageList.Create(16, 16, ILC_COLOR32 | ILC_MASK, 0, 10);
	m_imageList.SetBkColor(CLR_NONE);
	FillImgLstWith16x16Icons(&m_imageList, s_auIconResID, ELEMENT_COUNT(s_auIconResID));
	SetImageList(&m_imageList, LVSIL_SMALL);

	SetExtendedStyle(LVS_EX_FULLROWSELECT);

	InsertColumn(SL_COLUMN_SERVERNAME, GetResString(IDS_SL_SERVERNAME), LVCFMT_LEFT, 150);
	InsertColumn(SL_COLUMN_SERVERIP, GetResString(IDS_IP), LVCFMT_LEFT, 140);
	InsertColumn(SL_COLUMN_DESCRIPTION, GetResString(IDS_DESCRIPTION) , LVCFMT_LEFT, 150);
	InsertColumn(SL_COLUMN_PING, GetResString(IDS_PING), LVCFMT_RIGHT, 50);

	CString	strTemp;

	strTemp.Format(_T("%s (%s)"), GetResString(IDS_UUSERS), GetResString(IDS_MAXUSERS));
	InsertColumn(SL_COLUMN_NUMUSERS, strTemp, LVCFMT_RIGHT, 50);
	InsertColumn(SL_COLUMN_NUMFILES, GetResString(IDS_FILES) , LVCFMT_RIGHT, 50);
	InsertColumn(SL_COLUMN_PREFERENCES, GetResString(IDS_PRIORITY), LVCFMT_LEFT, 60);
	InsertColumn(SL_COLUMN_FAILEDCOUNT, GetResString(IDS_UFAILED), LVCFMT_RIGHT, 50);
	InsertColumn(SL_COLUMN_STATIC, GetResString(IDS_STATICSERVER), LVCFMT_LEFT, 50);
	InsertColumn(SL_COLUMN_SOFTFILELIMIT, GetResString(IDS_SERVER_SOFTHARDLIMIT), LVCFMT_RIGHT, 100);
	InsertColumn(SL_COLUMN_SOFTWAREVER, GetResString(IDS_SERVER_VERSION), LVCFMT_LEFT, 150);
	InsertColumn(SL_COLUMN_COUNTRY, GetResString(IDS_COUNTRY) , LVCFMT_LEFT, 150);

	m_bSortAscending[SL_COLUMN_NUMUSERS] = false;
	m_bSortAscending[SL_COLUMN_NUMFILES] = false;
	m_bSortAscending[SL_COLUMN_FAILEDCOUNT] = false;
	m_bSortAscending[SL_COLUMN_SOFTFILELIMIT] = false;
	m_bSortAscending[SL_COLUMN_SOFTWAREVER] = false;

	LoadSettings(CPreferences::TABLE_SERVER);

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CServerListCtrl::~CServerListCtrl()
{}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerListCtrl::OnDestroy()
{
	m_imageList.DeleteImageList();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	SortFirstInit() starts initial sorting after all initial information is added w/o sorting.
//	This is faster than adding every item with sorting rules.
void CServerListCtrl::SortFirstInit()
{
	if (g_eMuleApp.m_pGlobPrefs->DoUseSort())
	{
		SortInit(g_eMuleApp.m_pGlobPrefs->GetServerSortCol());
	}
	else
	{
		int	iSortColumn = g_eMuleApp.m_pGlobPrefs->GetColumnSortItem(CPreferences::TABLE_SERVER);
		bool	bSortAscending = g_eMuleApp.m_pGlobPrefs->GetColumnSortAscending(CPreferences::TABLE_SERVER);

		SortInit(iSortColumn | (bSortAscending ? MLC_SORTASC : MLC_SORTDESC));
	}
	ShowFilesCount();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerListCtrl::SortInit(int iSortCode)
{
//	Get the sort column
	int		iSortColumn = (iSortCode & MLC_COLUMNMASK);

//	Check max. column in case of configuration file corruption
	if (static_cast<uint32>(iSortColumn) >= SL_COLUMN_NUMCOLUMNS)
		iSortCode = iSortColumn = SL_COLUMN_SERVERNAME;

//	Get the sort order
	bool	bSortAscending = (iSortCode & MLC_SORTDESC) == 0;
//	Get the alternate sort
	int		iSortAltFlag = (iSortCode & MLC_SORTALT);

	m_bSortAscending[iSortColumn] = bSortAscending;
	if (iSortAltFlag == 0)
		SetSortArrow(iSortColumn, bSortAscending);
	else
	{
		SetSortArrow(iSortColumn, bSortAscending ? arrowDoubleUp : arrowDoubleDown);
	}
	SortItems(SortProc, iSortCode);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerListCtrl::Localize()
{
	if (GetSafeHwnd())
	{
		CHeaderCtrl	* pHeaderCtrl = GetHeaderCtrl();
		HDITEM	hdi;

		hdi.mask = HDI_TEXT;

		CString	strRes;

		strRes = GetResString(IDS_SL_SERVERNAME);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_SERVERNAME, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_IP);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_SERVERIP, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_DESCRIPTION);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_DESCRIPTION, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_PING);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_PING, &hdi);
		strRes.ReleaseBuffer();

		strRes.Format(_T("%s (%s)"), GetResString(IDS_UUSERS), GetResString(IDS_MAXUSERS));
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_NUMUSERS, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_FILES);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_NUMFILES, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_PRIORITY);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_PREFERENCES, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_UFAILED);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_FAILEDCOUNT, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_STATICSERVER);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_STATIC, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_SERVER_SOFTHARDLIMIT);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_SOFTFILELIMIT, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_SERVER_VERSION);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_SOFTWAREVER, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_COUNTRY);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(SL_COLUMN_COUNTRY, &hdi);
		strRes.ReleaseBuffer();

		for (POSITION pos = m_pServerList->m_serverList.GetHeadPosition(); pos != NULL;)
		{
			CServer*	pServer = m_pServerList->m_serverList.GetNext(pos);
			int			iIndex = FindIndex(pServer);

			if (iIndex != -1)
				RefreshServerDesc(iIndex, *pServer);
		}
		if (GetItemCount() != 0)
			ShowFilesCount();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CServerListCtrl::FindIndex(const CServer *pServer) const
{
	LVFINDINFO	find;

	memzero(&find, sizeof(LVFINDINFO));
	find.flags = LVFI_PARAM;
	find.lParam = (LPARAM)pServer;

	return FindItem(&find);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	RefreshServerDesc() refreshes the server description for the 'iIndex'th server from 'server'.
void CServerListCtrl::RefreshServerDesc(int iIndex, const CServer &server)
{
	CServer		*cur_srv;
	int			iImageIndex = 0;
	CString		strTemp;
	UINT		iItemState = 0;

	if (g_eMuleApp.m_pServerConnect->IsConnected()
		&& (cur_srv = g_eMuleApp.m_pServerConnect->GetCurrentServer()) != NULL
		&& cur_srv->GetPort() == server.GetPort()
		&& stricmp(cur_srv->GetAddress(), server.GetAddress()) == 0)
	{
		(server.GetFailedCount() == 0) ? iImageIndex = 2 : iImageIndex = 4;
		iItemState |= LVIS_GLOW;
	}
	else if (server.GetFailedCount() > 0)
	{
		iImageIndex = 6;
		if (!server.IsStaticMember())
			iItemState |= LVIS_CUT;
	}

	if (server.IsStaticMember())
		iImageIndex++;

	SetItemState(iIndex, iItemState, LVIS_GLOW | LVIS_CUT);
	SetItem(iIndex, 0, LVIF_IMAGE, 0, iImageIndex, 0, 0, 0, 0);

	if (server.GetListName())
	{
		SetItemText(iIndex, SL_COLUMN_SERVERNAME, server.GetListName());
	}

	if (server.GetAuxPort() != 0)
	{
		strTemp.Format(_T("%s : %u (%u)"), server.GetAddress(), server.GetPort(), server.GetAuxPort());
	}
	else
	{
		strTemp.Format(_T("%s : %u"), server.GetAddress(), server.GetPort());
	}
	SetItemText(iIndex, SL_COLUMN_SERVERIP, strTemp);

	if (server.GetDescription())
	{
		SetItemText(iIndex, SL_COLUMN_DESCRIPTION, server.GetDescription());
	}

	if (server.GetPing())
	{
		strTemp.Format(_T("%i"), server.GetPing());
		SetItemText(iIndex, SL_COLUMN_PING, strTemp);
	}

	if (server.GetNumUsers())
	{
		if (server.GetMaxUsers())
			strTemp.Format(_T("%i (%i)"), server.GetNumUsers(), server.GetMaxUsers());
		else
			strTemp.Format(_T("%i"), server.GetNumUsers());
		SetItemText(iIndex, SL_COLUMN_NUMUSERS, strTemp);
	}

	if (server.GetFiles())
	{
		strTemp.Format(_T("%i"), server.GetFiles());
		SetItemText(iIndex, SL_COLUMN_NUMFILES, strTemp);
	}

	switch (server.GetPreferences())
	{
		case PR_LOW:
		{
			SetItemText(iIndex, SL_COLUMN_PREFERENCES, GetResString(IDS_PRIOLOW));
			break;
		}
		case PR_NORMAL:
		{
			SetItemText(iIndex, SL_COLUMN_PREFERENCES, GetResString(IDS_PRIONORMAL));
			break;
		}
		case PR_HIGH:
		{
			SetItemText(iIndex, SL_COLUMN_PREFERENCES, GetResString(IDS_PRIOHIGH));
			break;
		}
		default:
		{
			SetItemText(iIndex, SL_COLUMN_PREFERENCES, GetResString(IDS_PRIONOPREF));
		}
	}

	strTemp.Format(_T("%i"), server.GetFailedCount());
	SetItemText(iIndex, SL_COLUMN_FAILEDCOUNT, strTemp);

	SetItemText(iIndex, SL_COLUMN_STATIC, YesNoStr(server.IsStaticMember()));

	if (server.GetHardMaxFiles())
	{
		if (server.GetSoftMaxFiles())
			strTemp.Format(_T("%i (%i)"), server.GetSoftMaxFiles(), server.GetHardMaxFiles());
		else
			strTemp.Format(_T("(%i)"), server.GetHardMaxFiles());
		SetItemText(iIndex, SL_COLUMN_SOFTFILELIMIT, strTemp);
	}

	SetItemText(iIndex, SL_COLUMN_SOFTWAREVER, server.GetVersion());

	SetItemText(iIndex, SL_COLUMN_COUNTRY, server.GetCountryName());
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerListCtrl::RefreshAllServer()
{
	for(POSITION pos = m_pServerList->m_serverList.GetHeadPosition(); pos != NULL;)
	{
		CServer	* pServer = m_pServerList->m_serverList.GetAt(pos);
		RefreshServer(*pServer);
		m_pServerList->m_serverList.GetNext(pos);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerListCtrl::RemoveServer(CServer *pServer, bool bDelToList)
{
	int	iIndex = FindIndex(pServer);

	if (iIndex != -1)
	{
		m_pServerList->RemoveServer(GetServerAt(iIndex));
		DeleteItem(iIndex);
	}
	ShowFilesCount();
	return;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Remove Dead Servers
void CServerListCtrl::RemoveDeadServer()
{
	if (g_eMuleApp.m_pGlobPrefs->DeadServer())
	{
		SetRedraw(FALSE);

#ifndef USE_SAFE_ARRAY

		for (POSITION pos = m_pServerList->m_serverList.GetHeadPosition(); pos != NULL; m_pServerList->m_serverList.GetNext(pos))
		{
			CServer	* pServer = m_pServerList->m_serverList.GetAt(pos);

			if ( pServer->GetFailedCount() > g_eMuleApp.m_pGlobPrefs->GetDeadserverRetries()
			     && !pServer->IsStaticMember() )
			{
				RemoveServer(pServer);
				pos = m_pServerList->m_serverList.GetHeadPosition();
			}
		}
#else
		for (INT_PTR pos = 0; pos < m_pServerList->array.GetCount(); pos++)
		{
			CServer	*pServer = m_pServerList->GetServerAt(pos);

			if ( pServer->GetFailedCount() > g_eMuleApp.m_pGlobPrefs->GetDeadserverRetries()
				 && !pServer->IsStaticMember() )
			{
				RemoveServer(pServer);
				pos = 0;
			}
		}
#endif
		SetRedraw(TRUE);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerListCtrl::AppendServer(CServer *pServer)
{
	if (::IsWindow(GetSafeHwnd()))
	{
		int	iNumServers = GetItemCount();
		int	iIndex = 0;

		iIndex = InsertItem( LVIF_TEXT | LVIF_PARAM, iNumServers, pServer->GetListName(), 0, 0, 1,
							reinterpret_cast<LPARAM>(pServer) );
		if (pServer->GetFailedCount() < 0)
			pServer->ResetFailedCount();
		RefreshServerDesc(iIndex, *pServer);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CServerListCtrl::AddServer(CServer *pServer, bool bAddToList, bool bBulkLoad, bool bChangeServerInfo/* = false*/)
{
	if (pServer == NULL)
		return false;

	if (!m_pServerList->AddServer(pServer, bChangeServerInfo))
		return false;

	if (bAddToList && !bChangeServerInfo)
		AppendServer(pServer);
	if (!bBulkLoad)
		ShowFilesCount();
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerListCtrl::RefreshServer(CServer &server)
{
	if (!g_eMuleApp.m_pdlgEmule->IsRunning())
		return;

	if (!::IsWindow(m_hWnd))
		return;

	int	iIndex = FindIndex(&server);

	if (iIndex == -1)
		return;

	if (server.GetFailedCount() < 0)
		server.ResetFailedCount();
	RefreshServerDesc(iIndex, server);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//	CServerListCtrl message handlers

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerListCtrl::OnContextMenu(CWnd *pWnd, CPoint point)
{
//	On right click, we also want to change the current selection like the left click does
	CPoint	p = point;

	ScreenToClient(&p);

	int	iIndex = HitTest(p);

	if (iIndex != -1)
		SetSelectionMark(iIndex);

//	Create up-to-date popupmenu
	UINT	iMenuFlags;
	CTitleMenu	m_ServerMenu;
	CMenu	m_ServerPrioMenu;

	int			iSelectionMark = GetSelectionMark();
	CServer		*pServer = (iSelectionMark != -1) ? GetSelectedServer() : NULL;
	UINT		dwSelectedCount = GetSelectedCount();
	bool		bSelectedSrv = (dwSelectedCount > 0) && (pServer != NULL);

//	Set state of selection-dependent menuitems
	iMenuFlags = MF_STRING | (((iSelectionMark != -1) && bSelectedSrv) ? 0 : MF_GRAYED);

	m_ServerPrioMenu.CreateMenu();

	bool		bJustOne = (dwSelectedCount == 1);

	m_ServerPrioMenu.AppendMenu( MF_STRING |
		((bJustOne && pServer->GetPreferences() == PR_LOW) ? MF_CHECKED : MF_UNCHECKED),
		MP_PRIOLOW, GetResString(IDS_PRIOLOW) );
	m_ServerPrioMenu.AppendMenu( MF_STRING |
		((bJustOne && pServer->GetPreferences() == PR_NORMAL) ? MF_CHECKED : MF_UNCHECKED),
		MP_PRIONORMAL, GetResString(IDS_PRIONORMAL) );
	m_ServerPrioMenu.AppendMenu( MF_STRING |
		((bJustOne && pServer->GetPreferences() == PR_HIGH) ? MF_CHECKED : MF_UNCHECKED),
		MP_PRIOHIGH, GetResString(IDS_PRIOHIGH) );

	m_ServerMenu.CreatePopupMenu();
	m_ServerMenu.AddMenuTitle(GetResString(IDS_EM_SERVER));
	m_ServerMenu.AppendMenu(iMenuFlags, MP_CONNECTTO, GetResString(IDS_CONNECTTHIS));
	m_ServerMenu.AppendMenu(iMenuFlags | MF_POPUP, (UINT_PTR)m_ServerPrioMenu.m_hMenu, GetResString(IDS_PRIORITY));

//	Hide greyed-out "add /remove from static server"
	if (bSelectedSrv)
	{
		if (pServer->IsStaticMember())
			m_ServerMenu.AppendMenu(iMenuFlags, MP_REMOVEFROMSTATIC, GetResString(IDS_REMOVEFROMSTATIC));
		else
			m_ServerMenu.AppendMenu(iMenuFlags, MP_ADDTOSTATIC, GetResString(IDS_ADDTOSTATIC));
	}

	m_ServerMenu.AppendMenu(MF_STRING | MF_SEPARATOR);
	m_ServerMenu.AppendMenu( iMenuFlags |
		(((dwSelectedCount != 1) || pServer->IsStaticMember()) ? MF_GRAYED : MF_ENABLED),
		MP_EDIT, GetResString(IDS_FNCEDIT) );
	m_ServerMenu.AppendMenu(iMenuFlags, MP_REMOVE, GetResString(IDS_REMOVETHIS));
	m_ServerMenu.AppendMenu(MF_STRING, MP_REMOVEALL, GetResString(IDS_REMOVEALL));
	m_ServerMenu.AppendMenu(MF_SEPARATOR);
	m_ServerMenu.AppendMenu(iMenuFlags, MP_GETED2KLINK, GetResString(IDS_DL_LINK1));

	m_ServerMenu.SetDefaultItem(MP_CONNECTTO);
	m_ServerMenu.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);

//	Menu objects are destroyed in their destructor
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CServerListCtrl::OnCommand(WPARAM wParam, LPARAM lParam)
{
	if (wParam == MP_REMOVEALL)
	{
#ifdef OLD_SOCKETS_ENABLED
		if (g_eMuleApp.m_pServerConnect->IsConnecting())
		{
			g_eMuleApp.m_pDownloadQueue->StopUDPRequests();
			g_eMuleApp.m_pServerConnect->StopConnectionTry();
			g_eMuleApp.m_pServerConnect->Disconnect();
			g_eMuleApp.m_pdlgEmule->ShowConnectionState(false);
		}
#endif //OLD_SOCKETS_ENABLED
		SetRedraw(FALSE);
		m_pServerList->RemoveAllServers();
		DeleteAllItems();
		SetRedraw(TRUE);
		ShowFilesCount();
		return true;
	}

	int	iIndex = GetSelectionMark();

	if (iIndex != -1)
	{
		if (GetSelectedServer() != NULL)
		{
			switch (wParam)
			{
			//#region Connect To Server
				case MP_CONNECTTO:
				{
					if (GetSelectedCount() > 1)
					{
						CServer	* pServer;
#ifdef OLD_SOCKETS_ENABLED

						g_eMuleApp.m_pServerConnect->Disconnect();
#endif OLD_SOCKETS_ENABLED

						POSITION	pos = GetFirstSelectedItemPosition();

					//	Move selected servers to the start of the list, to connect them first
						while (pos != NULL)
						{
							iIndex = GetNextSelectedItem(pos);
							if (iIndex > -1)
							{
								pServer = GetServerAt(iIndex);
								g_eMuleApp.m_pServerList->MoveServerDown(pServer);
							}
						}
#ifdef OLD_SOCKETS_ENABLED
						g_eMuleApp.m_pServerConnect->ConnectToAnyServer((g_eMuleApp.m_pServerList->GetServerCount() - GetSelectedCount()), false);
#else OLD_SOCKETS_ENABLED

						g_eMuleApp.m_pEngine->ConnectToAnyServer();
#endif OLD_SOCKETS_ENABLED

					}
					else
					{
#ifdef OLD_SOCKETS_ENABLED
						if (g_eMuleApp.m_pServerConnect->IsConnected())
						{
							if ( !_tcscmp( GetSelectedServer()->GetAddress(),
							               g_eMuleApp.m_pServerConnect->GetCurrentServer()->GetAddress() ) )
							{
								AddLogLine(false, IDS_ALREADYCONNECTED, GetSelectedServer()->GetListName());
								break;
							}
						}
						g_eMuleApp.m_pServerConnect->ConnectToServer(*GetSelectedServer());
#endif OLD_SOCKETS_ENABLED
#ifdef NEW_SOCKETS

						CServer	*pServer = GetSelectedServer();

						if (pServer != NULL)
							g_eMuleApp.m_pEngine->ConnectToServer(pServer);
#endif NEW_SOCKETS

					}
					g_eMuleApp.m_pdlgEmule->ShowConnectionState(false);
					break;
				}
			//#endregion
			//#region Rename Server
				case MP_EDIT:
				{
					CServer	*pServer = GetSelectedServer();
					CNewServerDlg dlgEditServer;
					CString strPort;
					CString strAuxPort;

					dlgEditServer.SetParent(&g_eMuleApp.m_pdlgEmule->m_wndServer);
					dlgEditServer.SetServerEditMode();
					strPort.Format(_T("%u"), pServer->GetPort());
					strAuxPort.Format(_T("%u"), pServer->GetAuxPort());
					dlgEditServer.SetLabels(pServer->GetAddress(), strPort, pServer->GetListName(), strAuxPort);
					if (dlgEditServer.DoModal() == IDOK)
					{
						g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
					}
					break;
				}
			//#endregion
			//#region Remove Server
				case MP_REMOVE:
				{
					SetRedraw(FALSE);

					POSITION	pos;

					while (GetFirstSelectedItemPosition() != NULL)
					{
						pos = GetFirstSelectedItemPosition();
						iIndex = GetNextSelectedItem(pos);
						m_pServerList->RemoveServer(GetServerAt(iIndex));
						DeleteItem(iIndex);
					}
					ShowFilesCount();
					SetRedraw(TRUE);
					break;
				}
			//#endregion
			//#region Add to Static Server List
				case MP_ADDTOSTATIC:
				{
					POSITION	pos = GetFirstSelectedItemPosition();

					while (pos != NULL)
					{
						CServer	* pServer = GetServerAt(GetNextSelectedItem(pos));

						if (!StaticServerFileAppend(pServer))
							return false;
						pServer->SetIsStaticMember(true);
						g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
					}
					break;
				}
			//#endregion
			//#region Remove from Static Server List
				case MP_REMOVEFROMSTATIC:
				{
					POSITION	pos = GetFirstSelectedItemPosition();

					while (pos != NULL)
					{
						CServer	* pServer = GetServerAt(GetNextSelectedItem(pos));

						if (!StaticServerFileRemove(pServer))
							return false;
						pServer->SetIsStaticMember(false);
						g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
					}
					break;
				}
			//#endregion
			//#region Set Server Priority to "Low"
				case MP_PRIOLOW:
				{
					POSITION	pos = GetFirstSelectedItemPosition();

					while (pos != NULL)
					{
						CServer	* pServer = GetServerAt(GetNextSelectedItem(pos));

						pServer->SetPreference(CServer::SERVERPRIORITY_LOW);
						if (pServer->IsStaticMember())
							StaticServerFileAppend(pServer);
						g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
					}
					break;
				}
			//#endregion
			//#region Set Server Priority to "Normal"
				case MP_PRIONORMAL:
				{
					POSITION	pos = GetFirstSelectedItemPosition();

					while (pos != NULL)
					{
						CServer	* pServer = GetServerAt(GetNextSelectedItem(pos));

						pServer->SetPreference(CServer::SERVERPRIORITY_NORMAL);
						if (pServer->IsStaticMember())
							StaticServerFileAppend(pServer);
						g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
					}
					break;
				}
			//#endregion
			//#region Set Server Priority to "High"
				case MP_PRIOHIGH:
				{
					POSITION	pos = GetFirstSelectedItemPosition();

					while (pos != NULL)
					{
						CServer	* pServer = GetServerAt(GetNextSelectedItem(pos));

						pServer->SetPreference(CServer::SERVERPRIORITY_HIGH);
						if (pServer->IsStaticMember())
							StaticServerFileAppend(pServer);
						g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
					}
					break;
				}
			//#endregion
			//#region Get ED2K Link for Server
				case MP_GETED2KLINK:
				{
					POSITION	pos = GetFirstSelectedItemPosition();
					CString	strBuffer, strLink;

					while (pos != NULL)
					{
						CServer	* pServer = GetServerAt(GetNextSelectedItem(pos));

						strBuffer.Format(_T("ed2k://|server|%s|%u|/"), pServer->GetFullIP(), pServer->GetPort());
						if (strLink.GetLength() > 0)
							strLink += _T("\r\n");
						strLink += strBuffer;
					}
					g_eMuleApp.CopyTextToClipboard(strLink);
					break;
				}
			//#endregion
			}
		}
	}
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	OnNMLdblclk() is the left-double-click message handler
void CServerListCtrl::OnNMLdblclk(NMHDR *pNMHDR, LRESULT *pResult)
{
	if (GetSelectionMark() != ( -1))
	{
#ifdef OLD_SOCKETS_ENABLED
		if (g_eMuleApp.m_pServerConnect->IsConnected())
		{
			if ( !_tcscmp( GetSelectedServer()->GetAddress(),
			               (g_eMuleApp.m_pServerConnect->GetCurrentServer())->GetAddress() ) )
			{
				AddLogLine(false, IDS_ALREADYCONNECTED, GetSelectedServer()->GetListName());
				return;
			}
		}
		g_eMuleApp.m_pServerConnect->ConnectToServer(*GetSelectedServer());
#endif //OLD_SOCKETS_ENABLED
#ifdef NEW_SOCKETS

		CServer	*pServer = GetSelectedServer();

		if (pServer != NULL)
			g_eMuleApp.m_pEngine->ConnectToServer(pServer);
#else
		g_eMuleApp.m_pdlgEmule->ShowConnectionState(false);
#endif //NEW_SOCKETS
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	OnLvnColumnclickServlist() is the column header click message handler
void CServerListCtrl::OnLvnColumnclickServlist(NMHDR *pNMHDR, LRESULT *pResult)
{
	NM_LISTVIEW	*pNMListView = reinterpret_cast<NM_LISTVIEW*>(pNMHDR);
	int			iFlags = 0, iSubItem = pNMListView->iSubItem;
	bool		bSortOrder = m_bSortAscending[iSubItem];

//	For the two double columns, toggle the Alt column flag when the arrow is facing up
	if (iSubItem == SL_COLUMN_NUMUSERS)
	{
		if (((m_dwParamSort & MLC_COLUMNMASK) == iSubItem) && bSortOrder)
			m_bSortAscending[SL_COLUMN_NUMCOLUMNS] = !m_bSortAscending[SL_COLUMN_NUMCOLUMNS];
		iFlags = m_bSortAscending[SL_COLUMN_NUMCOLUMNS] ? 0 : MLC_SORTALT;
	}
	if (iSubItem == SL_COLUMN_SOFTFILELIMIT)
	{
		if (((m_dwParamSort & MLC_COLUMNMASK) == iSubItem) && bSortOrder)
			m_bSortAscending[SL_COLUMN_NUMCOLUMNS + 1] = !m_bSortAscending[SL_COLUMN_NUMCOLUMNS + 1];
		iFlags = m_bSortAscending[SL_COLUMN_NUMCOLUMNS + 1] ? 0 : MLC_SORTALT;
	}

// Reverse sorting direction for the same column and keep the same if column was changed
	if ((m_dwParamSort & MLC_COLUMNMASK) == iSubItem)
		m_bSortAscending[iSubItem] = bSortOrder = !bSortOrder;

	if (iFlags == 0)
		SetSortArrow(iSubItem, bSortOrder);
	else
		SetSortArrow(iSubItem, (bSortOrder) ? arrowDoubleUp : arrowDoubleDown);
	SortItems(SortProc, iSubItem + iFlags + ((bSortOrder) ? MLC_SORTASC : MLC_SORTDESC));

	g_eMuleApp.m_pGlobPrefs->SetColumnSortItem(CPreferences::TABLE_SERVER, iSubItem);
	g_eMuleApp.m_pGlobPrefs->SetColumnSortAscending(CPreferences::TABLE_SERVER, bSortOrder);

	Invalidate();
	*pResult = 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CServerListCtrl::SortProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	CServer	*pServer1 = reinterpret_cast<CServer*>(lParam1);
	CServer	*pServer2 = reinterpret_cast<CServer*>(lParam2);

	if (pServer1 == NULL || pServer2 == NULL)
		return 0;

	int		iTemp = 0;
	int		iCounter1;
	int		iCounter2;
	int		iCompare = 0;
	int		iSortColumn = (lParamSort & MLC_COLUMNMASK);
	int		iSortAltFlag = (lParamSort & MLC_SORTALT);
	int		iSortMod = ((lParamSort & MLC_SORTDESC) == 0) ? 1 : -1;

	switch (iSortColumn + iSortAltFlag)
	{
		//#region Server Name
		case SL_COLUMN_SERVERNAME:
		{
			bool	bServerHasNoName1 = _tcsstr(pServer1->GetListName(), pServer1->GetFullIP());
			bool	bServerHasNoName2 = _tcsstr(pServer2->GetListName(), pServer2->GetFullIP());

			if (!bServerHasNoName1 && !bServerHasNoName2)
				iCompare = _tcsicmp(pServer1->GetListName(), pServer2->GetListName());
			else
				iCompare = bServerHasNoName1 ? (bServerHasNoName2 ? 0 : 1) : -1;

			if (iCompare == 0)
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERIP);
			break;
		}
		//#endregion
		//#region Server IP
		case SL_COLUMN_SERVERIP:
		{
		//	If both are dynamic IP...
			if (pServer1->HasDynIP() && pServer2->HasDynIP())
			{
				//iCompare = CString(pServer1->GetDynIP()).CompareNoCase(pServer2->GetDynIP());
				iCompare = _tcsicmp(pServer1->GetDynIP(), pServer2->GetDynIP());
			}
		//	If only the first is dynamic IP...
			else if (pServer1->HasDynIP())
			{
				iCompare = -1;
			}
		//	If only the second is dynamic IP...
			else if (pServer2->HasDynIP())
			{
				iCompare = 1;
			}
		//	If both have static IPs...
			else
			{
				CString		sIP1, sIP2, sTemp1, sTemp2;

				iCounter1 = iCounter2 = iTemp = 0;
				sIP1 = pServer2->GetFullIP();
				sIP2 = pServer1->GetFullIP();
				for (int i = 0; iTemp == 0 ; i++)
				{
					sTemp1 = sIP1.Tokenize(_T("."), iCounter1);
					sTemp2 = sIP2.Tokenize(_T("."), iCounter2);
					iTemp = _tstoi(sTemp2) - _tstoi(sTemp1);
				//	If the IPs are the same...
					if (i > 2)
					{
						iCompare = pServer1->GetPort() - pServer2->GetPort();
						break;
					}
					else if (iTemp != 0)
					{
						iCompare = iTemp;
					}
				}
			}
			break;
		}
		//#endregion
		//#region Description
		case SL_COLUMN_DESCRIPTION:
		{
			CString strServerDescription1 = pServer1->GetDescription();
			CString strServerDescription2 = pServer2->GetDescription();

			if (!strServerDescription1.IsEmpty() && !strServerDescription2.IsEmpty())
				iCompare = _tcsicmp(strServerDescription1, strServerDescription2);
			else
			{
				iCompare = strServerDescription1.IsEmpty() ? (strServerDescription2.IsEmpty() ? 0 : 1) : -1;
				iSortMod = 1;		//empty entries at the bottom in any case
			}

			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		//#endregion
		//#region Ping
		case SL_COLUMN_PING:
		{
			iCompare = pServer1->GetPing() - pServer2->GetPing();
			if (!pServer1->GetPing() || !pServer2->GetPing())
				iSortMod = -1;
			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		//#endregion
		//#region Num Users / Max Users
		case SL_COLUMN_NUMUSERS:
		{
			iCompare = pServer1->GetNumUsers() - pServer2->GetNumUsers();
			if (!pServer1->GetNumUsers() || !pServer2->GetNumUsers())
				iSortMod = -1;
			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		case SL_COLUMN_NUMUSERS + MLC_SORTALT:    //max. users
		{
			iCompare = pServer1->GetMaxUsers() - pServer2->GetMaxUsers();
			if (!pServer1->GetMaxUsers() || !pServer2->GetMaxUsers())
				iSortMod = -1;
			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		//#endregion
		//#region Number of Files
		case SL_COLUMN_NUMFILES:
		{
			iCompare = pServer1->GetFiles() - pServer2->GetFiles();
			if (!pServer1->GetFiles() || !pServer2->GetFiles())
				iSortMod = -1;
			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		//#endregion
		//#region Preferences
		case SL_COLUMN_PREFERENCES:
		{
			iCompare = pServer1->GetPreferences() - pServer2->GetPreferences();
			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		//#endregion
		//#region Failed Count
		case SL_COLUMN_FAILEDCOUNT:
		{
			iCompare = pServer1->GetFailedCount() - pServer2->GetFailedCount();
			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		//#endregion
		//#region Static
		case SL_COLUMN_STATIC:
		{
			//	If both are static...
			int iStatic1 = pServer1->IsStaticMember() ? 1:0;
			int iStatic2 = pServer2->IsStaticMember() ? 1:0;

			iCompare = iStatic1 - iStatic2;
			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		//#endregion
		//#region Soft File Limit / Hard File Limit
		case SL_COLUMN_SOFTFILELIMIT:
		{
			iCompare = pServer1->GetSoftMaxFiles() - pServer2->GetSoftMaxFiles();
			if (!pServer1->GetSoftMaxFiles() || !pServer2->GetSoftMaxFiles())
				iSortMod = -1;
			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		case SL_COLUMN_SOFTFILELIMIT + MLC_SORTALT:    //Hard files asc
		{
			iCompare = pServer1->GetHardMaxFiles() - pServer2->GetHardMaxFiles();
			if (!pServer1->GetHardMaxFiles() || !pServer2->GetHardMaxFiles())
				iSortMod = -1;
			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		//#endregion
		//#region Software Version
		case SL_COLUMN_SOFTWAREVER:
		{
			CString strServerVersion1 = pServer1->GetVersion();
			CString strServerVersion2 = pServer2->GetVersion();

			if (!strServerVersion1.IsEmpty() && !strServerVersion2.IsEmpty())
				iCompare = _tcsicmp(strServerVersion1, strServerVersion2);
			else
			{
				iCompare = strServerVersion1.IsEmpty() ? (strServerVersion2.IsEmpty() ? 0 : 1) : -1;
				iSortMod = 1;		//empty entries at the bottom in any case
			}

			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//eklmn: sort allways in ascending order
			}
			break;
		}
		//#endregion
		//#region Country
		case SL_COLUMN_COUNTRY:
		{
			CString strServerCountry1 = pServer1->GetCountryName();
			CString strServerCountry2 = pServer2->GetCountryName();

			if (!strServerCountry1.IsEmpty() && !strServerCountry2.IsEmpty())
				iCompare = _tcsicmp(strServerCountry1, strServerCountry2);
			else
			{
				iCompare = strServerCountry1.IsEmpty() ? (strServerCountry2.IsEmpty() ? 0 : 1) : -1;
				iSortMod = 1;		//empty entries at the bottom in any case
			}

			if (iCompare == 0)
			{
				iCompare = SortProc(lParam1,lParam2,SL_COLUMN_SERVERNAME);
				iSortMod = 1;		//sort allways in ascending order
			}
			break;
		}
		//#endregion
		default:
		{
			iCompare = 0;
		}
	}

	return iCompare * iSortMod;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CServerListCtrl::StaticServerFileAppend(CServer *pServer)
{
	try
	{
	//	Remove any entry before writing to avoid duplicates
		StaticServerFileRemove(pServer);

		FILE	*pStaticServersFile = _tfopen(g_eMuleApp.m_pGlobPrefs->GetConfigDir() + CString(_T("staticservers.dat")), _T("a"));

		if (pStaticServersFile == NULL)
		{
			AddLogLine(false, IDS_ERROR_SSF);
			return false;
		}

		if ( fprintf( pStaticServersFile,
		              "%s:%i,%i,%s\n",
		              pServer->GetAddress(),
		              pServer->GetPort(),
		              pServer->GetPreferences(),
		              pServer->GetListName() ) != EOF )
		{
			if (!pServer->IsStaticMember())
				AddLogLine(false, _T("'%s:%i,%s' %s"), pServer->GetAddress(), pServer->GetPort(), pServer->GetListName(), GetResString(IDS_ADDED2SSF));
			pServer->SetIsStaticMember(true);
			g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
		}

		fclose(pStaticServersFile);
	}
	catch (...)
	{
		return false;
	}
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CServerListCtrl::StaticServerFileRemove(CServer *pServer)
{
	try
	{
		if (!pServer->IsStaticMember())
			return true;

		CString		strLine;
		CString		strTest;
		char		arrcBuffer[1024];
		int			iBufLen = 1024;
		int			iPos;
		CString		strStaticFilePath = g_eMuleApp.m_pGlobPrefs->GetConfigDir() + CString("staticservers.dat");
		CString		strStaticTempPath = g_eMuleApp.m_pGlobPrefs->GetConfigDir() + CString("statictemp.dat");
		FILE	   *pStaticServersFile = _tfopen(strStaticFilePath , _T("r"));
		FILE	   *pStaticTempFile = _tfopen(strStaticTempPath , _T("w"));

		if ((pStaticServersFile == NULL) || (pStaticTempFile == NULL))
		{
			AddLogLine(false, IDS_ERROR_SSF);
			return false;
		}

		while (!feof(pStaticServersFile))
		{
			if (fgets(arrcBuffer, iBufLen, pStaticServersFile) == 0)
				break;

			strLine = arrcBuffer;

		//	Ignore comments or invalid lines
			if (strLine.GetAt(0) == '#' || strLine.GetAt(0) == '/')
			{
				fprintf(pStaticTempFile, arrcBuffer);
				continue;
			}
			if (strLine.GetLength() < 5)
			{
				fprintf(pStaticTempFile, arrcBuffer);
				continue;
			}

		//	Only interested in "host:port"
			iPos = strLine.Find(',');
			if (iPos == -1)
				continue;
			strLine = strLine.Left(iPos);

		//	Get host and port from given pServer
			strTest.Format(_T("%s:%i"), pServer->GetAddress(), pServer->GetPort());

		//	Compare, if not the same pServer write original line to temp file
			if (strLine.Compare(strTest) != 0)
				fprintf(pStaticTempFile, arrcBuffer);
		}

		fclose(pStaticServersFile);
		fclose(pStaticTempFile);

	//	All ok, remove the existing file and replace with the new one
		CFile::Remove(strStaticFilePath);
		CFile::Rename(strStaticTempPath, strStaticFilePath);

		AddLogLine(false, _T("'%s:%i,%s' %s"), pServer->GetAddress(), pServer->GetPort(), pServer->GetListName(), GetResString(IDS_REMOVEDFROMSSF));
		pServer->SetIsStaticMember(false);
		g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
	}
	catch (...)
	{
		return false;
	}
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerListCtrl::ShowFilesCount()
{
	CString		strTitle;

	strTitle.Format(_T("%s (%u)"), GetResString(IDS_SV_SERVERLIST), GetItemCount());
	g_eMuleApp.m_pdlgEmule->m_wndServer.SetDlgItemText(IDC_SERVLIST_TEXT, strTitle);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
