//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#pragma once
#include "stdafx.h"
#include "types.h"
#ifdef OLD_SOCKETS_ENABLED
#include "ListenSocket.h"
#endif //OLD_SOCKETS_ENABLED
#include "partfile.h"
#include "ClientCredits.h"
#include "safefile.h"
#include "BarShader.h"
#include "Loggable.h"
#include <vector>
#include <list>
#include <hash_map>

enum _EnumChatStates
{
	MS_NONE				= 0,
	MS_CHATTING			= 1,
	MS_CONNECTING		= 2,
	MS_UNABLETOCONNECT	= 3,
};
typedef EnumDomain<_EnumChatStates>	EnumChatStates;

enum EnumSecureIdentState
{
	IS_UNAVAILABLE		= 0,
	IS_ALLREQUESTSSEND  = 0,
	IS_SIGNATURENEEDED	= 1,
	IS_KEYANDSIGNEEDED	= 2,
};
enum _EnumInfoPacketState
{
	IP_NONE				= 0,
	IP_EDONKEYPROTPACK  = 1,
	IP_EMULEPROTPACK	= 2,
	IP_BOTH				= 3,
};

struct IPRange_Struct2;

class EnumInfoPacketState : public EnumDomain<_EnumInfoPacketState>
{
public:
							EnumInfoPacketState() {};
							EnumInfoPacketState(const _EnumInfoPacketState eValue)
								: EnumDomain<_EnumInfoPacketState>(eValue) {}
	EnumInfoPacketState		&operator |=(const EnumInfoPacketState &e2) { m_value |= e2.m_value; return *this; }
};

#ifdef OLD_SOCKETS_ENABLED
class CClientReqSocket;
#endif //OLD_SOCKETS_ENABLED
class CFriend;

class CUpDownClient : public CLoggable
{
	friend class CUploadQueue;

public:
	bool				m_bIsHashThief;
#ifdef OLD_SOCKETS_ENABLED
	CClientReqSocket   *m_pRequestSocket;
#endif //OLD_SOCKETS_ENABLED
	CClientCredits	   *m_pCredits;
	CFriend			   *m_pFriend;
//	Upload
	uint32				m_dwCompressionGain;
	uint32				m_dwUncompressed;
	uchar				m_reqFileHash[16];
	CPartFile		   *m_pReqPartFile;
	int					m_iSourcesSlot;
	byte			   *m_pbyteUpPartStatuses;

	CTypedPtrList<CPtrList, CPartFile*>		m_otherRequestsList;
	CTypedPtrMap<CMapPtrToWord,CPartFile*, uint32>		m_otherNoNeededMap;

private:
	bool				m_bIsCommunity;
	bool				m_bCommunityIsCached;
	bool				m_bScoreBaseIsCached;
	bool				m_bIsOnLan;
	bool				m_bIsLowID;
	bool				m_bEmuleProtocol;
	bool				m_bCommentDirty;
	bool				m_bIsHybrid;
	bool				m_bIsML;
	bool				m_bSupportsAskSharedDirs;
	uint32				m_dwConnectIP;
	uint32				m_dwUserIP;
	uint32				m_dwServerIP;
	uint32				m_dwUserID;
	uint16				m_uUserPort;
	uint16				m_uServerPort;
	uint32				m_dwClientVersion;
	sint32				m_lUpDataRate;
	byte				m_byteEmuleVersion;
	byte				m_byteDataCompVer;
	byte				m_byteDataTarodVer;
	uint32				m_dwDataPlusVers;
	CString				m_strModVersion;
	CString				m_strModExtended;
	CString				m_strUserName;
	CString				m_strFullUserIP;
	BYTE				m_userHash[16];
	uint16				m_uUDPPort;
	byte				m_byteUDPVer;
	byte				m_byteSourceExchangeVer;
	byte				m_byteAcceptCommentVer;
	byte				m_byteExtendedRequestsVer;
	byte				m_byteFailedCount;
	EnumClientTypes		m_eClientSoft;
	uint32				m_dwLastSourceRequest;
	uint32				m_dwLastSourceAnswer;
	uint32				m_dwEnteredConnectedState;
	int					m_iFileListRequested;
	int					m_iFileListRequestedSave;
	uint32				m_dwCompatibleClient;

	uint32				m_dwRemoteScore;
	uint16				m_uRemoteRatio;
	uint16				m_uRemoteBaseModifier;
	bool				m_bRemoteInfoAvailable;
	uint16				m_uRemoteClientSpeed;

	CTypedPtrList<CPtrList, Packet*>		 m_waitingPacketsList;

	DWORD					m_dwLastRefreshedDLDisplayTime;

	EnumSecureIdentState	m_eSecureIdentState;
	EnumInfoPacketState		m_eInfoPacketsReceived;// have we received the edonkeyprot and emuleprot packet already (see InfoPacketsReceived() )
	uint32					m_dwLastSignatureIP;
	byte					m_byteSupportSecIdent;

	uint16					m_uLastPartAsked;

	bool				m_bBanned;
	uint32				m_dwTransferredUp;
	EnumULQState		m_eUploadState;
	uint32				m_dwWaitTime;
	uint32				m_dwUploadTime;
	uint32				m_dwMaxSendAllowed;
	uint32				m_dwAvgUpDataRate;
	uint32				m_dwAskedCount;
	uint32				m_dwLastUpRequest;
	uint32				m_dwBanTime;
	uint32				m_dwCurSessionUp;
	std::list<int>		m_averageUDRList;
	std::list<DWORD>	m_averageULTickList;

	CTypedPtrList<CPtrList, Packet*>		 			m_blockSendQueue;
	CTypedPtrList<CPtrList, Requested_Block_Struct*> 	m_blockRequestsQueue;
	CTypedPtrList<CPtrList, Requested_Block_Struct*> 	m_doneBlocksList;
	CTypedPtrList<CPtrList, Requested_File_Struct*>		m_requestedFilesList;
	DWORD				m_dwLastRefreshedULDisplayTime;
	DWORD				m_dwLastGotULDataTime;

//	Remote client's parts
	uint16				m_uUpPartCount;
	uint16				m_uAvailUpPartCount;
	uint16				m_uUpCompleteSourcesCount;
	uint32				m_dwUpCompleteSourcesTime;

	byte				m_byteFailedFileRequestsCount;
//	Download
	bool				m_bRemoteQueueFull;
	EnumDLQState		m_eDownloadState;
	uint16				m_uPartCount;
	uint32				m_dwDownAskedCount;
	byte			   *m_pbytePartStatuses;
	uint32				m_dwLastAskedTime;
	CString				m_strClientFilename;
	uint32				m_dwTransferredDown;
	bool				m_bTransferredDownMini;
	uint32				m_dwDownStartTime;
	uint32				m_dwLastBlockOffset;
	uint32				m_dwDownDataRate;
	uint32				m_dwAvgDownDataRate;
	uint32				m_dwLastBlockReceived;
	uint16				m_uRemoteQueueRank;
	int					m_iDeltaQueueRank;
	bool				m_bCompleteSource;
	bool				m_bReaskPending;
	bool				m_bUDPPending;
	uint32				m_dwQueueWaitTime;
	std::list<int>		m_averageDLDataRateList;
	std::list<DWORD>	m_averageDLTickList;
	uint64				m_qwClientSumDLDataRateOverLastNMeasurements;
	uint64				m_qwClientSumDLTickOverLastNMeasurements;
	uint32				m_dwLastDataRateCalculationTime;
	uint32				m_dwTransferredInLastPeriod;

	CTypedPtrList<CPtrList, Pending_Block_Struct*>	 	m_pendingBlocksList;
	CTypedPtrList<CPtrList, Requested_Block_Struct*> 	m_downloadBlocksList;

	uint16				m_uUserCountryIdx;

//	Chat
	EnumChatStates		m_eChatState;
	CString				m_strComment;
	EnumPartFileRating	m_eRating;
	uint32				m_dwAwayMessageResendCount;

//	L2HAC Mod: Low ID to High ID automatic callback
	uint32				m_dwL2HACTime;
	uint32				m_dwLastL2HACExec;
	bool				m_bL2HACEnabled;
	bool				m_bRequestingHashSet;

public:
//
//	Base
//
#ifdef OLD_SOCKETS_ENABLED
					CUpDownClient(CClientReqSocket* sender = NULL);
#endif //OLD_SOCKETS_ENABLED
					CUpDownClient(uint16 in_port, uint32 in_userid, uint32 in_serverup, uint16 in_serverport, CPartFile* in_reqfile);
				   ~CUpDownClient();
	bool			Disconnected(bool bRetryConnection = true);
	bool			TryToConnect(bool bIgnoreMaxCon = false);
	void			ConnectionEstablished();
//	Accessors
	uint32			GetUserID()	const			{return m_dwUserID;}
	CString			GetUserName() const			{return m_strUserName;}
	uint32			GetIP()	const				{return m_dwUserIP;}
	bool			HasLowID() const			{return m_bIsLowID;}
	CString			GetFullIP() const 			{return m_strFullUserIP;}
	uint16			GetUserPort() const			{return m_uUserPort;}
	uint32			GetTransferredUp() const	{return m_dwTransferredUp;}
	uint32			GetTransferredDown() const	{return m_dwTransferredDown;}
	uint32			GetServerIP() const			{return m_dwServerIP;}
	void			SetServerIP(uint32 dwIP)	{ m_dwServerIP = dwIP; }
	uint16			GetServerPort()	const		{return m_uServerPort;}
	void			SetServerPort(uint16 uPort)	{m_uServerPort = uPort;}
	uchar*			GetUserHash()				{return m_userHash;}
	bool			HasValidHash() const		{return (((int*)m_userHash)[0] != 0) || (((int*)m_userHash)[1] != 0) ||
														(((int*)m_userHash)[2] != 0) || (((int*)m_userHash)[3] != 0); }
	void			SetUserHash(uchar* m_achTempUserHash);
	uint32			GetVersion() const			{return m_dwClientVersion;}
	int				GetHashType();
	uint64			GetUID() const				{ return (static_cast<uint64>(m_dwUserIP)<<32)+(static_cast<uint64>(m_uUserPort)<<16); }

	byte			GetMuleVersion() const		{return m_byteEmuleVersion;}
	CString			GetModVersion()				{return m_strModVersion;}
	bool			ExtProtocolAvailable()		{return m_bEmuleProtocol;}
	bool			TarodVersion()				{return (m_byteDataTarodVer == 1 || m_byteDataTarodVer == 2);}
	uint32			GetPlusVersion()			{return m_dwDataPlusVers;}
	bool			IsEmuleClient()				{return m_byteEmuleVersion;}
	CClientCredits* Credits()					{return m_pCredits;}
	bool			IsBanned(bool bIgnoreDS = false) const		{return (m_bBanned && (bIgnoreDS || m_eDownloadState != DS_DOWNLOADING));}
	CString			GetClientFilename() const	{return m_strClientFilename;}
	bool			SupportsUDP()				{return m_byteUDPVer != 0 && m_uUDPPort != 0;}
	uint16			GetUDPPort()				{return m_uUDPPort;}
	void			SetUDPPort(uint16 nPort)	{ m_uUDPPort = nPort; }
	byte			GetUDPVersion()				{return m_byteUDPVer;}
	byte			GetExtendedRequestsVersion(){return m_byteExtendedRequestsVer;}
	bool			IsFriend()	const			{return m_pFriend != NULL;}
	double			GetCompression()	{return static_cast<double>(m_dwCompressionGain)/m_dwUncompressed*100.0;} // Add rod show compression
	void			ResetCompressionGain() {m_dwCompressionGain = 0; m_dwUncompressed=1;} // Add show compression

	void			RequestSharedFileList();
	void			ProcessSharedFileList(char* pachPacket, uint32 nSize, LPCTSTR pszDirectory = NULL);

	void			SetUserID(uint32 newUserID)
	{
		m_dwUserID = newUserID;
		m_bIsLowID = (m_dwUserID < 0x1000000);
	}
	void			SetUserName(const CString& pszNewName) {m_strUserName = pszNewName;}
	bool            IsCommunity();
	void            ResetValueCache();
	EnumClientTypes	GetClientSoft()				{return m_eClientSoft;}
	void			ReGetClientSoft();
	void			ProcessHelloAnswer(char* pachPacket, uint32 nSize);
	bool			ProcessHelloPacket(BYTE* pbytePacket, uint32 nSize);
	void			SendHelloAnswer();
	void			SendHelloPacket();
	void			SendMuleInfoPacket(bool bAnswer);
	void			ProcessMuleInfoPacket(BYTE* pbytePacket, uint32 dwSize);
	void			ProcessMuleCommentPacket(char* pbytePacket, uint32 dwSize);
	bool			Compare(CUpDownClient* tocomp);
	void			SetLastSrcReqTime()			{m_dwLastSourceRequest = ::GetTickCount();}
	void			SetLastSrcAnswerTime()		{m_dwLastSourceAnswer = ::GetTickCount();}
	uint32			GetLastSrcReqTime()			{return m_dwLastSourceRequest;}
	uint32			GetLastSrcAnswerTime()		{return m_dwLastSourceAnswer;}
	uint32			GetEnteredConnectedState()	{return m_dwEnteredConnectedState;}
	void			SetCommentDirty(bool bDirty = true) {m_bCommentDirty = bDirty;}
	byte			GetSourceExchangeVersion()	{return m_byteSourceExchangeVer;}

	void			SendPublicKeyPacket();
	void			SendSignaturePacket();
	void			ProcessPublicKeyPacket(uchar* pachPacket, uint32 nSize);
	void			ProcessSignaturePacket(uchar* pachPacket, uint32 nSize);
	byte			GetSecureIdentState()		{return m_eSecureIdentState;}
	void			SendSecIdentStatePacket();
	void			ProcessSecIdentStatePacket(uchar* pachPacket, uint32 nSize);
	void			InfoPacketsReceived();
	void			ResetFileStatusInfo();

	EnumULQState	GetUploadState()			{return m_eUploadState;}
	void			SetUploadState(EnumULQState eNewState);
	uint32			GetWaitStartTime()			{return m_dwWaitTime;}
	uint32			GetWaitTime()				{return m_dwUploadTime-m_dwWaitTime;}
	bool			IsDownloading()				{return (m_eUploadState == US_UPLOADING);}
	bool			HasBlocks()					{return !(m_blockSendQueue.IsEmpty() && m_blockRequestsQueue.IsEmpty());}
	sint32			GetDataRate()				{return m_lUpDataRate;}
	uint32			GetScore(bool sysvalue, bool isdownloading = false, bool onlybasevalue = false);
	double			GetRareFilePushRatio(void);
	double			GetSmallFilePushRatio(void);
	void			AddReqBlock(Requested_Block_Struct* reqblock);
	bool			CreateNextBlockPackage();
	void			SetUpStartTime(uint32 dwTime = 0);
	uint32			GetUpStartTimeDelay()		{return ::GetTickCount() - m_dwUploadTime;}
	void			SetWaitStartTime(uint32 dwTime = 0);
	void			SendHashsetPacket(char* forfileid);
	void			SetUploadFileID(uchar* tempreqfileid);
	uchar*			GetUploadFileID()	{return m_reqFileHash;}
	uint32			SendBlockData(uint32 maxammount, bool& wantsMoreBandwidth);
	void			ClearUploadBlockRequests();
	void			SendRankingInfo();
	void			SendCommentInfo(CKnownFile *file);
	uint32			GetLastAskedDelay();
	void			AddRequestCount(uchar* fileid);
	bool			IsDifferentPartBlock(bool silent=false);
    bool            IsForbiddenByJumpstart();
	void			UnBan();
	void			Ban(LPCTSTR line, const CString& reason,...); //katsyonak 16.06.03
	uint32			GetBanTime()				{return m_dwBanTime;}
	uint32			GetAskedCount()				{return m_dwAskedCount;}
	void			AddAskedCount()				{m_dwAskedCount++;}
	void			SetAskedCount( uint32 m_cInAsked)	{m_dwAskedCount = m_cInAsked;}
	void			FlushSendBlocks();			// call this when you stop upload, or the m_pRequestSocket might be not able to send
	void			SetLastUpRequest()			{m_dwLastUpRequest = ::GetTickCount();}
	uint32			GetLastUpRequest()			{return m_dwLastUpRequest;}
	void			UDPFileReasked();

	uint32			GetSessionUp()				{return m_dwTransferredUp - m_dwCurSessionUp;}
	void			ResetSessionUp()			{m_dwCurSessionUp = m_dwTransferredUp;}
    void            SetLastGotULData()			{ m_dwLastGotULDataTime = ::GetTickCount(); };
    DWORD           GetLastGotULData()			{ return m_dwLastGotULDataTime; };
	uint32			GetCurrentlyUploadingPart();
#ifdef LATER
	uint32			GetCurrentlyDownloadingPart();
#endif LATER

//	Remote client's parts
	void			ProcessUpFileStatus(char* packet,uint32 size);
	uint16			GetUpPartCount()			{return m_uUpPartCount;}
	uint16			GetAvailUpPartCount()		{return m_uAvailUpPartCount;}
	void			DrawUpStatusBar(CDC* dc, RECT* rect, bool bFlat);
    void            CheckForUpPartsInfo();

//	File request counter
	byte			GetFailedFileRequests()		{return m_byteFailedFileRequestsCount;}
	void			UpdateFailedFileRequests()	{m_byteFailedFileRequestsCount++;}
	void			ResetFailedFileRequests()	{m_byteFailedFileRequestsCount = 0;}

//	Download
	uint32			GetAskedCountDown()			{return m_dwDownAskedCount;}
	void			AddAskedCountDown()			{m_dwDownAskedCount++;}
	void			SetAskedCountDown(uint32 m_cInDownAsked)				{m_dwDownAskedCount = m_cInDownAsked;}
	EnumDLQState	GetDownloadState()			{return m_eDownloadState;}
	void			SetDownloadState(EnumDLQState eNewState);
	uint32			GetLastAskedTime()			{return m_dwLastAskedTime;}
	void			ResetLastAskedTime()		{m_dwLastAskedTime = 0;}
	bool			IsPartAvailable(uint16 iPart)	{return	( (iPart >= m_uPartCount) || (!m_pbytePartStatuses) )? 0:m_pbytePartStatuses[iPart];}
	byte*			GetPartStatus()				{return m_pbytePartStatuses;}
	uint32			GetDownloadDataRate()		{return m_dwDownDataRate;}
	uint16			GetRemoteQueueRank()		{return m_uRemoteQueueRank;}
	int   			GetDifference()    			{return m_iDeltaQueueRank;} //cendre-eau
	void			SetRemoteQueueFull(bool flag)	{m_bRemoteQueueFull = flag;}
	bool			IsRemoteQueueFull()			{return m_bRemoteQueueFull;}
	void			SetRemoteQueueRank(uint16 nr, bool bUpdateGUI = true);
	void			DrawStatusBar(CDC* dc, RECT* rect, bool onlygreyrect, bool  bFlat);
	bool			AskForDownload();
	void			SendFileRequest();
//	void			SendFileRequestForUpload();
	void			ProcessFileInfo(char* packet,uint32 size);
	void			ProcessFileStatus(CSafeMemFile *packetStream, CPartFile *pReqPartFile, bool bUDPPacket = false);
	void			ProcessHashSet(char* packet,uint32 size);
	bool			AddRequestForAnotherFile(CPartFile* file);
	void			RemoveRequestForAnotherFile(CPartFile* pPartFile);
	void			SendBlockRequests();
	void			ProcessBlockPacket(char *ucPacketBuf, uint32 dwPacketSize, bool bPacked = false);
	uint32			CalculateDownloadRate();
//	void			ResetMeasurementPeriodTimer(uint32 time)	{m_dwLastDataRateCalculationTime = time;}
	uint16			GetAvailablePartCount();
	bool			SwapToAnotherFile(CPartFile* SwapTo, bool bIgnoreReasktime = false);
	uint32			GetQueueWaitTime()		{return ::GetTickCount() - m_dwQueueWaitTime;}
	void			StartQueueWaitTimer()	{m_dwQueueWaitTime = ::GetTickCount();}
	void			UDPReaskACK(uint16 nNewQR);
	void			UDPReaskFNF();
	void			UDPReaskForDownload();
	bool			IsSourceRequestAllowed();
	uint32			GetRemoteScore();
	uint16			GetRemoteRatio();
	double			GetRemoteBaseModifier();
	bool			IsRemoteInfoAvaiable(){return m_bRemoteInfoAvailable;}
	uint16			GetRemoteClientSpeed(){return m_uRemoteClientSpeed;}
	bool			IsCompleteSource()			{return m_bCompleteSource;}
	uint16			GetUpCompleteSourcesCount()	{return m_uUpCompleteSourcesCount;}
	uint32			GetUpCompleteSourcesTime()	{return m_dwUpCompleteSourcesTime;}
	void			SetUpCompleteSourcesCount(uint16 n)	{m_uUpCompleteSourcesCount= n; m_dwUpCompleteSourcesTime= ::GetTickCount();}

	CString			GetDownloadStateAsString(void);

	uint16			GetLastDownPartAsked() {return m_uLastPartAsked;};
	void			SetLastDownPartAsked(uint16 uLastPart) { m_uLastPartAsked = uLastPart; };

//	Returns the count of available parts that this user has but we not have
	uint16			GetAvailableOfNeededPartsCount(); //added by Mr.Fry

//	chat
	EnumChatStates		GetChatState()				{return m_eChatState;}
	void				SetChatState(EnumChatStates nNewS)	{m_eChatState = nNewS;}

//	File Comment
	CString				GetFileComment()         	{return m_strComment;}
	void		   		SetFileComment(const CString& desc) 	{m_strComment = desc;}
	EnumPartFileRating	GetFileRating()            	{return m_eRating;}
	void		   		SetFileRating(EnumPartFileRating eNewRating) 	{m_eRating = eNewRating;}

	void			ShowDownloadingParts(std::_Bvector& parts);
	uint32			GetAwayMessageResendCount()				{return m_dwAwayMessageResendCount;};
	void			SetAwayMessageResendCount(uint32 newVal) {m_dwAwayMessageResendCount = newVal;};
	void			UpdateDisplayedInfo(boolean force=false);

    int             GetFileListRequested() { return m_iFileListRequested; }
    void            SetFileListRequested(int iFileListRequested) { m_iFileListRequested = m_iFileListRequestedSave = iFileListRequested; }

	bool			IsOnLAN()						{return m_bIsOnLan;};

	CString			GetUploadFileInfo();

	uint32			GetL2HACTime()							{return m_dwL2HACTime ? (m_dwL2HACTime - L2HAC_CALLBACK_PRECEDE) : 0;}
	void			SetLastL2HACExecution(uint32 m_set_to = 0)	{m_dwLastL2HACExec = m_set_to ? m_set_to : ::GetTickCount();}
	uint32			GetLastL2HACExecution()					{return m_dwLastL2HACExec;}
	void			EnableL2HAC()								{m_bL2HACEnabled = true;}
	void			DisableL2HAC()								{m_bL2HACEnabled = false;}
	bool			IsL2HACEnabled()							{return m_bL2HACEnabled;}

	HICON			GetClientInfo4Tooltips(CString &strInfo, bool bForUpload = false);
	void			UpdateLastBlockReceivedTime()		{m_dwLastBlockReceived = ::GetTickCount();}

	void			ClearPendingBlocksList(void);
	void			ClearDownloadBlocksList(void);

	CString			GetCountryName() const;
	int				GetCountryIndex() const			{return m_uUserCountryIdx;}
	void			ResetIP2Country();

private:
//	base
	void			Init();
	bool			ProcessHelloTypePacket(CSafeMemFile& data);
	void			SendHelloTypePacket(CMemFile& data);
	double			GetClientSpeed();

	int				unzip(Pending_Block_Struct *pRequestedBlock, BYTE *pbyteZipped, uint32 dwLenZipped, uint32 *pdwLenUnzipped);

//	upload
	void			CreateStandardPackets(byte* data,uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF = true);
	void			CreatePackedPackets(byte* data,uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF = true);
	void			SplitAndAddPacketsToSendQueue(Packet* packet, bool bFromPF);
};
